// Ryan Schmidt - rms@unknownroad.com
// Copyright (c) 2005-2010
// Distributed under the Boost Software License, Version 1.0.
// http://www.boost.org/LICENSE_1_0.txt


#include "stdafx.h"
#include "BlobShop.h"
#include "ParameterEditor.h"
#include "SystemConstants.h"

#include "CappedLinearSweepSceneObject.h"


class DraggableFloatParamButtonHandler : public rmsui::HUDButtonEventHandler
{
public:
	DraggableFloatParamButtonHandler( rmsui::GenericFloatParameter * pParameter,
		float fMinDeltaScale = 0.1f, float fMaxDeltaScale = 0.5f, float fFalloffRange = 1.0f)
	{
		m_pParameter = pParameter;
		m_fMinDeltaScale = fMinDeltaScale;
		m_fFalloffRange = fFalloffRange;
		m_fMaxDeltaScale = fMaxDeltaScale - fMinDeltaScale;
	}
	~DraggableFloatParamButtonHandler() {}

	void SetTarget( rmsui::GenericFloatParameter * pParameter ) {
		m_pParameter = pParameter;
	}

	virtual void OnDragDelta( const Wml::Vector2f & vDelta ) 
	{
		if ( m_pParameter ) {
			float fCurValue = m_pParameter->GetValueFloat();
			float fMinValue = m_pParameter->GetMinValue();

			float fDeltaScale = m_fMinDeltaScale;
			float fDist = fabsf(fCurValue - fMinValue);
			if ( fDist < m_fFalloffRange ) {
				fDist /= m_fFalloffRange;
				float fValue = (1.0f - fDist*fDist);
				if ( fValue > 0 ) {
					fValue = fValue*fValue*fValue;
					fDeltaScale += (1.0f - fValue) * m_fMaxDeltaScale;
				}

			} else {
				fDeltaScale += m_fMaxDeltaScale;
			}

			float fDelta = vDelta.X() * fDeltaScale + vDelta.Y() * fDeltaScale;
			m_pParameter->SetValueFloat( 
				m_pParameter->GetValueFloat() + fDelta );
		}
	}

protected:
	rmsui::GenericFloatParameter * m_pParameter;
	float m_fMinDeltaScale;
	float m_fMaxDeltaScale;
	float m_fFalloffRange;
};


#include "SurfaceManager.h"
class PolygonizerResolutionButtonHandler : public rmsui::HUDButtonEventHandler
{
public:
	PolygonizerResolutionButtonHandler(CRenderingManager * pManager) {m_pRenderManager = pManager;}
	~PolygonizerResolutionButtonHandler() {}

	void SetRenderManager(CRenderingManager * pManager) {m_pRenderManager = pManager;}

	virtual void OnDragDelta( const Wml::Vector2f & vDelta ) 
	{
		float fDelta = vDelta.X() * 0.2f + vDelta.Y() * 0.2f;
		if ( fDelta > 0 )
			m_pRenderManager->GetSurfaceManager()->IncreaseSurfacingAccuracy( 1.0f - fabsf(fDelta), true );
		else
			m_pRenderManager->GetSurfaceManager()->DecreaseSurfacingAccuracy( 1.0f + fabsf(fDelta), true );

	}
	CRenderingManager * m_pRenderManager;
};









ParameterEditor * ParameterEditor::s_pSingleton = NULL;

ParameterEditor * ParameterEditor::Get()
{
	if ( s_pSingleton == NULL )
		s_pSingleton = new ParameterEditor();
	return s_pSingleton;
}

ParameterEditor::ParameterEditor()
{
	m_pSelectionParamToolbar = new rmsui::HUDToolbar( "SelectionParameterToolbar", false, rmsui::HUDToolbar::Vertical );
	m_pSelectionParamToolbar->Translate( -1.0f, 0.0f, false );
	m_pSelectionParamToolbar->SetDockMode( rmsui::HUDItem::DockLeft );

	m_pExtraParamObject = NULL;

	InitializeSelectionParamButtons();

	m_pRenderManager = NULL;
	InitializeGlobalParamToolbar();
}




void ParameterEditor::SetRenderManager( CRenderingManager * pManager ) 
{ 
	m_pRenderManager = pManager; 
	m_pMeshResolutionButtonHandler->SetRenderManager( pManager) ;
}



/*
 * Current Selection Parameters
 */


void ParameterEditor::OnTreeSelectionChanged( const std::vector<rmsui::SceneObject *> & newSelection )
{
	SetSelection( newSelection );
}

void ParameterEditor::OnMousePickSelectionChanged( const std::vector<rmsui::SceneObject *> & newSelection )
{
	SetSelection( newSelection );
}

void ParameterEditor::SetSelection( const std::vector< rmsui::SceneObject * > & vSceneObjects )
{
	m_pSelectionParamToolbar->ClearButtons();

	if ( vSceneObjects.size() != 1 )
		return;
	UpdateSelectionParamToolbar( vSceneObjects[0] );

	if ( m_pExtraParamObject != NULL )
		UpdateSelectionParamToolbar( m_pExtraParamObject );
	m_pExtraParamObject = NULL;
}


void ParameterEditor::SetExtraParamObject(rmsui::SceneObject * pObject)
{
	m_pExtraParamObject = pObject;
}

void ParameterEditor::UpdateSelectionParamToolbar( rmsui::SceneObject * pObject )
{
	const std::vector<rmsui::IParameter *> & vParams = pObject->GetParameterSet().GetParameters();
	size_t nCount = vParams.size();
	for ( unsigned int i = 0; i < nCount; ++i ) {

		CString sName = vParams[i]->GetName();
		DoParameterNameRewritingHacks( pObject, sName );

		ButtonInfo buttonInfo = FindButtonByName( sName );
		if ( buttonInfo.pButton != NULL ) {

			switch  ( vParams[i]->GetType() ) {
				case rmsui::IParameter::FloatParameter:
					buttonInfo.pHandler->SetTarget( (rmsui::GenericFloatParameter * ) vParams[i] );
					break;
				default:
					buttonInfo.pHandler->SetTarget(NULL);
					break;
			}

			m_pSelectionParamToolbar->AppendButton( buttonInfo.pButton );
		}

	}
}

ParameterEditor::ButtonInfo ParameterEditor::FindButtonByName( const char * pName )
{
	std::string str(pName);
	ButtonNameMap::iterator found( m_buttonMap.find(str) );
	if ( found == m_buttonMap.end() )
		return ButtonInfo(NULL,NULL);		// dummy button?
	return (*found).second;
}




/*
 * initialilze global param buttons
 */
void ParameterEditor::InitializeGlobalParamToolbar()
{
	m_pGlobalParamToolbar = new rmsui::HUDToolbar( "GlobalParameterToolbar", false, rmsui::HUDToolbar::Horizontal );
	m_pGlobalParamToolbar->Translate( -1.0f, 0.95f, false );
	m_pGlobalParamToolbar->SetDockMode( rmsui::HUDItem::DockLeft );		// really want to dock topleft (?)

	static const float fButtonSize = 0.125f;	
	
	rmsui::HUDFileIcon * pIcon = new rmsui::HUDFileIcon(CSystemConstants::GetIconPath() + "meshResolution.png");
	m_pMeshResolutionButton = new rmsui::HUDButton("MeshResButton", pIcon, true, false);
	m_pMeshResolutionButton->Bounds() = Wml::AxisAlignedBox2f( 0.0f, fButtonSize, 0.0f, fButtonSize );

	m_pMeshResolutionButtonHandler = new PolygonizerResolutionButtonHandler( m_pRenderManager );
	m_pMeshResolutionButton->AddListener( m_pMeshResolutionButtonHandler );

	m_pGlobalParamToolbar->AppendButton( m_pMeshResolutionButton );
}



void ParameterEditor::InitializeSelectionParamButtons()
{
	static const float fButtonSize = 0.125f;

	// create buttons
	rmsui::HUDFileIcon *pIcon = new rmsui::HUDFileIcon(CSystemConstants::GetIconPath() + "sweepResize.png");
	m_pWidthButton = new rmsui::HUDButton("WidthButton", pIcon, true, false);
	m_pWidthButton->Bounds() = Wml::AxisAlignedBox2f( 0.0f, fButtonSize, 0.0f, fButtonSize );

	pIcon = new rmsui::HUDFileIcon(CSystemConstants::GetIconPath() + "blendPower.png");
	m_pBlendPowerButton = new rmsui::HUDButton("BlendPowerButton", pIcon, true, false);
	m_pBlendPowerButton->Bounds() = Wml::AxisAlignedBox2f( 0.0f, fButtonSize, 0.0f, fButtonSize );

	pIcon = new rmsui::HUDFileIcon(CSystemConstants::GetIconPath() + "surfacePointsLine.png");
	m_pTaperButton = new rmsui::HUDButton("TaperButton", pIcon, true, false);
	m_pTaperButton->Bounds() = Wml::AxisAlignedBox2f( 0.0f, fButtonSize, 0.0f, fButtonSize );

	pIcon = new rmsui::HUDFileIcon(CSystemConstants::GetIconPath() + "sphereRadius.png");
	m_pRadiusButton = new rmsui::HUDButton("RadiusButton", pIcon, true, false);
	m_pRadiusButton->Bounds() = Wml::AxisAlignedBox2f( 0.0f, fButtonSize, 0.0f, fButtonSize );

	pIcon = new rmsui::HUDFileIcon(CSystemConstants::GetIconPath() + "smoothDiffSharpness.png");
	m_pSharpnessButton = new rmsui::HUDButton("SharpnessButton", pIcon, true, false);
	m_pSharpnessButton->Bounds() = Wml::AxisAlignedBox2f( 0.0f, fButtonSize, 0.0f, fButtonSize );

	// add handlers for each button
	m_pWidthButtonHandler = new DraggableFloatParamButtonHandler( NULL );
	m_pWidthButton->AddListener( m_pWidthButtonHandler );

	m_pBlendPowerButtonHandler = new DraggableFloatParamButtonHandler( NULL, 0.1f, 2.0f );
	m_pBlendPowerButton->AddListener( m_pBlendPowerButtonHandler );

	m_pTaperButtonHandler = new DraggableFloatParamButtonHandler( NULL );
	m_pTaperButton->AddListener( m_pTaperButtonHandler );

	m_pRadiusButtonHandler = new DraggableFloatParamButtonHandler( NULL, 0.1f, 0.5f );
	m_pRadiusButton->AddListener( m_pRadiusButtonHandler );

	m_pSharpnessButtonHandler = new DraggableFloatParamButtonHandler( NULL, 0.05f, 0.1f );
	m_pSharpnessButton->AddListener( m_pSharpnessButtonHandler );

	// add buttons to map
	m_buttonMap[ std::string("Width") ] = ButtonInfo(m_pWidthButton, m_pWidthButtonHandler);
	m_buttonMap[ std::string("Length") ] = ButtonInfo(m_pWidthButton, m_pWidthButtonHandler);

	m_buttonMap[ std::string("Blend Power") ] = ButtonInfo(m_pBlendPowerButton, m_pBlendPowerButtonHandler);;
	m_buttonMap[ std::string("Taper") ] = ButtonInfo(m_pTaperButton, m_pTaperButtonHandler);;
	m_buttonMap[ std::string("Radius") ] = ButtonInfo(m_pRadiusButton, m_pRadiusButtonHandler);
	m_buttonMap[ std::string("Sharpness") ] = ButtonInfo(m_pSharpnessButton, m_pSharpnessButtonHandler);
}



void ParameterEditor::DoParameterNameRewritingHacks(rmsui::SceneObject * pObject, CString & paramName )
{
	rmsui_bt::CappedLinearSweepSceneObject * pSweepSO = 
		dynamic_cast<rmsui_bt::CappedLinearSweepSceneObject *>( pObject );
	if ( pSweepSO != NULL ) {
		if ( pSweepSO->GetCapStyleParameter()->GetValueInteger() == 1 ) {
			if ( paramName == "Falloff Length" )
				paramName = "Length";
			else if ( paramName == "Length" )
				paramName = "IgnoreMe";
		} 
	}
}