// Ryan Schmidt - rms@unknownroad.com
// Copyright (c) 2005-2010
// Distributed under the Boost Software License, Version 1.0.
// http://www.boost.org/LICENSE_1_0.txt


// RenderingManager.cpp : implementation file
//

#include "stdafx.h"
#include "BlobShop.h"
#include "RenderingManager.h"
#include ".\renderingmanager.h"

#include <gl/GL.h>
#include <gl/GLU.h>

#include <limits>

#include "rmsprofile.h"

#include <BlobTreeStatistics.h>
#include <WmlExtVectorUtil.h>
#include "BlobTreeNodeSceneObject.h"

#include <Image.h>
#include <Texture.h>

// CRenderingManager

IMPLEMENT_DYNAMIC(CRenderingManager, CWnd)
CRenderingManager::CRenderingManager()
	: m_PointSetRendererDialog(this, this),
	  m_LightingControlDialog(this, this)
{
	m_pCamera = NULL;
	m_cBackgroundColor = Wml::Vector3f(1.0f, 1.0f, 1.0f);

	m_bDrawCartesianAxes = true;

	// set up lighting
	m_eLightingMode = DiffuseOnly;
	for (unsigned int i = 0; i < 8; ++i) {
		m_Lights[i].LightIndex() = i;
		m_Lights[i].On() = false;
	}
	m_fAmbientLight = 0.1f;
		
	m_bCuttingPlaneEnabled = false;

	// set up defaults for mesh rendering
	m_bDrawMesh = true;
	m_bDrawMeshNormals = false;
	m_bDrawMeshWireframe = false;
	m_eMeshColorMode = SelectionColor; //ConstantColor;
	m_eMeshShadingMode = SmoothShading;
	m_bMeshColorsValid = false;
	m_cMeshColor = Wml::Vector3f(0.5f, 1.0f, 0.5f);
	m_cWireframeColor = Wml::Vector3f(1.0f, 0.0f, 0.0f);
	m_cNormalColor = Wml::Vector3f(0.0f, 1.0f, 0.0f);
	m_fNormalLength = 0.05f;

	// selection surface
	m_bDrawSelectionSurface = false;
	m_bCurSelectionMeshValid = false;
	m_SelectionSurfacePolygonizer.SetConvergenceType( rmsimplicit::CPortGG4Polygonizer3D::BinarySearch_Converge );
	m_SelectionSurfacePolygonizer.SetConvergenceSteps( 10 );
	m_SelectionSurfacePolygonizer.SetGridResolution( 25 );

	// point set stuff
	m_bDrawSharpEdges = false;
	m_bPointSetValid = false;
	m_fSilhouetteWidth = 2.0f;
	m_fStippleSize = 2.0f;
	m_cSilhouetteColor = Wml::Vector3f::ZERO;

	m_bDrawAllPoints = false;
	m_bDrawLightCulledPoints = false;
	m_bDrawSuggestiveContours = false;
	m_bDrawSilhouette = false;
	m_bDrawZFillMesh = true;
	m_bDrawDiscZFill = false;
	m_bFreezeSilhouette =  false;

	// surface analysis
//	m_SurfaceAnalyzer.SetPolygonizer( &m_DynamicPolygonizer );

	// set up message pipe from child texture manager
	m_texManager.RegisterChangeListener(this);
	m_eCurParamType = ITextureManagerListener::NoTextureCoords;
	m_bVertexColorOverride = false;
}

CRenderingManager::~CRenderingManager()
{
	m_PointSetRendererDialog.DestroyWindow();
}


void CRenderingManager::OnCreateGL()
{
	m_texManager.SetCheckerboardTexture();
}

void CRenderingManager::Render()
{
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	glClearColor(m_cBackgroundColor.X(), m_cBackgroundColor.Y(), m_cBackgroundColor.Z(), 0.0);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	// initialize lighting
	InitializeLighting();

	// set camera params (does perspective setup and lookat)
	m_pCamera->Update();

	// draw axes
	if (m_bDrawCartesianAxes)
		DrawCartesianAxes(1.0f);

	bool bDoLineDrawing = m_bDrawSilhouette || m_bDrawSuggestiveContours || 
						  m_bDrawAllPoints || m_bDrawLightCulledPoints || 
						  m_bDrawSharpEdges || m_bDrawMeshWireframe;

	if ( m_bDrawZFillMesh && bDoLineDrawing && !(m_bDrawMesh && !m_bDrawMeshWireframe) )
		RenderMeshesAsZFill();

	if (m_bDrawMesh) {

		std::vector< IsoSurfaceMesh * > vMeshes;
		m_pSurfaceManager->GetSurfaceList(vMeshes);
		size_t nCount = vMeshes.size();

		for ( unsigned int i = 0; i < nCount; ++i ) {
			IsoSurfaceMesh * pSurface = vMeshes[i];

			if ( m_bCuttingPlaneEnabled ) {

				glEnable(GL_CLIP_PLANE0);
				Wml::Vector3f vNormal( m_vCuttingPlaneFrame.Axis( Wml::Frame3f::AxisZ ) );
				double dClipPlane[4] = {
					-vNormal.X(), -vNormal.Y(), -vNormal.Z(), vNormal.Dot( m_vCuttingPlaneFrame.Origin() ) };
				glClipPlane( GL_CLIP_PLANE0, dClipPlane );


				// draw back-facing polys
				glEnable(GL_CULL_FACE);
				glCullFace(GL_FRONT);

				glColor3fv(m_cMeshColor);
				RenderMeshAsCuttingPlane(pSurface);

				glCullFace(GL_BACK);
			
				//glLightModeli( GL_LIGHT_MODEL_TWO_SIDE, 1 );

				RenderSurface(pSurface, bDoLineDrawing);

				glDisable(GL_CULL_FACE);


				// flip orientation
				dClipPlane[0] *= -1.0;
				dClipPlane[1] *= -1.0;
				dClipPlane[2] *= -1.0;
				dClipPlane[3] = -vNormal.Dot( m_vCuttingPlaneFrame.Origin() );
				glClipPlane( GL_CLIP_PLANE0, dClipPlane );

				//glEnable(GL_POLYGON_STIPPLE);
				//GLubyte mask[32*4];
				//for ( int i = 0; i < 32*4; ++i )
				//	mask[i] = ( (i/4) % 2 == 0 ) ? 0x00 : 0xFF;
				//glPolygonStipple( mask );

				glEnable(GL_CULL_FACE);
				glEnable(GL_BLEND);
				glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

				Wml::Vector3f vTmp = m_cMeshColor;
				m_cMeshColor = Wml::Vector3f(0.7f, 0.7f, 1.0f);

				RenderSurface(pSurface, bDoLineDrawing, 0.4f);

				m_cMeshColor = vTmp;

				glDisable(GL_CLIP_PLANE0);
				glDisable(GL_POLYGON_STIPPLE);
				glDisable(GL_CULL_FACE);
				glDisable(GL_BLEND);

			} else
				RenderSurface(pSurface, bDoLineDrawing);

		}

	}


	glPushAttrib(GL_ENABLE_BIT);
	//glDisable(GL_DEPTH_TEST);

	if (bDoLineDrawing) {

		ValidatePointSet();
		m_PointSetRenderer.NewFrame( m_pCamera );

		if ( m_bDrawDiscZFill )
			RenderDiscsAsZFill();

		if (m_bDrawSharpEdges)
			RenderSharpEdges();

		if (m_bDrawSilhouette || m_bDrawSuggestiveContours)
			RenderContours();

		if (m_bDrawAllPoints)
			RenderAllPoints();

		if (m_bDrawLightCulledPoints)
			RenderLightCulledPoints();
	}


	if (m_bDrawSelectionSurface) {
		ValidateSelectionSurface();
		if ( m_bCurSelectionMeshValid )
			RenderSelectionSurface();
	}

	glPopAttrib();


	// render surface analyzer
//	m_SurfaceAnalyzer.Render();

	// render texture manager
	m_texManager.Render();

	//if ( m_bCuttingPlaneEnabled )
	//	DrawCuttingPlane();

	//if ( m_eSurfaceType == BoundedMC )
	//	m_DynamicPolygonizer.RenderData();
}

void CRenderingManager::DrawCartesianAxes( float fLength )
{
	glPushAttrib(GL_ENABLE_BIT);
	glDisable(GL_LIGHTING);

	// draw cartesian axes
	glBegin(GL_LINES);
	// red x axis
	glColor3f(1.f,0.f,0.f);
	glVertex3f(0.0, 0.0, 0.0);
	glVertex3f(fLength, 0.0, 0.0);
	// green y axis
	glColor3f(0.f,1.f,0.f);
	glVertex3f(0.0, 0.0, 0.0);
	glVertex3f(0.0, fLength, 0.0);
	// blue z axis
	glColor3f(0.f,0.f,1.f);
	glVertex3f(0.0, 0.0, 0.0);
	glVertex3f(0.0, 0.0, fLength);
	glEnd();

	glPopAttrib();
}



void CRenderingManager::InvalidateSelectionSurface()
{
	m_bCurSelectionMeshValid = false;
}

void CRenderingManager::ValidateSelectionSurface()
{
	if ( m_bCurSelectionMeshValid == true )
		return;

	// get selection...
	if ( m_vSelectedObjects.size() != 1 )
		return;
	rmsui_bt::BlobTreeNodeSceneObject * pObject = 
		dynamic_cast<rmsui_bt::BlobTreeNodeSceneObject *>( m_vSelectedObjects[0] );
	if ( pObject == NULL )
		return;

	// figure out resolution factor
	Wml::AxisAlignedBox3f bounds1;
	pObject->GetBlobTreeNode()->FieldBounds(bounds1);

	m_SelectionSurfacePolygonizer.SetSurfaceIsoValue( 0.5f );
	m_SelectionSurfacePolygonizer.SetCellSize( bounds1, 0.075f );

	m_SelectionSurfacePolygonizer.SetOutputMesh( &m_curSelectionMesh );
	m_SelectionSurfacePolygonizer.SetScalarField( pObject->GetBlobTreeNode() );

	m_SelectionSurfacePolygonizer.Polygonize();

	m_bCurSelectionMeshValid = true;
}


void CRenderingManager::ComputeMeshColors()
{
	// [RMS: CONV   (should be done in surface manager!)
	if (m_bMeshColorsValid == true)
		return;
	//switch (m_eMeshColorMode) {
	//	case VertexErrorColor:
	//		ColorMeshByVertexError();
	//		break;
	//	case ApproxErrorColor:
	//		ColorMeshByApproxError();
	//		break;
	//}
	
	m_bMeshColorsValid = true;
}



void CRenderingManager::RenderSurface( IsoSurfaceMesh * pSurface, bool bDoPolyOffset, float fAlpha )
{
	glPushAttrib(GL_ENABLE_BIT | GL_POLYGON_BIT);

	glEnable(GL_NORMALIZE);

	if (bDoPolyOffset) {
		glEnable(GL_POLYGON_OFFSET_FILL );
		glPolygonOffset(1.0f, 1.0f);
	}

	if (m_bDrawMeshWireframe) {
		glDisable(GL_LIGHTING);
		glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
		glColor4f( m_cWireframeColor.X(), m_cWireframeColor.Y(), m_cWireframeColor.Z(), fAlpha );

	} else {
		if (m_eLightingMode == WarmToCoolDiffuse)
			glColor4f( 1.0f, 1.0f, 1.0f, fAlpha);
		else
			glColor4f( m_cMeshColor.X(), m_cMeshColor.Y(), m_cMeshColor.Z(), fAlpha );
		glEnable(GL_LIGHTING);
	}

	// make sure colors are valid
	if (m_eMeshColorMode != ConstantColor)
		ComputeMeshColors();

	// set shading mode
	bool bUseTriangleNormals = false;
	if (m_eMeshShadingMode == SmoothShading)
		glShadeModel(GL_SMOOTH);
	else {
		glShadeModel(GL_FLAT);
		bUseTriangleNormals = true;
	}

	// draw mesh
	if (m_eMeshColorMode != ConstantColor || m_bVertexColorOverride == true){
		if ( m_eMeshColorMode == VertexErrorColor || m_eMeshColorMode == ApproxErrorColor )
			glDisable(GL_LIGHTING);

		if ( pSurface->GetPolygonizerType() == IsoSurfaceMesh::DynamicMC )
			pSurface->GetVFMesh()->DrawGL(true);
		else {
			TriangleMeshRenderer r( pSurface->GetSimpleMesh() );
			r.DrawImmediateWithColor(bUseTriangleNormals);
		}

	}else {

		bool bHasTexCoords = (m_eCurParamType != ITextureManagerListener::NoTextureCoords);
		if ( bHasTexCoords )
			m_texManager.EnableTexture();


		if ( pSurface->GetPolygonizerType() == IsoSurfaceMesh::StaticMC ) {
			TriangleMeshRenderer r( pSurface->GetSimpleMesh() );
			if ( bHasTexCoords )
				glColor4f(m_cMeshColor.X(), m_cMeshColor.Y(), m_cMeshColor.Z(), fAlpha);

			r.DrawImmediate(bUseTriangleNormals, false, bHasTexCoords, 
				( m_eCurParamType == ITextureManagerListener::PerVertexTexCoords ) ? true :  false );
		} else {
			pSurface->GetVFMesh()->DrawGL(true);
		}

//		if ( m_bTexCoordsValid )
			m_texManager.DisableTexture();

	}

	// draw mesh normals
	if (m_bDrawMeshNormals) {
		glDisable(GL_LIGHTING);
		glColor3f( m_cNormalColor.X(), m_cNormalColor.Y(), m_cNormalColor.Z() );

		if ( pSurface->GetPolygonizerType() == IsoSurfaceMesh::StaticMC ) {
			TriangleMeshRenderer r( pSurface->GetSimpleMesh() );
			r.DrawNormals( m_fNormalLength );
			//r.DrawCurvatureDirections( m_fNormalLength, false, true, m_pCurField );
		} 
	}

	glPopAttrib();
}


void CRenderingManager::RenderSelectionSurface()
{
	glPushAttrib(GL_ENABLE_BIT | GL_POLYGON_BIT | GL_LIGHTING_BIT );

	glEnable(GL_NORMALIZE);

	// create triangle mesh renderer
	TriangleMeshRenderer meshRenderer(&m_curSelectionMesh);

	glDisable(GL_DEPTH_TEST);
	glEnable(GL_CULL_FACE);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	//glEnable(GL_POLYGON_STIPPLE);
	//GLubyte mask[32*4];
	//for ( int i = 0; i < 32*4; ++i )
	//	mask[i] = ( (i/4) % 2 == 0 ) ? 0x00 : 0xFF;
	//glPolygonStipple( mask );

	glColor4f(0.0f, 0.25f, 0.25f, 0.25f);
	float fSpecular[4] = {0,0,0,0};
	glMaterialfv( GL_FRONT_AND_BACK, GL_SPECULAR, fSpecular );

	meshRenderer.DrawImmediate();

	glPopAttrib();
}





/*
 * Lighting
 */
void CRenderingManager::InitializeLighting()
{
	// turn off all lights
	for (unsigned int i = 0; i < 8; ++i) 
		m_Lights[i].On() = false;

	switch(m_eLightingMode) {
		case DiffuseOnly:
			SetDiffuseOnlyLighting();
			break;
		case DiffuseAndSpecular:
			SetDiffuseAndSpecularLighting();
			break;
		case WarmToCoolDiffuse:
			SetWarmToCoolDiffuseLighting();
			break;
		case NPRLighting:
			SetNPRLighting();
			break;
	}

	if (m_eLightingMode != NoLighting) {
		glEnable(GL_LIGHTING);

		for (unsigned int i = 0; i < 8; ++i) 
			m_Lights[i].SetupLight();
	} else {
		glDisable(GL_LIGHTING);
	}
}


void CRenderingManager::SetNPRLighting()
{
	GLfloat ambient[] = { 0, 0, 0, 1.0f };
	glLightModelfv(GL_LIGHT_MODEL_AMBIENT, ambient);

	float fSpecular[] = {0.0f, 0.0f, 0.0f, 1.0f};
	glMaterialfv( GL_FRONT_AND_BACK, GL_SPECULAR, fSpecular );

	float fEye = (m_pCamera->GetLocation() - m_pCamera->GetTarget()).Length();
	Wml::Vector3f vUpLeft( -2.0f*fEye, fEye, fEye );
	Wml::Vector3f vUpRight( 2.0f*fEye, fEye, fEye );

	m_Lights[0].On() = true;
	m_Lights[0].Position() = vUpLeft;
	m_Lights[0].Ambient() = Wml::ColorRGB(0.0f, 0.0f, 0.0f);
	m_Lights[0].Diffuse() = Wml::ColorRGB(1.0f, 1.0f, 1.0f);
	m_Lights[0].Specular() = Wml::ColorRGB(0.0f, 0.0f, 0.0f);
}

void CRenderingManager::SetDiffuseOnlyLighting()
{
	// initialize one point light at 10, 10, 10
	GLfloat ambient[] = { m_fAmbientLight, m_fAmbientLight, m_fAmbientLight, 1.0f };
	glLightModelfv(GL_LIGHT_MODEL_AMBIENT, ambient);

	// kill any specular lighting
	float fSpecular[] = {0.0f, 0.0f, 0.0f, 1.0f};
	glMaterialfv( GL_FRONT_AND_BACK, GL_SPECULAR, fSpecular );

	//m_Lights[0].On() = true;
	//m_Lights[0].Position() = Wml::Vector3f(100.0, 100.0, 100.0);
	//m_Lights[0].Ambient() = Wml::ColorRGB(0.2f, 0.2f, 0.2f);
	//m_Lights[0].Diffuse() = Wml::ColorRGB(0.6f, 0.6f, 0.6f);
	//m_Lights[0].Specular() = Wml::ColorRGB(1.0, 1.0, 1.0);

	float fEye = (m_pCamera->GetLocation() - m_pCamera->GetTarget()).Length();
	Wml::Vector3f vUpLeft( -2.0f*fEye, fEye, fEye );
	Wml::Vector3f vUpRight( 2.0f*fEye, fEye, fEye );

	m_Lights[0].On() = true;
	m_Lights[0].Position() = vUpLeft;
	m_Lights[0].Ambient() = Wml::ColorRGB(0.0f, 0.0f, 0.0f);
	m_Lights[0].Diffuse() = Wml::ColorRGB(0.6f, 0.6f, 0.6f);
	m_Lights[0].Specular() = Wml::ColorRGB(0.5f, 0.5f, 0.5f);

	m_Lights[1].On() = true;
	m_Lights[1].Position() = vUpRight;
	m_Lights[1].Ambient() = Wml::ColorRGB(0.0f, 0.0f, 0.0f);
	m_Lights[1].Diffuse() = Wml::ColorRGB(0.5f, 0.5f, 0.5f);
	m_Lights[1].Specular() = Wml::ColorRGB(0.5f, 0.5f, 0.5f);
}

void CRenderingManager::SetDiffuseAndSpecularLighting()
{
	SetDiffuseOnlyLighting();

	float fSpecular[] = {1.0f, 1.0f, 1.0f, 1.0f};
	glMaterialfv( GL_FRONT_AND_BACK, GL_SPECULAR, fSpecular );
	glMaterialf( GL_FRONT_AND_BACK, GL_SHININESS, 40.0f );
}

void CRenderingManager::SetWarmToCoolDiffuseLighting()
{
	// [RMS] this code adapted from thesis of Amy Gooch (google it...)

	// kill any specular lighting
	float fSpecular[] = {0.0f, 0.0f, 0.0f, 1.0f};
	glMaterialfv( GL_FRONT_AND_BACK, GL_SPECULAR, fSpecular );

	GLfloat R_warm, G_warm, B_warm,R_cool, G_cool, B_cool;
	R_warm=207/255.0f; G_warm=207/255.0f; B_warm=145/255.0f;
	R_cool=80/255.0f;  G_cool=80/255.0f;  B_cool=145/255.0f; 

	GLfloat hi_diffuse[] = { (R_warm-R_cool)/2.0f, 
		(G_warm-G_cool)/2.0f,
		(B_warm-B_cool)/2.0f };
	GLfloat lo_diffuse[] = { (R_cool-R_warm)/2.0f,
		(G_cool-G_warm)/2.0f, 
		(B_cool-B_warm)/2.0f };
	//GLfloat hi_position[] = {  1,  1, EYE, 1 };
	//GLfloat lo_position[] = { -1, -1, EYE, 1 };

	GLfloat ambient[] = { 0.5, 0.5, 0.5 };

	glLightModelfv(GL_LIGHT_MODEL_AMBIENT, ambient);

	float fEye = (m_pCamera->GetLocation() - m_pCamera->GetTarget()).Length();
	Wml::Vector3f vUpLeft( -fEye, -fEye, fEye );
	Wml::Vector3f vUpRight( fEye, fEye, fEye );

	m_Lights[0].On() = true;
	m_Lights[0].Position() = vUpLeft;
	m_Lights[0].Ambient() = Wml::ColorRGB(0.0f, 0.0f, 0.0f);
	m_Lights[0].Diffuse() = Wml::ColorRGB( hi_diffuse[0], hi_diffuse[1], hi_diffuse[2] );

	m_Lights[1].On() = true;
	m_Lights[1].Position() = vUpRight;
	m_Lights[1].Ambient() = Wml::ColorRGB(0.0f, 0.0f, 0.0f);
	m_Lights[1].Diffuse() = Wml::ColorRGB( lo_diffuse[0], lo_diffuse[1], lo_diffuse[2] );
}



/*
void CStandard3DView::InitializeCubeMap()
{
	glEnable( GL_TEXTURE_CUBE_MAP_ARB );

	unsigned char * pCubeMap = new unsigned char[512 * 512 * 3];

	// make left face
	for (unsigned int yi = 0; yi < 512; ++yi) {
		for (unsigned int xi = 0; xi < 512; ++xi) {
			pCubeMap[ yi*512*3 + xi*3 ] = (xi > 100 && xi < 200) ? 255 : 0;
			pCubeMap[ yi*512*3 + xi*3 + 1] = 0;
			pCubeMap[ yi*512*3 + xi*3 + 2] = 0;
		}
	}

	glTexImage2D( GL_TEXTURE_CUBE_MAP_POSITIVE_X_ARB,
		0, GL_RGBA8, 512, 512, 0, GL_RGB, GL_UNSIGNED_BYTE, pCubeMap);


	// just do rest of them right now..
	glTexImage2D( GL_TEXTURE_CUBE_MAP_NEGATIVE_X_ARB,
		0, GL_RGBA8, 512, 512, 0, GL_RGB, GL_UNSIGNED_BYTE, pCubeMap);

	glTexImage2D( GL_TEXTURE_CUBE_MAP_POSITIVE_Y_ARB,
		0, GL_RGBA8, 512, 512, 0, GL_RGB, GL_UNSIGNED_BYTE, pCubeMap);
	glTexImage2D( GL_TEXTURE_CUBE_MAP_NEGATIVE_Y_ARB,
		0, GL_RGBA8, 512, 512, 0, GL_RGB, GL_UNSIGNED_BYTE, pCubeMap);

	glTexImage2D( GL_TEXTURE_CUBE_MAP_POSITIVE_Z_ARB,
		0, GL_RGBA8, 512, 512, 0, GL_RGB, GL_UNSIGNED_BYTE, pCubeMap);
	glTexImage2D( GL_TEXTURE_CUBE_MAP_NEGATIVE_Z_ARB,
		0, GL_RGBA8, 512, 512, 0, GL_RGB, GL_UNSIGNED_BYTE, pCubeMap);

	// set up texgen ??

	glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_REFLECTION_MAP_ARB);
    glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_REFLECTION_MAP_ARB);
    glTexGeni(GL_R, GL_TEXTURE_GEN_MODE, GL_REFLECTION_MAP_ARB);
    glEnable(GL_TEXTURE_GEN_S);
    glEnable(GL_TEXTURE_GEN_T);
    glEnable(GL_TEXTURE_GEN_R);


	delete [] pCubeMap;
}
*/


void CRenderingManager::InvalidatePointSet()
{
	m_bPointSetValid = false;
}

void CRenderingManager::ValidatePointSet()
{
	if (m_bPointSetValid == true)
		return;

#if 0
	if ( m_bDrawSelectionSurface && m_vSelectedObjects.size() == 1) {
		m_PointSetRenderer.SetScalarField( ((rmsui_bt::BlobTreeNodeSceneObject *)m_vSelectedObjects[0])->GetBlobTreeNode() );
		m_PointSetRenderer.SetTriangleMesh( & m_curSelectionMesh );
		m_PointSetRenderer.SetSurfaceIsoValue( 0.5f );
		m_PointSetRenderer.MeshChanged();

		m_bPointSetValid = true;
	}
#else
	std::vector< IsoSurfaceMesh * > vMeshes;
	m_pSurfaceManager->GetSurfaceList(vMeshes);
	size_t nCount = vMeshes.size();
	if ( nCount == 1 && vMeshes[0]->GetPolygonizerType() == IsoSurfaceMesh::StaticMC ) {

		m_PointSetRenderer.SetScalarField( m_pCurField );
		m_PointSetRenderer.SetTriangleMesh( vMeshes[0]->GetSimpleMesh() );
		m_PointSetRenderer.SetSurfaceIsoValue( 0.5f );
		m_PointSetRenderer.MeshChanged();

		m_bPointSetValid = true;
	}
#endif
}



void CRenderingManager::RenderMeshesAsZFill()
{
	std::vector< IsoSurfaceMesh * > vMeshes;
	m_pSurfaceManager->GetSurfaceList(vMeshes);
	size_t nCount = vMeshes.size();

	if (m_bDrawMesh == false || m_bDrawMeshWireframe == true) {
		glPushAttrib(GL_ENABLE_BIT);
		glEnable(GL_POLYGON_OFFSET_FILL);
		glPolygonOffset(1.0f, 1.0f);
		glDisable(GL_LIGHTING);
		glColorMask(GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE);

		for ( unsigned int i = 0; i < nCount; ++i ) {
			if ( vMeshes[i]->GetPolygonizerType() == IsoSurfaceMesh::DynamicMC )
				vMeshes[i]->GetVFMesh()->DrawGL(false);
			else {
				TriangleMeshRenderer r( vMeshes[i]->GetSimpleMesh() );
				r.DrawAsZFill();
			}
		}
	
		glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);
		glPopAttrib();
	}

}


rmsui::Texture * pTexture = NULL;
unsigned int nCount = 0;

void CRenderingManager::RenderAllPoints()
{
	//ValidateSurface();
	ValidatePointSet();

	if (!m_bPointSetValid)
		return;

	glPushAttrib( GL_ENABLE_BIT | GL_POINT_BIT | GL_POLYGON_BIT | GL_DEPTH_BUFFER_BIT );

	glPointSize( m_fSilhouetteWidth );
	glColor3f( m_cSilhouetteColor.X(), m_cSilhouetteColor.Y(), m_cSilhouetteColor.Z() );

	//glColorMask(GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE);
	//glDisable(GL_LIGHTING);

	if ( pTexture == NULL ) {
		rmsui::Image img;
		img.Load("C:\\Projects\\BlobShop\\images\\stipple.png");
		pTexture = new rmsui::Texture();
		pTexture->InitializeBuffer( img.Width(), img.Height(), img.Buffer(), 
			img.Channels() == 4 ? GL_RGBA : GL_RGB );

		pTexture->SetTextureInfo( GL_TEXTURE_2D, GL_LINEAR, GL_REPEAT );
		pTexture->TextureEnvMode() = GL_REPLACE;
	}
	pTexture->Enable();

	//glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	//glEnable(GL_BLEND);

	glAlphaFunc( GL_NOTEQUAL, 0.0f );
	glEnable( GL_ALPHA_TEST );

	glColor4f(1.0f, 1.0f, 1.0f, 1.0f);

//	glDisable(GL_DEPTH_TEST);
//	glDepthMask(GL_TRUE);
//	glDepthFunc(GL_ALWAYS);

	//m_PointSetRenderer.RenderDiscs(true, true);
	//m_PointSetRenderer.RenderCulledDiscs( PointSetRenderer::CullByLighting, true, m_pCamera);

	pTexture->Disable();

	//m_PointSetRenderer.RenderPoints( m_pCamera );
	glColor4f(0.0f, 0.0f, 0.0f, 1.0f);
	//m_PointSetRenderer.RenderCulledPoints( PointSetRenderer::CullByLighting, m_pCamera);
	//m_PointSetRenderer.RenderPoints( m_pCamera );

	//m_PointSetRenderer.RenderDiscs_GeometryOnly();
	m_PointSetRenderer.RenderCulledDiscPoints( PointSetRenderer::CullByLighting, m_pCamera);

//	m_PointSetRenderer.RenderDiscPoints();


	glPopAttrib();
}


void CRenderingManager::RenderLightCulledPoints()
{
	//ValidateSurface();
	ValidatePointSet();
	if (!m_bPointSetValid)
		return;

	glPushAttrib( GL_ENABLE_BIT | GL_POINT_BIT | GL_POLYGON_BIT | GL_DEPTH_BUFFER_BIT );
	glDisable(GL_LIGHTING);

	glPointSize( m_fStippleSize );
//	_RMSInfo("Stipple size is %f\n", m_fStippleSize);
	glColor4f(0.0f, 0.0f, 0.0f, 1.0f);

	m_PointSetRenderer.RenderCulledDiscPoints( PointSetRenderer::CullByLighting, m_pCamera);

	glPopAttrib();
}


void CRenderingManager::RenderDiscsAsZFill()
{
	ValidatePointSet();
	if (!m_bPointSetValid)
		return;

	glPushAttrib( GL_ENABLE_BIT | GL_POINT_BIT | GL_POLYGON_BIT );

	// fastest rendering...
	glDisable(GL_LIGHTING);
	glShadeModel(GL_FLAT);

	// enable polygon offset
	glEnable(GL_POLYGON_OFFSET_FILL);
	glPolygonOffset( 5.0f, 5.0f );

	// prevent color writing (just doing Z-fill...)
	glColorMask(GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE);

	// render w/ flat color
	//glColor3f(1.0f, 1.0f, 0.0f);

	m_PointSetRenderer.RenderDiscs_GeometryOnly();

	glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);

	glShadeModel(GL_SMOOTH);

	glPopAttrib();
}





void CRenderingManager::RenderSharpEdges()
{
	std::vector< IsoSurfaceMesh * > vMeshes;
	m_pSurfaceManager->GetSurfaceList(vMeshes);
	size_t nCount = vMeshes.size();
	if ( nCount == 1 && vMeshes[0]->GetPolygonizerType() == IsoSurfaceMesh::StaticMC ) {

		rmsimplicit::TriangleMesh * pMesh = vMeshes[0]->GetSimpleMesh();

		const std::vector< std::pair<int,int> > * pvFeatureEdges = vMeshes[0]->GetFeatureEdges();
		if (pvFeatureEdges == NULL)
			return;

		glPushAttrib(GL_ENABLE_BIT | GL_LINE_BIT);

		glDisable(GL_LIGHTING);

		glLineWidth( m_fSilhouetteWidth );
		glColor3f( m_cSilhouetteColor.X(), m_cSilhouetteColor.Y(), m_cSilhouetteColor.Z() );

		size_t nSize = pvFeatureEdges->size();
		Wml::Vector3f v1,v2;
		glBegin(GL_LINES);
		for (unsigned int i = 0; i < nSize; ++i) {
			const std::pair<int,int> & edge = (*pvFeatureEdges)[i];

			pMesh->GetVertex( edge.first, v1 );
			pMesh->GetVertex( edge.second, v2 );

			glVertex3f( v1.X(), v1.Y(), v1.Z() );
			glVertex3f( v2.X(), v2.Y(), v2.Z() );
		}
		glEnd();


		glPopAttrib();
	}

}


void CRenderingManager::RenderContours()
{
	std::vector< IsoSurfaceMesh * > vMeshes;
	m_pSurfaceManager->GetSurfaceList(vMeshes);
	size_t nCount = vMeshes.size();
	if ( nCount > 0 ) {

		rmsimplicit::TriangleMesh * pTriMesh = vMeshes[0]->GetSimpleMesh();

		if ( !m_bFreezeSilhouette ) {
			m_silFinder.SetCurrentMesh( pTriMesh );
			m_silFinder.SetCurrentCamera( m_pCamera );
			m_silFinder.SetCurrentScalarField( m_pCurField );
			m_silFinder.ComputeLines( m_bDrawSilhouette, m_bDrawSuggestiveContours );
		}

		glPushAttrib(GL_ENABLE_BIT | GL_LINE_BIT | GL_POINT_BIT);

		glHint( GL_LINE_SMOOTH_HINT, GL_NICEST );
		//glEnable( GL_LINE_SMOOTH );
		glDisable(GL_LIGHTING);

		glLineWidth( m_fSilhouetteWidth * 2.0f );
		glColor3f( m_cSilhouetteColor.X(), m_cSilhouetteColor.Y(), m_cSilhouetteColor.Z() );

		//glDisable(GL_DEPTH_TEST);
		glDepthMask(GL_FALSE);

		m_silFinder.RenderLines();

		glDepthMask(GL_TRUE);

		glPopAttrib();
	}
}


/*
 * Message and dialog handling stuff
 */








void CRenderingManager::ComputeStatistics( CRenderingManager::StatisticType eType, const char * pFilename )
{
	// RMS CONV

	//switch(eType) {
	//	case FieldValueErrorAtMeshVertices:
	//		{
	//		rmsbt::CachedScalarFieldWrapper * pWrapper = dynamic_cast<rmsbt::CachedScalarFieldWrapper *>(m_pCurBlobTree);
	//		ASSERT(pWrapper);
	//		rmsbt::BlobTreeNode * pNode = pWrapper->GetRealNode();
	//		rmsbt::BlobTreeStatistics::ComputeFieldValueErrorAtMeshVertices(&m_curMesh, pNode, pFilename);
	//		}
	//		break;

	//	case CacheMemoryUsage:
	//		{
	//		rmsbt::CachedScalarFieldWrapper * pWrapper = dynamic_cast<rmsbt::CachedScalarFieldWrapper *>(m_pCurBlobTree);
	//		ASSERT(pWrapper);
	//		rmsbt::BlobTreeNode * pNode = pWrapper->GetRealNode();
	//		rmsbt::BlobTreeStatistics::ComputeCellAllocationStats(pNode, pFilename);
	//		}
	//		break;

	//	case CacheEvaluations:
	//		{
	//		rmsbt::CachedScalarFieldWrapper * pWrapper = dynamic_cast<rmsbt::CachedScalarFieldWrapper *>(m_pCurBlobTree);
	//		ASSERT(pWrapper);
	//		rmsbt::BlobTreeNode * pNode = pWrapper->GetRealNode();
	//		rmsbt::BlobTreeStatistics::ComputeCellEvaluationStats(pNode, pFilename);
	//		}
	//		break;

	//	default:
	//		ASSERT(0);
	//}
}







void CRenderingManager::ShowPointSetRendererControlDialog( )
{
	m_PointSetRendererDialog.m_fPointSetDensityEdit = m_PointSetRenderer.GetPointSpacing();
	m_PointSetRendererDialog.m_fDiscSizeEdit = m_PointSetRenderer.GetDiscSize();

	m_PointSetRendererDialog.m_fStippleSizeEdit = m_fStippleSize;
	m_PointSetRendererDialog.m_nStippleDensityEdit = m_PointSetRenderer.GetDiscPointSubdivSteps();

	m_PointSetRendererDialog.m_fContourWidth = m_fSilhouetteWidth;

	m_PointSetRendererDialog.m_nDrawAllPointsCheck = (int)m_bDrawAllPoints;
	m_PointSetRendererDialog.m_nDrawLightCulledPointsCheck = (int)m_bDrawLightCulledPoints;

	m_PointSetRendererDialog.m_nDrawSilhouetteContoursCheck = (int)m_bDrawSilhouette;
	m_PointSetRendererDialog.m_nDrawSuggestiveContoursCheck = (int)m_bDrawSuggestiveContours;

	m_PointSetRendererDialog.m_nDrawMeshZFillCheck = (int)m_bDrawZFillMesh;
	m_PointSetRendererDialog.m_nDrawDiscZFillCheck = (int)m_bDrawDiscZFill;

	m_PointSetRendererDialog.m_nFreezeSilhouette = (int)m_bFreezeSilhouette;

	m_PointSetRendererDialog.m_fNPRLowThreshold = m_PointSetRenderer.GetNPRLowThreshold();
	m_PointSetRendererDialog.m_fNPRHighThreshold = m_PointSetRenderer.GetNPRHighThreshold();

	m_PointSetRendererDialog.UpdateDialogValues();
	m_PointSetRendererDialog.ShowWindow(SW_SHOW);
}

void CRenderingManager::OnPointSetRendererDialogApplyButton()
{
	m_PointSetRendererDialog.UpdateClassValues();

	m_bDrawAllPoints = m_PointSetRendererDialog.m_nDrawAllPointsCheck != 0;
	m_bDrawLightCulledPoints = m_PointSetRendererDialog.m_nDrawLightCulledPointsCheck != 0;

	m_bDrawSilhouette = m_PointSetRendererDialog.m_nDrawSilhouetteContoursCheck != 0;
	m_bDrawSuggestiveContours = m_PointSetRendererDialog.m_nDrawSuggestiveContoursCheck != 0;

	m_bDrawZFillMesh = m_PointSetRendererDialog.m_nDrawMeshZFillCheck != 0;
	m_bDrawDiscZFill = m_PointSetRendererDialog.m_nDrawDiscZFillCheck != 0;
	
	m_bFreezeSilhouette = m_PointSetRendererDialog.m_nFreezeSilhouette != 0;

	m_fStippleSize = m_PointSetRendererDialog.m_fStippleSizeEdit;
	m_fSilhouetteWidth = m_PointSetRendererDialog.m_fContourWidth;

	float fSpacing = m_PointSetRendererDialog.m_fPointSetDensityEdit;
	if (fSpacing != m_PointSetRenderer.GetPointSpacing()) {
		m_PointSetRenderer.SetPointSpacing( fSpacing );
		InvalidatePointSet();
	}

	float fDiscSize = m_PointSetRendererDialog.m_fDiscSizeEdit;
	if ( fDiscSize != m_PointSetRenderer.GetDiscSize() ) {
		m_PointSetRenderer.SetDiscSize( fDiscSize );
	}

	int nStippleDensity = m_PointSetRendererDialog.m_nStippleDensityEdit;
	if ( nStippleDensity != m_PointSetRenderer.GetDiscPointSubdivSteps() ) {
		m_PointSetRenderer.SetDiscPointSubdivSteps( nStippleDensity );
	}

	m_PointSetRenderer.SetNPRLowThreshold( m_PointSetRendererDialog.m_fNPRLowThreshold ) ;
	m_PointSetRenderer.SetNPRHighThreshold( m_PointSetRendererDialog.m_fNPRHighThreshold ) ;

	NotifyRenderParametersChanged();
}

void CRenderingManager::ShowLightingControlDialog()
{
	if ( m_eLightingMode == DiffuseOnly )
		m_LightingControlDialog.m_nLightingMode = 0;
	else if (m_eLightingMode == DiffuseAndSpecular )
		m_LightingControlDialog.m_nLightingMode = 1;
	else if (m_eLightingMode == WarmToCoolDiffuse )
		m_LightingControlDialog.m_nLightingMode = 2;
	else
		ASSERT(0);

	if (m_eMeshShadingMode == FlatShading)
		m_LightingControlDialog.m_nShadingMode = 1;
	else 
		m_LightingControlDialog.m_nShadingMode = 0;

	m_LightingControlDialog.m_nShowCoordAxes = (m_bDrawCartesianAxes == true);

	if ( m_eMeshColorMode == SelectionColor )
		m_LightingControlDialog.m_nMeshColorMode = 1;
	else if ( m_eMeshColorMode == VertexErrorColor)
		m_LightingControlDialog.m_nMeshColorMode = 2;
	else if ( m_eMeshColorMode == ApproxErrorColor)
		m_LightingControlDialog.m_nMeshColorMode = 3;
	else
		m_LightingControlDialog.m_nMeshColorMode = 0;

	COLORREF bgColor = 
		RGB( (int)(m_cBackgroundColor.X() * 255.0f), (int)(m_cBackgroundColor.Y() * 255.0f), (int)(m_cBackgroundColor.Z() * 255.0f ) );
	m_LightingControlDialog.m_bgColor = bgColor;

	COLORREF meshColor = 
		RGB( (int)(m_cMeshColor.X() * 255.0f), (int)(m_cMeshColor.Y() * 255.0f), (int)(m_cMeshColor.Z() * 255.0f ) );
	m_LightingControlDialog.m_meshColor = meshColor;

	m_LightingControlDialog.m_fAmbientValue = m_fAmbientLight;

	m_LightingControlDialog.UpdateDialogValues();
	m_LightingControlDialog.ShowWindow(SW_SHOW);
}


void CRenderingManager::OnLightingDialogApplyButton()
{
	m_LightingControlDialog.UpdateClassValues();

	if (m_LightingControlDialog.m_nLightingMode == 0)
		m_eLightingMode = DiffuseOnly;
	else if (m_LightingControlDialog.m_nLightingMode == 1)
		m_eLightingMode = DiffuseAndSpecular;
	else if (m_LightingControlDialog.m_nLightingMode == 2)
		m_eLightingMode = WarmToCoolDiffuse;
	else
		ASSERT(0);

	if (m_LightingControlDialog.m_nShadingMode == 1)
		m_eMeshShadingMode = FlatShading;
	else 
		m_eMeshShadingMode = SmoothShading;

	if (m_LightingControlDialog.m_nMeshColorMode == 1)
		m_eMeshColorMode = SelectionColor;
	else if (m_LightingControlDialog.m_nMeshColorMode == 2)
		m_eMeshColorMode = VertexErrorColor;
	else if (m_LightingControlDialog.m_nMeshColorMode == 3)
		m_eMeshColorMode = ApproxErrorColor;
	else
		m_eMeshColorMode = ConstantColor;
	m_bMeshColorsValid = false;

	m_bDrawCartesianAxes = (m_LightingControlDialog.m_nShowCoordAxes == 1);


	m_cBackgroundColor.X() = ((float)GetRValue(m_LightingControlDialog.m_bgColor)) / 255.0f;
	m_cBackgroundColor.Y() = ((float)GetGValue(m_LightingControlDialog.m_bgColor)) / 255.0f;
	m_cBackgroundColor.Z() = ((float)GetBValue(m_LightingControlDialog.m_bgColor)) / 255.0f;


	m_cMeshColor.X() = ((float)GetRValue(m_LightingControlDialog.m_meshColor)) / 255.0f;
	m_cMeshColor.Y() = ((float)GetGValue(m_LightingControlDialog.m_meshColor)) / 255.0f;
	m_cMeshColor.Z() = ((float)GetBValue(m_LightingControlDialog.m_meshColor)) / 255.0f;

	m_fAmbientLight = m_LightingControlDialog.m_fAmbientValue;

	// update surface manager color parameter
	m_pSurfaceManager->EnableSelectionColoring() = (m_eMeshColorMode == SelectionColor);

	NotifyRenderParametersChanged();
}


/* 
 * cutting plane stuff
 */
void CRenderingManager::OnCuttingPlaneEnabled( const Wml::Frame3f & vPlaneFrame )
{
	m_bCuttingPlaneEnabled = true;
	m_vCuttingPlaneFrame = vPlaneFrame;
	NotifyRenderParametersChanged();
}

void CRenderingManager::OnCuttingPlaneDisabled( )
{
	m_bCuttingPlaneEnabled = false;
}

void CRenderingManager::DrawCuttingPlane()
{
	glPushAttrib(GL_ENABLE_BIT);
	glDisable(GL_LIGHTING);

	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	// get scene bounds
	Wml::AxisAlignedBox3f bounds;
	m_pCurField->FieldBounds(bounds);
	float fDim = (Wml::GetDimension(bounds, 0) + Wml::GetDimension(bounds, 1) + Wml::GetDimension(bounds, 2)) / 3.0f;

	static const float vVertices[4][2] = { {-1.0f,-1.0f}, {1.0f,-1.0f}, {1.0f,1.0f}, {-1.0f,1.0f} };

	glColor4f(0.5f, 0.5f, 0.5f, 0.2f);
	glBegin(GL_QUADS);
	for ( unsigned int i = 0; i < 4; ++i ) {
		Wml::Vector3f vVertex( vVertices[i][0] * fDim, vVertices[i][1] * fDim, 0.0f );
		m_vCuttingPlaneFrame.ToWorld(vVertex);
		vVertex += m_vCuttingPlaneFrame.Origin();
		glVertex3fv(vVertex);
	}
	glEnd();	

	glPopAttrib();
}

void CRenderingManager::RenderMeshAsCuttingPlane( IsoSurfaceMesh * pSurface )
{
	glPushAttrib(GL_ENABLE_BIT);
	glDisable(GL_LIGHTING);

	// get cutting plane normal
	Wml::Vector3f vNormal( m_vCuttingPlaneFrame.Axis( Wml::Frame3f::AxisX ) );
	vNormal *= -1.0f;

	glNormal3fv( vNormal );
	if ( pSurface->GetPolygonizerType() == IsoSurfaceMesh::DynamicMC ) {
		pSurface->GetVFMesh()->DrawGL(false);
	} else {
		TriangleMeshRenderer r( pSurface->GetSimpleMesh() );
		r.DrawAsZFill();
	}

	glPopAttrib();
}

// CRenderingManager message handlers


BEGIN_MESSAGE_MAP(CRenderingManager, CWnd)
	ON_COMMAND(ID_VIEW_SHOWMESH, OnViewShowmesh)
	ON_UPDATE_COMMAND_UI(ID_VIEW_SHOWMESH, OnUpdateViewShowmesh)
	ON_COMMAND(ID_VIEW_SHOWGRADIENTS, OnViewShowgradients)
	ON_UPDATE_COMMAND_UI(ID_VIEW_SHOWGRADIENTS, OnUpdateViewShowgradients)
	ON_COMMAND(ID_VIEW_WIREFRAME, OnViewWireframe)
	ON_UPDATE_COMMAND_UI(ID_VIEW_WIREFRAME, OnUpdateViewWireframe)
	ON_COMMAND(ID_VIEW_SHOWSILHOUETTE, OnViewShowsilhouette)
	ON_UPDATE_COMMAND_UI(ID_VIEW_SHOWSILHOUETTE, OnUpdateViewShowsilhouette)
	ON_COMMAND(ID_VIEW_POINT_SET_CONTROL, OnViewPointSetControl)
	ON_COMMAND(ID_VIEW_LIGHTINGCONTROL, OnViewLightingcontrol)
	ON_COMMAND(ID_VIEW_SHOWCSGEDGES, OnViewShowcsgedges)
	ON_UPDATE_COMMAND_UI(ID_VIEW_SHOWCSGEDGES, OnUpdateViewShowcsgedges)
	ON_COMMAND(ID_VIEW_SURFACEANALYZER, OnViewSurfaceanalyzer)
	ON_COMMAND(ID_VIEW_SHOWSELECTIONSURFACE, OnViewShowselectionsurface)
	ON_UPDATE_COMMAND_UI(ID_VIEW_SHOWSELECTIONSURFACE, OnUpdateViewShowselectionsurface)
	ON_COMMAND(ID_VIEW_NPRMODE, OnViewNprmode)
END_MESSAGE_MAP()



void CRenderingManager::OnSceneChanged()
{
	//InvalidateSurface();
	InvalidatePointSet();
	InvalidateSelectionSurface();
	NotifyRenderParametersChanged();
}

void CRenderingManager::OnTreeSelectionChanged( const std::vector<rmsui::SceneObject *> & newSelection )
{
	m_vSelectedObjects = newSelection;
	InvalidateSelectionSurface();
	//if ( newSelection.size() > 0 ) {
	//	InvalidatePointSet();
	//}
	NotifyRenderParametersChanged();
}

void CRenderingManager::OnMousePickSelectionChanged( const std::vector<rmsui::SceneObject *> & newSelection )
{
	m_vSelectedObjects = newSelection;
	InvalidateSelectionSurface();
	//if ( newSelection.size() > 0 ) {
	//	InvalidatePointSet();
	//}

	NotifyRenderParametersChanged();
}

void CRenderingManager::OnSelectionParametersChanged( const std::vector<rmsui::SceneObject *> & vModified )
{
	//InvalidateSurface();
	InvalidatePointSet();
	InvalidateSelectionSurface();

	NotifyRenderParametersChanged();	
}

void CRenderingManager::TexManRenderParametersChanged()
{
	NotifyRenderParametersChanged();
}

void CRenderingManager::TexManTextureParamsUpdated( ITextureManagerListener::ParamType eType )
{
	m_eCurParamType = eType;
	NotifyRenderParametersChanged();
}

void CRenderingManager::TexManVertexColorsUpdated()
{
	m_bMeshColorsValid = true;
	m_bVertexColorOverride = true;
	NotifyRenderParametersChanged();
}


void CRenderingManager::OnSurfaceManagerMeshesChanged()
{
	// [RMS: this is a hack....]
	//  RMS: for some reason, this causes a huge slowdown. No idea why...

	InvalidatePointSet();
	InvalidateSelectionSurface();;
	m_silFinder.NotifyMeshChanged();

	m_eCurParamType = ITextureManagerListener::NoTextureCoords;

	NotifyRenderParametersChanged();
}

void CRenderingManager::OnSurfaceManagerMeshAttribsUpdated()
{
	NotifyRenderParametersChanged();
}


void CRenderingManager::NotifyRenderParametersChanged()
{
	std::set<IChangeListener *>::iterator cur(m_setListeners.begin()), end(m_setListeners.end());
	while (cur != end)
		(*cur++)->OnRenderParametersChanged();
}

void CRenderingManager::OnViewShowmesh()
{
	m_bDrawMesh = ! m_bDrawMesh;
	NotifyRenderParametersChanged();
}
void CRenderingManager::OnUpdateViewShowmesh(CCmdUI *pCmdUI)
{ pCmdUI->SetCheck( m_bDrawMesh == true ); }


void CRenderingManager::OnViewShowgradients()
{
	m_bDrawMeshNormals = ! m_bDrawMeshNormals;
	NotifyRenderParametersChanged();
}
void CRenderingManager::OnUpdateViewShowgradients(CCmdUI *pCmdUI)
{ pCmdUI->SetCheck( m_bDrawMeshNormals == true ); }


void CRenderingManager::OnViewWireframe()
{
	m_bDrawMeshWireframe = ! m_bDrawMeshWireframe;
	NotifyRenderParametersChanged();
}
void CRenderingManager::OnUpdateViewWireframe(CCmdUI *pCmdUI)
{ pCmdUI->SetCheck( m_bDrawMeshWireframe == true ); }

void CRenderingManager::OnViewShowsilhouette()
{
	m_bDrawSilhouette = ! m_bDrawSilhouette;
	NotifyRenderParametersChanged();
}
void CRenderingManager::OnUpdateViewShowsilhouette(CCmdUI *pCmdUI)
{ pCmdUI->SetCheck( m_bDrawSilhouette == true ); }


void CRenderingManager::OnViewPointSetControl()
{
	ShowPointSetRendererControlDialog();
}

void CRenderingManager::OnViewLightingcontrol()
{
	ShowLightingControlDialog();
}

void CRenderingManager::OnViewShowcsgedges()
{
	m_bDrawSharpEdges = ! m_bDrawSharpEdges;
	NotifyRenderParametersChanged();
}
void CRenderingManager::OnUpdateViewShowcsgedges(CCmdUI *pCmdUI)
{ pCmdUI->SetCheck( m_bDrawSharpEdges == true ); }


void CRenderingManager::OnViewShowselectionsurface()
{
	m_bDrawSelectionSurface = ! m_bDrawSelectionSurface;
	InvalidateSelectionSurface();
	NotifyRenderParametersChanged();
}
void CRenderingManager::OnUpdateViewShowselectionsurface(CCmdUI *pCmdUI)
{ pCmdUI->SetCheck( m_bDrawSelectionSurface == true ); }



void CRenderingManager::OnViewSurfaceanalyzer()
{
	m_SurfaceAnalyzer.ShowControlDialog(true);
}

BOOL CRenderingManager::OnCmdMsg(UINT nID, int nCode, void* pExtra, AFX_CMDHANDLERINFO* pHandlerInfo)
{
	// give sub-managers a chance...
	if ( m_texManager.OnCmdMsg(nID, nCode, pExtra, pHandlerInfo) )
		return TRUE;

	return __super::OnCmdMsg(nID, nCode, pExtra, pHandlerInfo);
}


void CRenderingManager::OnViewNprmode()
{
	m_bDrawCartesianAxes = false;
	m_bDrawMesh = false;
	m_bDrawLightCulledPoints = true;
	m_bDrawSilhouette = true;
	m_bDrawSuggestiveContours = true;
	m_bDrawDiscZFill = true;
	m_bDrawZFillMesh = false;

	// this hides bboxes and scene object dots/lines
	CSketchManager::Get()->SetSketchingEnabled(true);

	NotifyRenderParametersChanged();
}

void CRenderingManager::ExternalHackFunc()
{
	m_fStippleSize = 1.0f;
	m_fSilhouetteWidth = 1.0f;
	m_PointSetRenderer.SetDiscPointSubdivSteps(2);
}