// Ryan Schmidt - rms@unknownroad.com
// Copyright (c) 2004-2010
// Distributed under the Boost Software License, Version 1.0.
// http://www.boost.org/LICENSE_1_0.txt


#pragma once


#include <OpenGLWmlLight.h>

#include <ExtendedWmlCamera.h>

#include "ChangeListener.h"
#include "TriangleMeshRenderer.h"

#include "LightingControlDialog.h"


#include <vector>
#include <set>



// CRenderingManager

class CRenderingManager : public CWnd, public IChangeListener, 
	public CLightingControlDialogListener
{
	DECLARE_DYNAMIC(CRenderingManager)

public:
	CRenderingManager();
	virtual ~CRenderingManager();

	void RegisterChangeListener( IChangeListener * pListener );

	virtual void OnSceneChanged();
	virtual void OnTreeSelectionChanged( const std::vector<rmsui::SceneObject *> & newSelection );
	virtual void OnMousePickSelectionChanged( const std::vector<rmsui::SceneObject *> & newSelection );
	virtual void OnSelectionParametersChanged( const std::vector<rmsui::SceneObject *> & vModified );
	virtual void OnCuttingPlaneEnabled( const Wml::Frame3f & vPlaneFrame );
	virtual void OnCuttingPlaneDisabled( );


	//! surface manager listener interface
	virtual void OnSurfaceManagerMeshesChanged();
	virtual void OnSurfaceManagerMeshAttribsUpdated();

	void SetCamera( Wml::ExtendedWmlCamera * pCamera );

	void SetCurrentBlobTree( rmsimplicit::ScalarField * pField );

	void SetSurfaceManager( CSurfaceManager * pManager ) {m_pSurfaceManager = pManager;}
	CSurfaceManager * GetSurfaceManager() { return m_pSurfaceManager; }

	// notify that OpenGL is created
	void OnCreateGL();

	void Render();

	void ShowPointSetRendererControlDialog( );
	void ShowLightingControlDialog();

	// statistics stuff
	enum StatisticType {
		FieldValueErrorAtMeshVertices, CacheMemoryUsage, CacheEvaluations
	};
	void ComputeStatistics( StatisticType eType, const char * pFilename );

	// texture generation
	CTextureManager * GetTextureManager();


	void ExternalHackFunc();

protected:

	
/*
 * camera
 */ 
	Wml::ExtendedWmlCamera * m_pCamera;
	Wml::Vector3f m_cBackgroundColor;


/*
 * misc gfx stuff
 */
	bool m_bDrawCartesianAxes;
	void DrawCartesianAxes( float fLength );

/*
 * lighting
 */
	enum LightingMode {
		NoLighting, DiffuseOnly, DiffuseAndSpecular, WarmToCoolDiffuse, NPRLighting
	};
	LightingMode m_eLightingMode;

	Wml::OpenGLWmlLight m_Lights[8];
	float m_fAmbientLight;

	void InitializeLighting();
	void SetDiffuseOnlyLighting();
	void SetDiffuseAndSpecularLighting();
	void SetWarmToCoolDiffuseLighting();
	void SetNPRLighting();
//	void InitializeCubeMap();

/*
 * scene objects  (TODO: this maybe should me more elaborate??)
 */
	rmsimplicit::ScalarField * m_pCurField; 


/*
 * polygonization
 */
	CSurfaceManager * m_pSurfaceManager;

	// mesh drawing control flags
	bool m_bDrawMesh;
	bool m_bDrawMeshNormals;
	bool m_bDrawMeshWireframe;

	enum MeshColorMode {
		ConstantColor, VertexErrorColor, ApproxErrorColor, SelectionColor
	};
	MeshColorMode m_eMeshColorMode;
	bool m_bMeshColorsValid;

	enum MeshShadingMode {
		SmoothShading, FlatShading
	};
	MeshShadingMode m_eMeshShadingMode;

	Wml::Vector3f m_cMeshColor;
	Wml::Vector3f m_cWireframeColor;
	Wml::Vector3f m_cNormalColor;
	float m_fNormalLength;
		
	void ComputeMeshColors();
	void RenderSurface( IsoSurfaceMesh * pSurface, bool bDoPolyOffset, float fAlpha = 1.0f );

/*
 * selection mesh support
 */
	bool m_bDrawSelectionSurface;
	rmsimplicit::CPortGG4Polygonizer3D m_SelectionSurfacePolygonizer;
	rmsimplicit::TriangleMesh m_curSelectionMesh;
	bool m_bCurSelectionMeshValid;

	std::vector<rmsui::SceneObject *> m_vSelectedObjects;

	// selection mesh
	void InvalidateSelectionSurface();
	void ValidateSelectionSurface();


	void RenderSelectionSurface();

/*
 * cutting plane stuff
 */
	bool m_bCuttingPlaneEnabled;
	Wml::Frame3f m_vCuttingPlaneFrame;

	void DrawCuttingPlane();
	void RenderMeshAsCuttingPlane( IsoSurfaceMesh * pSurface );

/*
 * NPR rendering stuff
 */
	PointSetRenderer m_PointSetRenderer;
	CPointSetRendererControlDialog m_PointSetRendererDialog;
	CLightingControlDialog m_LightingControlDialog;

	// NPR-style rendering control flags
	bool m_bPointSetValid;

	bool m_bDrawAllPoints;
	bool m_bDrawLightCulledPoints;

	bool m_bDrawSharpEdges;

	bool m_bDrawZFillMesh;
	bool m_bDrawDiscZFill;

	bool m_bDrawSilhouette;
	bool m_bDrawSuggestiveContours;

	Wml::Vector3f m_cSilhouetteColor;
	float m_fSilhouetteWidth;
	float m_fStippleSize;

	void InvalidatePointSet();
	void ValidatePointSet();
	void RenderMeshesAsZFill();
	void RenderSharpEdges();
	void RenderDiscsAsZFill();

	void RenderAllPoints();
	void RenderLightCulledPoints();

	rmsimplicit::BruteForceSilhouetteFinder m_silFinder;
	bool m_bFreezeSilhouette;
	void RenderContours();

/*
 * surface analysis stuff
 */
	SurfaceAnalyzer m_SurfaceAnalyzer;

/*
 * texture stuff
 */
	CTextureManager m_texManager;
	ITextureManagerListener::ParamType m_eCurParamType;
	bool m_bVertexColorOverride;


/*
 * change listeners
 */
	std::set<IChangeListener *> m_setListeners;
	void NotifyRenderParametersChanged();

	virtual void OnPointSetRendererDialogApplyButton();
	virtual void OnLightingDialogApplyButton();

protected:
	DECLARE_MESSAGE_MAP()
public:
	afx_msg void OnViewShowmesh();
	afx_msg void OnUpdateViewShowmesh(CCmdUI *pCmdUI);
	afx_msg void OnViewShowgradients();
	afx_msg void OnUpdateViewShowgradients(CCmdUI *pCmdUI);
	afx_msg void OnViewWireframe();
	afx_msg void OnUpdateViewWireframe(CCmdUI *pCmdUI);
	afx_msg void OnViewShowsilhouette();
	afx_msg void OnUpdateViewShowsilhouette(CCmdUI *pCmdUI);
	afx_msg void OnViewPointSetControl();
	afx_msg void OnViewLightingcontrol();
	afx_msg void OnViewShowcsgedges();
	afx_msg void OnUpdateViewShowcsgedges(CCmdUI *pCmdUI);
	afx_msg void OnViewSurfaceanalyzer();
	virtual BOOL OnCmdMsg(UINT nID, int nCode, void* pExtra, AFX_CMDHANDLERINFO* pHandlerInfo);
	afx_msg void OnViewShowselectionsurface();
	afx_msg void OnUpdateViewShowselectionsurface(CCmdUI *pCmdUI);
	afx_msg void OnViewNprmode();
};


inline void CRenderingManager::SetCamera( Wml::ExtendedWmlCamera * pCamera )
{
	m_pCamera = pCamera;
	ASSERT(m_pCamera);
	m_texManager.SetCamera(pCamera);
}

inline void CRenderingManager::SetCurrentBlobTree( rmsimplicit::ScalarField * pField )
{
	m_pCurField = pField;
	ASSERT(m_pCurField);

	m_texManager.SetCurrentBlobTree(pField);
}

inline void CRenderingManager::RegisterChangeListener( IChangeListener * pListener )
{
	m_setListeners.insert(pListener);
}


inline CTextureManager * CRenderingManager::GetTextureManager()
{
	return &m_texManager;
}