// Ryan Schmidt - rms@unknownroad.com
// Copyright (c) 2005-2010
// Distributed under the Boost Software License, Version 1.0.
// http://www.boost.org/LICENSE_1_0.txt


// SceneManager.cpp : implementation file
//

#include "stdafx.h"
#include "SketchFramework.h"
#include "SceneManager.h"

#include "ParametersDialogBar.h"
#include "TreeEditorDialogBar.h"
#include "ParameterEditor.h"

#include "GenericPointSceneObject.h"

#include "BlobTreeLocking.h"

#include <PolygonField2D.h>

// ack bad
#include <WmlBSplineCurve2.h>
#include <PolyLine.h>

#include <WmlExtVectorUtil.h>
#include <WmlPGMImage.h>

using namespace rmsui_bt;

static rmsimplicit::ScalarField2D * LoadField();

#define LOCK_BLOBTREE() SCOPELOCK_BLOBTREE( m_pCurrentBlobTree )


// CSceneManager

IMPLEMENT_DYNAMIC(CSceneManager, CWnd)
CSceneManager::CSceneManager()
{
	m_pObjectManager = new rmsui::SceneObjectManager();
	m_pTool3DManager = NULL;
	m_pCurrentBlobTree = NULL;
	m_pCurrentBlobTree_ThreadSafeField = NULL;

	InitializeFirstBlobTree();
}

CSceneManager::~CSceneManager()
{
	delete m_pObjectManager;
	m_pObjectManager = NULL;
}


std::map< CurrentSceneManagerKey, CSceneManager * > CSceneManager::s_streams;
CurrentSceneManagerKey CSceneManager::s_CurrentKey = 0;


void CSceneManager::SetCurrentSceneManager( CurrentSceneManagerKey key )
{
	s_CurrentKey = key;	
}



CSceneManager * CSceneManager::Get()
{
	ASSERT(s_CurrentKey != 0);

	if ( s_streams[s_CurrentKey] == NULL )
		s_streams[s_CurrentKey] = new CSceneManager();
	return s_streams[s_CurrentKey];
}





void CSceneManager::SetTool3DManager( rmsui::Tool3DManager * pTool3DManager )
{
	m_pTool3DManager = pTool3DManager;

	// assume tree is already created..
	m_pTool3DManager->SetToolSurface( m_pCurrentBlobTree->GetRootNode()->GetRootNode() );
}

BEGIN_MESSAGE_MAP(CSceneManager, CCmdTarget)
	ON_COMMAND(ID_ADDPRIM_POINT, OnAddprimPoint)

END_MESSAGE_MAP()



void CSceneManager::RegisterChangeListener( IChangeListener * pListener )
{
	m_setListeners.insert(pListener);
}

void CSceneManager::NotifySceneChanged()
{
	std::set<IChangeListener *>::iterator cur(m_setListeners.begin()), end(m_setListeners.end());
	while (cur != end)
		(*cur++)->OnSceneChanged();
}

void CSceneManager::NotifyParametersChangedPassThrough()
{
	NotifySelectionParametersChanged();
}

void CSceneManager::NotifyTreeSelectionChanged()
{
	std::set<IChangeListener *>::iterator cur(m_setListeners.begin()), end(m_setListeners.end());
	while (cur != end)
		(*cur++)->OnTreeSelectionChanged( m_pObjectManager->GetSelectedEntities() );
}

void CSceneManager::NotifySelectionParametersChanged()
{
	std::set<IChangeListener *>::iterator cur(m_setListeners.begin()), end(m_setListeners.end());
	while (cur != end)
		(*cur++)->OnSelectionParametersChanged( m_curSelection );
}

void CSceneManager::NotifyCurrentBlobTreeChanged()
{
	std::set<IChangeListener *>::iterator cur(m_setListeners.begin()), end(m_setListeners.end());
	while (cur != end)
		(*cur++)->OnCurrentBlobTreeChanged( );
}

void CSceneManager::NotifyMousePickSelectionChanged()
{
	std::set<IChangeListener *>::iterator cur(m_setListeners.begin()), end(m_setListeners.end());
	while (cur != end)
		(*cur++)->OnMousePickSelectionChanged( m_pObjectManager->GetSelectedEntities() );
}

void CSceneManager::NotifyCuttingPlaneEnabled( const Wml::Frame3f & vPlaneFrame )
{
	std::set<IChangeListener *>::iterator cur(m_setListeners.begin()), end(m_setListeners.end());
	while (cur != end)
		(*cur++)->OnCuttingPlaneEnabled( vPlaneFrame );
}

void CSceneManager::NotifyCuttingPlaneDisabled()
{
	std::set<IChangeListener *>::iterator cur(m_setListeners.begin()), end(m_setListeners.end());
	while (cur != end)
		(*cur++)->OnCuttingPlaneDisabled( );
}

void CSceneManager::NotifyRenderParametersChanged()
{
	std::set<IChangeListener *>::iterator cur(m_setListeners.begin()), end(m_setListeners.end());
	while (cur != end)
		(*cur++)->OnRenderParametersChanged( );
}


/*
 * these are notifications SceneManager gets from 3DView (at this point...)
 */


void CSceneManager::NotifyToolActionOcurred()
{
	NotifySelectionParametersChanged();
}
void CSceneManager::NotifyMousePickActionOccurred()
{
	m_curSelection = m_pObjectManager->GetSelectedEntities();
	NotifyMousePickSelectionChanged();
}

/*
 * rmsui::IToolActionEventListener handlers
 */
void CSceneManager::OnToolMoveEvent()
{
	NotifySelectionParametersChanged();
}


/* 
 * IGestureEventListener events
 */
bool CSceneManager::OnPointStroke( rmsui::GesturePointStroke * pPointStroke )
{
	LOCK_BLOBTREE();

	const unsigned int * pViewport = m_pTool3DManager->GetCamera()->GetViewport();
	Wml::Ray3f ray;
	m_pTool3DManager->GetCamera()->GetPickRay(
		(int)pPointStroke->Point().X(), 
		pViewport[3] - 1 - (int)pPointStroke->Point().Y(), 
		pViewport[2], pViewport[3],
		ray.Origin(), ray.Direction() );	

	if ( TryPickSurface( ray ) == true ) {

		// update tool manager
		GetSceneObjectManager()->UpdateToolManager( m_pTool3DManager );
		NotifyMousePickActionOccurred();
		m_pTool3DManager->EnableCurrentTool(true);

		return true;
	} else {

		m_pObjectManager->DeSelectAll();
		GetSceneObjectManager()->UpdateToolManager( m_pTool3DManager );
		NotifyMousePickActionOccurred();
		m_pTool3DManager->EnableCurrentTool(false);

		return true;
	}
}

bool CSceneManager::OnLineStroke( rmsui::GestureLineStroke * pLineStroke )
{
	LOCK_BLOBTREE();

	Wml::Vector2f vStroke( pLineStroke->EndPoint2() - pLineStroke->EndPoint1() );
	vStroke.Normalize();
	if ( vStroke.Dot( Wml::Vector2f::UNIT_Y ) > 0.9f ) {
		return MoveSelectionToParent();
	} else
		return false;
}

bool CSceneManager::OnScribbleStroke( rmsui::GestureScribbleStroke * pScribbleStroke )
{
	LOCK_BLOBTREE();

	// do NOT do if sketch manager is sketching...
	// NOTE: figure out a way to test this (need access to sketch manager ??)

	// update current selection set
	m_curSelection = m_pObjectManager->GetSelectedEntities();

	if ( m_curSelection.size() == 0 )
		return false;
	if ( m_curSelection[0] == m_pCurrentBlobTree->GetRootNode() )
		return false;

	// make sure selection is a BT node
	rmsui_bt::BlobTreeNodeSceneObject * pObject = 
		dynamic_cast<rmsui_bt::BlobTreeNodeSceneObject *>(m_curSelection[0]);
	if (pObject == NULL)
		return false;

	// ok, now make sure that ray through center of scribble box intersects with this object
	rmsbt::BlobTreeNode * pNode = pObject->GetBlobTreeNode();

	const unsigned int * pViewport = m_pTool3DManager->GetCamera()->GetViewport();
	Wml::Ray3f ray;
	m_pTool3DManager->GetCamera()->GetPickRay(
		(int)pScribbleStroke->Bounds().Center().X(),
		pViewport[3] - 1 - (int)pScribbleStroke->Bounds().Center().Y(), 
		pViewport[2], pViewport[3],
		ray.Origin(), ray.Direction() );	

	Wml::Vector3f vIntersect;
	if ( pNode->FindIntersection( ray.Origin(), ray.Direction(), vIntersect ) == false )
		return false;

	bool bRemoved = RemoveNode(pObject);

	return bRemoved;
}


/*
 * Setup / Loading / Saving
 */


void CSceneManager::InitializeFirstBlobTree()
{
	m_pCurrentBlobTree = new BlobTreeSceneObject( );
	rmsui_bt::BlobTreeFactory * pFactory = new BlobTreeFactory( m_pCurrentBlobTree );
	m_pCurrentBlobTree->SetFactory( pFactory );

	m_pObjectManager->AddObject(m_pCurrentBlobTree, false);
	m_pObjectManager->AddObject( m_pCurrentBlobTree->GetRootNode(), true);

	// notify bt dialog bar
	CBlobTreeEditorDialogBar::Get()->SetCurrentBlobTree( m_pCurrentBlobTree );

	m_pCurrentBlobTree_ThreadSafeField = 
		new rmsui_bt::ThreadSafeScalarField( m_pCurrentBlobTree, m_pCurrentBlobTree->GetRootNode()->GetRootNode() );

	NotifySceneChanged();
}


void CSceneManager::LoadBlobTree( rmsbt::RootNode * pRootNode )
{
	BlobTreeNodeSceneObject * pRootNodeObj = m_pCurrentBlobTree->GetFactory()->RestoreTree( pRootNode );

	NotifySceneChanged();

}


void CSceneManager::ForceToolManagerUpdate()
{
	m_pObjectManager->UpdateToolManager( m_pTool3DManager );
}

void CSceneManager::ForceReDraw()
{
	NotifyRenderParametersChanged();
}


void CSceneManager::SetBlobTreeEditorTreeSelectionSet( 
					 const std::vector<rmsui_bt::BlobTreeNodeSceneObject *> & selection )
{
	size_t nSize = selection.size();

	// clear current selection in object manager
	m_pObjectManager->DeSelectAll();

	// update current selection list (TODO: could we retreive this from object manger ??)
	m_curSelection.clear();
	for (unsigned int i = 0; i < nSize; ++i)
		m_curSelection.push_back( selection[i] );

	if (nSize) {

		// select new selection in object manager
		for (unsigned int i = 0; i < nSize; ++i) {
			m_pObjectManager->SelectObject( selection[i] );
		}

		// At this point we can assume selection is a blobtree node here, so set current blobtree
		ASSERT( m_pCurrentBlobTree == ((rmsui_bt::BlobTreeNodeSceneObject *)selection[0])->GetBlobTree() );
		//m_pCurrentBlobTree = ((rmsui_bt::BlobTreeNodeSceneObject *)selection[0])->GetBlobTree();

	} 

	// update tool manager. Could we do this with a notification?
	m_pObjectManager->UpdateToolManager( m_pTool3DManager );
	m_pTool3DManager->EnableCurrentTool( (nSize > 0) ? true : false );

//	NotifySceneChanged();			// no need for this. forces repolygonization...! (which is crap...)
	NotifyTreeSelectionChanged();
}

void CSceneManager::ClearBlobTreeEditorTreeSelectionSet()
{
	m_curSelection.clear();
	NotifyTreeSelectionChanged();
}



/*
 * picking support
 */
bool CSceneManager::TryPickSurface( const Wml::Ray3f & ray )
{
	LOCK_BLOBTREE();

	// RMS TODO: support more than one tree...

	rmsbt::RootNode * pRootNode = m_pCurrentBlobTree->GetRootNode()->GetRootNode();

	// try pick
	bool bFound = false;
	Wml::Vector3f vIsect;
	if ( pRootNode->FindIntersection( ray.Origin(), ray.Direction(), vIsect ) ) {

		rmsbt::BlobTreeNode * pSelectedNode = NULL;
		float fSelectedValue = 0.0f;
		pRootNode->SelectClosestPrimitive( vIsect, &pSelectedNode, fSelectedValue );

		if ( pSelectedNode != NULL ) {
			std::vector<SceneObject *> vObjects;
			unsigned int nSize = GetSceneObjectManager()->GetSceneObjectCount();
			vObjects.resize( nSize );
			GetSceneObjectManager()->GetSceneObjects(vObjects);

			// find object for node
			for ( unsigned int i = 0; i < nSize && !bFound; ++i ) {
				rmsui_bt::BlobTreeNodeSceneObject * pObject = 
					dynamic_cast<rmsui_bt::BlobTreeNodeSceneObject *>( vObjects[i] );
				if ( pObject != NULL && pObject->GetBlobTreeNode() == pSelectedNode ) {

					GetSceneObjectManager()->DeSelectAll();
					GetSceneObjectManager()->SelectObject( vObjects[i] );
					bFound = true;
				}
			}
		}

	}

	return bFound;
}


bool CSceneManager::MoveSelectionToParent()
{
	LOCK_BLOBTREE();

	// get selection list
	m_curSelection = m_pObjectManager->GetSelectedEntities();
	if (m_curSelection.size() != 1)
		return false;	

	// convert to node
	rmsui_bt::BlobTreeNodeSceneObject * pSelected = dynamic_cast<rmsui_bt::BlobTreeNodeSceneObject *>( m_curSelection[0] );
	if (!pSelected)
		return false;

	// get parent
	rmsui::SceneObject * pParent = pSelected->GetParent();
	rmsui_bt::BlobTreeNodeSceneObject * pParentNode = dynamic_cast<rmsui_bt::BlobTreeNodeSceneObject *>( pParent );
	if ( pParentNode == NULL )
		return false;

	// if parent is root node, do not move
	if ( pParentNode->GetBlobTreeNode()->GetType() == rmsbt::BlobTreeNode::RootNode )
		return false;

	m_pObjectManager->DeSelectAll();
	m_pObjectManager->SelectObject( pParent );
	m_pObjectManager->UpdateToolManager( m_pTool3DManager );
	m_pTool3DManager->EnableCurrentTool( true );

	NotifySceneChanged();
	NotifyCurrentBlobTreeChanged();
	NotifyTreeSelectionChanged();

	return true;
}



// CSceneManager message handlers

void CSceneManager::OnRemoveSelection()
{
	LOCK_BLOBTREE();

	// for now just "lose" it....crap crap crap!
	m_curSelection = m_pObjectManager->GetSelectedEntities();
	if (m_curSelection.size() == 0)
		return;

	// just erase first one for now  [RMS TODO: implement this properly...]
	rmsui_bt::BlobTreeNodeSceneObject * pRemove = dynamic_cast<rmsui_bt::BlobTreeNodeSceneObject *>( m_curSelection[0] );
	if (!pRemove)
		ASSERT(0);

	if (pRemove->GetParent() == NULL) {
		_RMSInfo("Ignoring attempt to remove root node...\n");
		return;
	}

	// remove from parent
	rmsui_bt::BlobTreeNodeSceneObject * pParent = NULL;
	bool bResult = TryRemoveNodeFromParent( pRemove, &pParent );
	if (!bResult) {
		ASSERT(bResult);
		return;
	}

	m_pObjectManager->DeSelectAll();
	m_pObjectManager->RemoveObjectAndChildren( pRemove );
	delete pRemove;

	m_pTool3DManager->ClearToolTargets();

	m_curSelection = m_pObjectManager->GetSelectedEntities();

	// optimize tree
	if ( pParent )
		OptimizeSubTree( pParent, SMOPT_REDUNDANT_COMPOSITIONS | SMOPT_CHILDLESS_COMPOSITIONS );

	NotifySceneChanged();
	NotifyCurrentBlobTreeChanged();
	NotifyTreeSelectionChanged();

}



rmsui_bt::BlobTreeNodeSceneObject * CSceneManager::AddSimpleChildNode( rmsui_bt::BlobTreeNodeSceneObject::Type eType,
																	   rmsui_bt::BlobTreeNodeSceneObject * pParent )
{
	LOCK_BLOBTREE();

	if (m_pCurrentBlobTree) {

		if ( pParent == NULL && !m_curSelection.empty() )
			pParent = dynamic_cast< rmsui_bt::CompositionNodeSceneObject *>( m_curSelection[0] );

		rmsui_bt::BlobTreeNodeSceneObject * pObject = 
			m_pCurrentBlobTree->GetFactory()->AddNodeChild( pParent, eType );

		m_pObjectManager->DeSelectAll();
		m_pObjectManager->AddObject( pObject, true );

		// notification??
		m_pObjectManager->UpdateToolManager( m_pTool3DManager );
		m_pTool3DManager->EnableCurrentTool(true);

		NotifySceneChanged();
		NotifyCurrentBlobTreeChanged();
		NotifyTreeSelectionChanged();

		return pObject;
	} else {
		ASSERT(0);
		return NULL;
	}
}



void CSceneManager::AddEricModel( int nType )
{
	LOCK_BLOBTREE();

	rmsui_bt::BlobTreeNodeSceneObject::Type eType = (rmsui_bt::BlobTreeNodeSceneObject::Type )nType;

	if (m_pCurrentBlobTree) {

		rmsui_bt::BlobTreeNodeSceneObject * pParent = NULL;

		// create new revolution object
		SumBlendSceneObject * pSumBlendObject = 
			(SumBlendSceneObject *)m_pCurrentBlobTree->GetFactory()->AddNodeChild(pParent, eType);

		m_pObjectManager->DeSelectAll();
		m_pObjectManager->AddObject( pSumBlendObject, true );

		// notification??
		m_pObjectManager->UpdateToolManager( m_pTool3DManager );
		m_pTool3DManager->EnableCurrentTool(true);

		NotifySceneChanged();
		NotifyCurrentBlobTreeChanged();
		NotifyTreeSelectionChanged();

	} else
		ASSERT(0);
}


void CSceneManager::OnAddprimPoint() 
{ 
	rmsui::GenericPointSceneObject * pObject =
		new rmsui::GenericPointSceneObject();

	
	m_pObjectManager->DeSelectAll();
	m_pObjectManager->AddObject( pObject, true );

	// notification??
	m_pObjectManager->UpdateToolManager( m_pTool3DManager );
	m_pTool3DManager->EnableCurrentTool(true);

	NotifySceneChanged();
	NotifyCurrentBlobTreeChanged();
	NotifyTreeSelectionChanged();
}





void CSceneManager::SetTemporaryCuttingPlane( const Wml::Frame3f & planeFrame )
{
	LOCK_BLOBTREE();

	if (! m_pCurrentBlobTree )
		return;

	m_pCurrentBlobTree->GetRootNode()->GetRootNode()->SetCuttingPlane( planeFrame );
//	m_pCurrentBlobTree->GetRootNode()->GetRootNode()->EnableCuttingPlane( true );

	NotifyCuttingPlaneEnabled( planeFrame );
	NotifySelectionParametersChanged();
}

bool CSceneManager::IsTemporaryCuttingPlaneEnabled()
{
	LOCK_BLOBTREE();

	if (! m_pCurrentBlobTree )
		return false;
	Wml::Frame3f vTmp;
	return m_pCurrentBlobTree->GetRootNode()->GetRootNode()->GetCuttingPlane(vTmp);
}

void CSceneManager::ClearTemporaryCuttingPlane()
{
	LOCK_BLOBTREE();

	if ( m_pCurrentBlobTree ) {
		m_pCurrentBlobTree->GetRootNode()->GetRootNode()->EnableCuttingPlane(false);
		NotifyCuttingPlaneDisabled();
		NotifySelectionParametersChanged();
	}
}




void CSceneManager::OnEditClearselection()
{
	m_pObjectManager->DeSelectAll();
	m_pObjectManager->UpdateToolManager( m_pTool3DManager );
	m_pTool3DManager->EnableCurrentTool(false);

	NotifyTreeSelectionChanged();
}

