// Ryan Schmidt - rms@unknownroad.com
// Copyright (c) 2005-2010
// Distributed under the Boost Software License, Version 1.0.
// http://www.boost.org/LICENSE_1_0.txt


#include "stdafx.h"
#include "SketchFramework.h"
#include "SketchManager.h"
#include ".\sketchmanager.h"

#include "MainFrm.h"
#include "ChildFrm.h"
#include "SystemConstants.h"

#include <gl/GL.h>
#include <gl/GLU.h>

#include <WmlExtVectorUtil.h>
#include <WmlExtIntrPln3Lin3.h>
#include <WmlSegment2.h>
#include <WmlDistVec2Lin2.h>
#include <WmlDistLin3Lin3.h>
#include <algorithm>

// sketch object scene object types
#include "CappedLinearSweepSceneObject.h"
#include "RevolutionSceneObject.h"
#include "MinMaxDifferenceSceneObject.h"
#include "RicciBlendSceneObject.h"
#include "CurveSweepSceneObject.h"
#include "PolyLineSceneObject.h"
#include "SumBlendSceneObject.h"

#include "BlobTreeLocking.h"


#define SKETCH_FIELD_RESOLUTION 128


rmsui::SceneObject * StrokeData::GetFirstHitObject() 
{
	size_t nHits = StrokeSize();
	for (unsigned int i = 0; i < nHits; ++i)
		if (m_vStrokeRays[i].pHitObject != NULL)
			return m_vStrokeRays[i].pHitObject;
	return NULL;
}

rmsui::SceneObject * StrokeData::GetLastHitObject() 
{
	for (unsigned int i = StrokeSize()-1; i >= 0; ++i)
		if (m_vStrokeRays[i].pHitObject != NULL)
			return m_vStrokeRays[i].pHitObject;
	return NULL;
}

bool StrokeData::IsCutOutStroke()
{
	if ( m_vStrokeRays.size() < 5 )
		return false;

	// this is a "crossing" stroke if we started outside, then went inside, then back outside...
	// (or if it was always inside...)
	bool bAlwaysInside = true;
	size_t nSize = m_vStrokeRays.size();
	std::vector<bool> bInOutVec;		// true is in, false is out
	int nState = 2;		// invalid initial state;
	for (unsigned int i = 0; i < nSize; ++i) {
		StrokeData::RayVertexInfo & rayInfo = m_vStrokeRays[i];
		if (nState == 2) {
			bInOutVec.push_back( (rayInfo.pHitObject != NULL) );
			nState = (rayInfo.pHitObject != NULL);
		} else if ( nState == 1 && rayInfo.pHitObject == NULL ) {
			bInOutVec.push_back(false);
			nState = 0;
		} else if (nState == 0 && rayInfo.pHitObject != NULL ) {
			bInOutVec.push_back(true);
			nState = 1;
		}
		if (nState == 0)
			bAlwaysInside = false;
	}

	// have to start out, then in, then out for this to be a crossing stroke
	if ( bAlwaysInside || bInOutVec.size() >= 3 && bInOutVec[0] == false && bInOutVec[1] == true && bInOutVec[2] == false )
		return true;
	else
		return false;
}


bool StrokeData::IsAddOnStroke()
{
	if ( m_vStrokeRays.size() < 5 )
		return false;

	// this is a "crossing" stroke if we started outside, then went inside, then back outside...
	bool bAlwaysInside = true;
	size_t nSize = m_vStrokeRays.size();
	std::vector<bool> bInOutVec;		// true is in, false is out
	int nState = 2;		// invalid initial state;
	for (unsigned int i = 0; i < nSize; ++i) {
		StrokeData::RayVertexInfo & rayInfo = m_vStrokeRays[i];
		if (nState == 2) {
			bInOutVec.push_back( (rayInfo.pHitObject != NULL) );
			nState = (rayInfo.pHitObject != NULL);
		} else if ( nState == 1 && rayInfo.pHitObject == NULL ) {
			bInOutVec.push_back(false);
			nState = 0;
		} else if (nState == 0 && rayInfo.pHitObject != NULL ) {
			bInOutVec.push_back(true);
			nState = 1;
		}
		if (nState == 0)
			bAlwaysInside = false;
	}

	// have to start out, then in, then out for this to be a crossing stroke
	if ( bAlwaysInside || bInOutVec.size() >= 3 && bInOutVec[0] == true && bInOutVec[1] == false && bInOutVec[2] == true )
		return true;
	else
		return false;
}


bool StrokeData::IsJoinStroke()
{
	rmsui::SceneObject * pFirst = NULL, * pLast = NULL;

	// find first selection object
	size_t nHits = m_vStrokeRays.size();
	for (unsigned int i = 0; i < nHits; ++i) {
		if (m_vStrokeRays[i].pHitObject != NULL) {
			pFirst = m_vStrokeRays[i].pHitObject;
			i = (unsigned int)nHits + 1;
		}
	}

	// find last selection obejct
	for (int j = (unsigned int)nHits-1; j >= 0; --j) {
		if (m_vStrokeRays[j].pHitObject != NULL) {
			pLast = m_vStrokeRays[j].pHitObject;
			j = -1;
		}
	}

	if ( pFirst != NULL && pLast != NULL && pFirst != pLast )
		return true;
	else
		return false;
}

bool StrokeData::IsCuttingPlaneStroke()
{
	size_t nSize = m_vStrokeRays.size();
	if (nSize < 3)
		return false;

	// this is a "crossing" stroke if we started outside, then went inside, then back outside...
	bool bAlwaysInside = true;
	std::vector<bool> bInOutVec;		// true is in, false is out
	int nState = 2;		// invalid initial state;
	for (unsigned int i = 0; i < nSize; ++i) {
		StrokeData::RayVertexInfo & rayInfo = m_vStrokeRays[i];
		if (nState == 2) {
			bInOutVec.push_back( (rayInfo.pHitObject != NULL) );
			nState = (rayInfo.pHitObject != NULL);
		} else if ( nState == 1 && rayInfo.pHitObject == NULL ) {
			bInOutVec.push_back(false);
			nState = 0;
		} else if (nState == 0 && rayInfo.pHitObject != NULL ) {
			bInOutVec.push_back(true);
			nState = 1;
		}
		if (nState == 0)
			bAlwaysInside = false;
	}

	// quit if we don't have the right crossing stuff
	if ( ! (bInOutVec.size() >= 3 && bInOutVec[0] == false && bInOutVec[1] == true && bInOutVec[2] == false) )
		return false;

	// ok, cutting stroke is L-shaped. split stroke into 'linear' segments (approximately...)
	std::vector< Wml::Vector2f > vDirChangePts;

	static const float fThreshold = 0.9f;
	vDirChangePts.push_back( m_vStrokeRays[0].vScreen );		// always add first point
	Wml::Vector2f vCurDir = m_vStrokeRays[1].vScreen - vDirChangePts.back();
	vCurDir.Normalize();
	for ( unsigned int i = 1; i < nSize-1 && vDirChangePts.size() < 5; ++i ) {
		Wml::Vector2f vNextDir( m_vStrokeRays[i+1].vScreen - m_vStrokeRays[i].vScreen );
		vNextDir.Normalize();
		if ( vNextDir.Dot( vCurDir ) < fThreshold ) {
			vDirChangePts.push_back( m_vStrokeRays[i].vScreen );
			vCurDir = vNextDir;
		}
	}
	vDirChangePts.push_back( m_vStrokeRays.back().vScreen );		// always add last point

	nSize = vDirChangePts.size();
	if ( nSize >= 5 || nSize < 3 )		// too many changes to be an L...
		return false;

	// ok, look at first and last vectors...
	Wml::Vector2f vFirst( vDirChangePts[0] - vDirChangePts[1] );
	vFirst.Normalize();
	Wml::Vector2f vLast( vDirChangePts.back() - vDirChangePts[ nSize - 2 ] );
	vLast.Normalize();

	// dot product should be near 0 if it's an L...
	float fDot = vFirst.Dot(vLast);
	if ( fabsf(fDot) < 0.3f )
		return true;
	else
		return false;
}

bool StrokeData::IsSurfaceStroke()
{
	if ( m_vStrokeRays.size() < 1 )
		return false;

	size_t nSize = m_vStrokeRays.size();
	for (unsigned int i = 0; i < nSize; ++i) {
		if ( m_vStrokeRays[i].pHitObject == NULL )
			return false;
	}
	return true;
}

bool StrokeData::IsSurfaceLoop()
{
	if ( m_vStrokeRays.size() < 1 )
		return false;

	size_t nSize = m_vPolyLineRays.size();
	for (unsigned int i = 0; i < nSize; ++i) {
		if ( m_vPolyLineRays[i].pHitObject == NULL )
			return false;
	}
	return true;
}


bool StrokeData::IsClosedLoop()
{
	return m_polyLine.GetNumVertices() > 3 && m_polyLine.IsClosed() && m_polyLine.IsSimple();
}

bool StrokeData::IsLineStroke()
{
	if ( m_vStrokeRays.size() == 0 )
		return false;

	Wml::Segment2f segLine;
	segLine.Origin() = m_vStrokeRays.front().vScreen;
	segLine.Direction() = m_vStrokeRays.back().vScreen - segLine.Origin();
	float fLineLenSqr = segLine.Direction().SquaredLength();
	fLineLenSqr /= 200.0f;

	int nLinearFailed = 0;
	size_t nVerts = m_vStrokeRays.size();
	for (unsigned int i = 0; i < nVerts; ++i) {
		float fDistSqr = SqrDistance( m_vStrokeRays[i].vScreen, segLine );
		if (fDistSqr > fLineLenSqr)
			++nLinearFailed;
	}

	if ( (float)nLinearFailed / (float)(nVerts) < 0.1f )
		return true;
	else
		return false;
}

void ISketchInterfaceComponent::AddSuggestionListItem( rmsui::HUDButton * pButton )
{
	// [RMS: probably should do some sanity checking here to ensure that
	//   we are actually in the UpdateSuggestionList() callback...
	m_pSketchManager->AddSuggestionToolbarButton( pButton );
}

// access stroke data (avoids having to copy it everywhere...)
StrokeData & ISketchInterfaceComponent::GetCurrentStrokeData()
{
	return m_pSketchManager->GetStrokeData();
}


class SketchActionButtonHandler : public rmsui::HUDButtonEventHandler
{
public:
	enum SketchActionType {
		ClearSketch,
		BlobbySweep,
		LinearSweep,
		BlendBlobbySweep,
		CutSurface,
		Revolution,
		SurfacePointsLine,
		ResizeSweep,
		JoinUnion,
		JoinBlend
	};

	SketchActionButtonHandler( CSketchManager * pSketchManager, SketchActionType eType )
	{
		m_pSketchManager = pSketchManager;
		m_eType = eType;
	}
	~SketchActionButtonHandler() {}

	virtual void OnCross() {
		switch(m_eType) {
			case ClearSketch:
				m_pSketchManager->OnEditClearstrokes();
				break;

			case BlobbySweep:
				m_pSketchManager->NotifyUserRequestedSketchOption( SketchOptionsDialogListener::WyvillSweep );
				break;

			case LinearSweep:
				m_pSketchManager->NotifyUserRequestedSketchOption( SketchOptionsDialogListener::LinearSweep );
				break;

			case BlendBlobbySweep:
				m_pSketchManager->NotifyUserRequestedSketchOption( SketchOptionsDialogListener::SurfaceAddBlendSweep );
				break;

			case CutSurface:
				m_pSketchManager->NotifyUserRequestedSketchOption( SketchOptionsDialogListener::SurfaceRemoveCut );
				break;

			case Revolution:
				m_pSketchManager->NotifyUserRequestedSketchOption( SketchOptionsDialogListener::Revolution );
				break;

			case SurfacePointsLine:
				m_pSketchManager->NotifyUserRequestedSketchOption( SketchOptionsDialogListener::SurfacePointsLine );
				break;

			case ResizeSweep:
				m_pSketchManager->NotifyUserRequestedSketchOption( SketchOptionsDialogListener::ResizeSweep );
				break;

			case JoinUnion:
				m_pSketchManager->NotifyUserRequestedSketchOption( SketchOptionsDialogListener::JoinUnion );
				break;

			case JoinBlend:
				m_pSketchManager->NotifyUserRequestedSketchOption( SketchOptionsDialogListener::JoinBlend );
				break;

			default:
				ASSERT(0);
		}
	}

protected:
	SketchActionType m_eType;
	CSketchManager * m_pSketchManager;
};




class DraggableParamButtonHandler : public rmsui::HUDButtonEventHandler
{
public:
	DraggableParamButtonHandler( rmsui::GenericFloatParameter * pParameter,
		float fMinDeltaScale = 0.1f, float fMaxDeltaScale = 0.5f, float fFalloffRange = 1.0f)
	{
		m_pParameter = pParameter;
		m_fMinDeltaScale = fMinDeltaScale;
		m_fFalloffRange = fFalloffRange;
		m_fMaxDeltaScale = fMaxDeltaScale - fMinDeltaScale;
	}
	~DraggableParamButtonHandler() {}

	void SetTarget( rmsui::GenericFloatParameter * pParameter ) {
		m_pParameter = pParameter;
	}

	virtual void OnDragDelta( const Wml::Vector2f & vDelta ) 
	{
		if ( m_pParameter ) {
			float fCurValue = m_pParameter->GetValueFloat();
			float fMinValue = m_pParameter->GetMinValue();

			float fDeltaScale = m_fMinDeltaScale;
			float fDist = fabsf(fCurValue - fMinValue);
			if ( fDist < m_fFalloffRange ) {
				fDist /= m_fFalloffRange;
				float fValue = (1.0f - fDist*fDist);
				if ( fValue > 0 ) {
					fValue = fValue*fValue*fValue;
					fDeltaScale += (1.0f - fValue) * m_fMaxDeltaScale;
				}

			} else {
				fDeltaScale += m_fMaxDeltaScale;
			}

			float fDelta = vDelta.X() * fDeltaScale + vDelta.Y() * fDeltaScale;
			m_pParameter->SetValueFloat( 
				m_pParameter->GetValueFloat() + fDelta );
		}
	}

protected:
	rmsui::GenericFloatParameter * m_pParameter;
	float m_fMinDeltaScale;
	float m_fMaxDeltaScale;
	float m_fFalloffRange;
};



CSketchManager * CSketchManager::s_pSingleton = NULL;
CSketchManager * CSketchManager::Get( )
{
	if ( s_pSingleton == NULL ) {
		s_pSingleton = new CSketchManager( CSceneManager::Get() );
	}
	return s_pSingleton;
}


IMPLEMENT_DYNAMIC(CSketchManager, CWnd)
CSketchManager::CSketchManager( CSceneManager * pSceneManager )
	: m_sketchOptionsDialog(this, this)
{
	m_pSceneManager = pSceneManager;
	m_bInSketch = false;
	m_pLastSketchObject = NULL;
	m_bCuttingPlaneEnabled = false;

	m_pSuggestionListToolbar = NULL;
	m_pClearSketchButton = NULL;
	m_pBlobbySweepItemButton = NULL;
	m_pLinearSweepItemButton = NULL;
	m_pBlendBlobbySweepItemButton = NULL;
	m_pCutItemButton = NULL;
	m_pRevolutionItemButton = NULL;
	m_pSurfacePointsLineItemButton = NULL;
	m_pResizeSweepItemButton = NULL;
	m_pJoinUnionItemButton = NULL;
	m_pJoinBlendItemButton = NULL;

	m_pClearSketchButtonHandler = NULL;
	m_pBlobbySweepItemButtonHandler = NULL;
	m_pLinearSweepItemButtonHandler = NULL;
	m_pBlendBlobbySweepItemButtonHandler = NULL;
	m_pCutItemButtonHandler = NULL;
	m_pRevolutionItemButtonHandler = NULL;
	m_pSurfacePointsLineItemButtonHandler = NULL;
	m_pResizeSweepItemButtonHandler = NULL;
	m_pJoinUnionItemButtonHandler = NULL;
	m_pJoinBlendItemButtonHandler = NULL;

	InitializeSuggestionListToolbar();

	// register ourself as a sketch component (temporary!)
	RegisterSketchComponent( this );
}

CSketchManager::~CSketchManager()
{
}

BEGIN_MESSAGE_MAP(CSketchManager, CWnd)
	ON_COMMAND(ID_EDIT_CLEARSTROKES, OnEditClearstrokes)
	ON_UPDATE_COMMAND_UI(ID_EDIT_CLEARSTROKES, OnUpdateEditClearstrokes)
	ON_COMMAND(ID_EDIT_POPSTROKE, OnEditPopstroke)
END_MESSAGE_MAP()


void CSketchManager::RegisterChangeListener( IChangeListener * pListener )
{
	m_setListeners.insert(pListener);
}

void CSketchManager::RegisterSketchComponent( ISketchInterfaceComponent * pComponent )
{
	m_vSketchComponents.push_back( pComponent );

	// set internal manager pointer
	pComponent->SetSketchManager( this );
}

/*
 * IGestureEventListener handlers
 */
bool CSketchManager::OnLineStroke( rmsui::GestureLineStroke * pLineStroke )
{
	// if line stroke is basically to left, we pop current sketch
	Wml::Vector2f vLine( pLineStroke->EndPoint2() - pLineStroke->EndPoint1() );
	vLine.Normalize();
	if ( vLine.Dot( - Wml::Vector2f::UNIT_X ) > 0.9f && m_strokeFactory.GetNumStrokes() > 0 ) {
		OnEditPopstroke();
		return true;
	}else if ( TryLineStrokeObjectInteractions( pLineStroke->EndPoint1(), pLineStroke->EndPoint2() ) ) {
		return true;
	} else
		return false;
}

bool CSketchManager::OnGeneralStroke( rmsui::GestureStroke * pStroke )
{
	const std::vector<Wml::Vector2f> & vVertices = pStroke->GetVertices();
	size_t nCount = vVertices.size();
	if ( nCount < 3 )
		return false;

	MouseDownEvent( vVertices[0] );
	for ( unsigned int i = 1; i < nCount-1; ++i )
		MouseMoveEvent( vVertices[i] );
	MouseUpEvent( vVertices[nCount-1] );

	return true;
}



void CSketchManager::OnCuttingPlaneEnabled( const Wml::Frame3f & vPlaneFrame )
{
	m_bCuttingPlaneEnabled = true;
}

void CSketchManager::OnCuttingPlaneDisabled( )
{
	m_bCuttingPlaneEnabled = false;
}


void CSketchManager::OnTreeSelectionChanged( const std::vector<rmsui::SceneObject *> & newSelection )
{
	m_vSelection = newSelection;
}

void CSketchManager::OnMousePickSelectionChanged( const std::vector<rmsui::SceneObject *> & newSelection )
{
	m_vSelection = newSelection;
}



void CSketchManager::NotifySketchChanged()
{
	std::set<IChangeListener *>::iterator cur(m_setListeners.begin()), end(m_setListeners.end());
	while (cur != end)
		(*cur++)->OnSketchChanged();
}



void CSketchManager::SetSketchingEnabled( bool bSketchingEnabled )
{
	m_bSketchingEnabled = bSketchingEnabled;
	ShowSketchOptionDialog( m_bSketchingEnabled );
}



void CSketchManager::TryStrokeObjectInteractions()
{
	// compute new intersection info
	UpdateStrokeData();

	if ( m_strokeData.IsCuttingPlaneStroke() ) {
		SetCuttingPlane();
		return;
	}

	UpdateSuggestionListToolbar( );
}


CSketchOptionsDialog::SketchOptions CSketchManager::GetSketchOptions( StrokeData & strokeData )
{
	// if stroke data is empty, just return
	CSketchOptionsDialog::SketchOptions newOptions;
	if ( strokeData.StrokeSize() == 0 )
		return newOptions;

	// get current polyline
	bool bClosedContours = strokeData.IsClosedLoop();

	bool bIsCutOutStroke =  strokeData.IsCutOutStroke();
	bool bIsAddOnStroke =  strokeData.IsAddOnStroke();
	bool bIsJoinStroke =  strokeData.IsJoinStroke();
	bool bIsSurfaceStroke =  strokeData.IsSurfaceStroke();

	if (bClosedContours && bIsCutOutStroke)
		newOptions.insert( SketchOptionsDialogListener::SurfaceRemoveCut );
	if (bClosedContours && bIsAddOnStroke) {
		newOptions.insert( SketchOptionsDialogListener::SurfaceAddBlendSweep );
		newOptions.insert( SketchOptionsDialogListener::SurfaceAddBlendRevolution );
	}

	if ( bIsJoinStroke ) {
		newOptions.insert( SketchOptionsDialogListener::JoinUnion );
		newOptions.insert( SketchOptionsDialogListener::JoinBlend );
	}

	if ( bClosedContours )
		newOptions.insert( SketchOptionsDialogListener::Revolution );

	// always insert these? Maybe do some stroke tests like closure, etc....
	if ( bClosedContours ) {
		newOptions.insert( SketchOptionsDialogListener::LinearSweep );
		newOptions.insert( SketchOptionsDialogListener::WyvillSweep );
		newOptions.insert( SketchOptionsDialogListener::BezierSweep );
	}

	if ( bIsSurfaceStroke )
		newOptions.insert( SketchOptionsDialogListener::SurfacePointsLine );

	return newOptions;
}



//! this is called from OnLineStroke....
bool CSketchManager::TryLineStrokeObjectInteractions( const Wml::Vector2f & vStart, const Wml::Vector2f & vEnd )
{
	// [RMS: this is disabled for now because it is tricky-er to port over to the new sketch interface
	//  architecture. However, it may not be worth fixing, because it is kind of awkward anyway...
#if 0

	// ok, currently the only thing we do here is re-size linear sweeps...

	if ( InMultiStrokeSketch() )
		return false;

	if ( m_vSelection.size() != 1 )
		return false;

	rmsui_bt::CappedLinearSweepSceneObject * pSweep = 
		dynamic_cast<rmsui_bt::CappedLinearSweepSceneObject *>( m_vSelection[0] );
	if ( pSweep == NULL )
		return false;

	// ok, check that line is approximately parallel w/ our sweep...
	const unsigned int * pViewport = m_pCamera->GetViewport();
	Wml::Ray3f ray1, ray2;
	m_pCamera->GetPickRay( (int)vStart.X(), pViewport[3] - 1 - (int)vStart.Y(), 
		pViewport[2], pViewport[3], ray1.Origin(), ray1.Direction() );
	m_pCamera->GetPickRay( (int)vEnd.X(), pViewport[3] - 1 - (int)vEnd.Y(), 
		pViewport[2], pViewport[3], ray2.Origin(), ray2.Direction() );
	Wml::Vector3f vPoint1( ray1.Origin() + ray1.Direction() );
	Wml::Vector3f vPoint2( ray2.Origin() + ray2.Direction() );
	Wml::Vector3f vPerpDir( vPoint2 - vPoint1 );
	vPerpDir.Normalize();

	Wml::Frame3f vFrame;
	pSweep->GetSweepNode()->GetFrame(vFrame);
	Wml::Vector3f vSweepAxis( vFrame.Axis( Wml::Frame3f::AxisZ ) );
	float fDot = fabsf(vSweepAxis.Dot(vPerpDir));
	if ( fDot < 0.97f ) {
		return false;
	}

	// ok, we want to make a resize sweep option. But (!) we need 
	// to figure out the resize length. So compute closest-points between
	// the rays and the segment of the sweep node
	Wml::Line3f sweepRay;
	float fCurLength = pSweep->GetSweepNode()->GetSweep()->GetLength();
	sweepRay.Origin() = vFrame.Origin();
	sweepRay.Direction() = vSweepAxis;
	float fParamRay, fParamSweep;
	Wml::SqrDistance( sweepRay, ray1, &fParamSweep, &fParamRay );
	m_vLastPoint1 = Wml::Vector3f (sweepRay.Origin() + sweepRay.Direction() * fParamSweep );
	Wml::SqrDistance( sweepRay, ray2, &fParamSweep, &fParamRay );
	m_vLastPoint2 = Wml::Vector3f (sweepRay.Origin() + sweepRay.Direction() * fParamSweep );


	CSketchOptionsDialog::SketchOptions newOptions;
	newOptions.insert( SketchOptionsDialogListener::ResizeSweep );
	UpdateSuggestionListToolbar( newOptions );

	return true;
#endif
	
	return false;
}


void CSketchManager::UpdateStrokeData()
{
	m_strokeData.Clear();

	// [RMS: this is an incredibly ugly hack...]

	// get blobtree from scene manager and lock it for the rest of this function...
	rmsui_bt::BlobTreeSceneObject * pBlobTreeSO = m_pSceneManager->GetCurrentBlobTree();
	SCOPELOCK_BLOBTREE( pBlobTreeSO );

	// find root node
	rmsui_bt::RootNodeSceneObject * pRootObject = pBlobTreeSO->GetRootNode();
	rmsbt::RootNode * pRootNode = pRootObject->GetRootNode();

	// get children of root node object
	const std::vector< rmsui::SceneObject *> & vGenericChildren = pRootObject->GetChildren();
	std::vector<rmsui_bt::BlobTreeNodeSceneObject *> vChildren;
	size_t nChildren = vGenericChildren.size();
	for ( unsigned int i = 0; i  < nChildren; ++i ) {
		rmsui_bt::BlobTreeNodeSceneObject * pObject = 
			dynamic_cast<rmsui_bt::BlobTreeNodeSceneObject *>( vGenericChildren[i] );
		if ( pObject != NULL )
			vChildren.push_back(pObject);
	}
	nChildren = vChildren.size();

	// this should be cached in blob tree nodes !
	Wml::AxisAlignedBox3f bounds;
	pRootNode->FieldBounds(bounds);

	// necessary stuff to construct screen rays...
	const unsigned int * pViewport = m_pCamera->GetViewport();
	Wml::Ray3f ray;

	// iterate over vertices of current sketch
	unsigned int nStrokes = m_strokeFactory.GetNumStrokes();
	for (unsigned int i = 0; i < nStrokes; ++i) {
		rmssketch::Stroke * pStroke = m_strokeFactory.GetStroke(i);

		unsigned int nVertices = pStroke->GetNumVertices();
		for (unsigned int j = 0; j < nVertices; ++j) {

			StrokeData::RayVertexInfo rayInfo;
			rayInfo.nStroke = i;
			rayInfo.pHitObject = NULL;
			rayInfo.vScreen = pStroke->GetVertex(j);
			
			// try ray intersection
			m_pCamera->GetPickRay( (int)rayInfo.vScreen.X(), pViewport[3] - 1 - (int)rayInfo.vScreen.Y(), 
				pViewport[2], pViewport[3], ray.Origin(), ray.Direction() );

			// [RMS: this is broken, since we are no longer doing 'findintersection' with
			//   root node, but with root-node children...
			if ( m_bCuttingPlaneEnabled )
				pRootNode->EnableCuttingPlane(true);

			float fMinParam = std::numeric_limits<float>::max();
			for ( unsigned int k = 0; k < nChildren; ++k ) {
				rmsbt::BlobTreeNode * pObject = vChildren[k]->GetBlobTreeNode();

				Wml::Vector3f vIntersect;
				if ( pObject->FindIntersection( ray.Origin(), ray.Direction(), vIntersect, &bounds) ) {
					float fLength = (vIntersect - ray.Origin()).Dot(ray.Direction());
					if ( fLength < fMinParam ) {
						fMinParam = fLength;
						rayInfo.pHitObject = vChildren[k];
						rayInfo.vWorld = vIntersect;
					}
				}
			}


			//if (pBlobTree->FindIntersection( ray.Origin(), ray.Direction(), rayInfo.vWorld, &bounds))
			//	rayInfo.pHitObject = m_pSceneManager->GetCurrentBlobTree();

			if ( m_bCuttingPlaneEnabled )
				pRootNode->EnableCuttingPlane(false);

			m_strokeData.AppendStrokeVertexInfo( rayInfo );
		}
	}

	// set stroke data polyline
	m_strokeData.PolyLine() = m_strokeFactory.GetCurrentPolyLine();

	// compute polyline / surface intersections
	rmssketch::PolyLine & polyLine = m_strokeFactory.GetCurrentPolyLine();
	unsigned int nVertices = polyLine.GetNumVertices();
	for (unsigned int i = 0; i < nVertices; ++i) {

		StrokeData::RayVertexInfo rayInfo;
		rayInfo.nStroke = 0xFFFFFFFF;
		rayInfo.pHitObject = NULL;
		rayInfo.vScreen = polyLine.GetVertex(i);

		// try ray intersection
		m_pCamera->GetPickRay( (int)rayInfo.vScreen.X(), pViewport[3] - 1 - (int)rayInfo.vScreen.Y(), 
			pViewport[2], pViewport[3], ray.Origin(), ray.Direction() );

		// [RMS: this is broken, since we are no longer doing 'findintersection' with
		//   root node, but with root-node children...
		if ( m_bCuttingPlaneEnabled )
			pRootNode->EnableCuttingPlane(true);

		float fMinParam = std::numeric_limits<float>::max();
		for ( unsigned int k = 0; k < nChildren; ++k ) {
			rmsbt::BlobTreeNode * pObject = vChildren[k]->GetBlobTreeNode();

			Wml::Vector3f vIntersect;
			if ( pObject->FindIntersection( ray.Origin(), ray.Direction(), vIntersect, &bounds) ) {
				float fLength = (vIntersect - ray.Origin()).Dot(ray.Direction());
				if ( fLength < fMinParam ) {
					fMinParam = fLength;
					rayInfo.pHitObject = vChildren[k];
					rayInfo.vWorld = vIntersect;
				}
			}
		}

		if ( m_bCuttingPlaneEnabled )
			pRootNode->EnableCuttingPlane(false);

		m_strokeData.AppendPolyLineVertexInfo( rayInfo );
	}

}





void CSketchManager::Compute3DFrame( Wml::AxisAlignedBox2f & initialBox, Wml::Frame & frame, 
									 float & fWidth, float & fHeight, 
									 bool bUseAverageIntersectionDepth, rmsui::SceneObject * pHitObject )
{
	const unsigned int * pViewport = m_pCamera->GetViewport();
	Wml::Ray3f ray;
	Wml::Vector3f vAxisX, vAxisY, vAxisZ;

	Wml::Vector3f vCenter( initialBox.GetXMin() + 0.5f * (initialBox.GetXMax() - initialBox.GetXMin() ),
							initialBox.GetYMin() + 0.5f * (initialBox.GetYMax() - initialBox.GetYMin() ), 0.0f );

	// project center point to get "z" vector of frame
	m_pCamera->GetPickRay( (int)vCenter.X(), pViewport[3] - 1 - (int)vCenter.Y(), 
							pViewport[2], pViewport[3], ray.Origin(), ray.Direction() );

	// reverse direction and normalize
	vAxisZ = ray.Direction();
	vAxisZ *= -1.0f;
	vAxisZ.Normalize();

	Wml::Matrix3f vAlignMatrix;
	Wml::ComputeAlignZAxisMatrix( vAxisZ, vAlignMatrix );

	Wml::Vector3f vTest( vAlignMatrix * vAxisZ );


	// Y axis is up vector...
	vAxisY = m_pCamera->GetUp();

	// x axis is  Y x Z normalized...
	vAxisX = vAxisY.Cross( vAxisZ );
	vAxisX.Normalize();

	frame.SetFrame( vAxisX, vAxisY, vAxisZ );

	Wml::Plane3f plane;
	plane.SetNormal( vAxisZ );
	plane.SetConstant( plane.GetNormal().Dot( Wml::Vector3f::ZERO ) );


	int nHits;
	Wml::Vector3f vHits[2];	
	if (bUseAverageIntersectionDepth && m_strokeData.StrokeSize() > 0) {

#if 1

		// HAAAAACK if pHitObject is NULL try getting current selection
		const std::vector<rmsui::SceneObject *> & vSelection = 
			CSceneManager::Get()->GetCurrentSelectionSet();
		rmsui_bt::BlobTreeNodeSceneObject * pHackObject = NULL;
		if ( vSelection.size() == 1 ) {
			_RMSInfo("HAAAAAACK in CSketchManager::Compute3DFrame\n");
			pHackObject = (rmsui_bt::BlobTreeNodeSceneObject *)vSelection[0];
		}


		// project all intersection points onto ray and take average as depth...
		float fMaxProj = 0.0f;
		float fSumProj = 0.0f;
		float fInitialProj = ray.Origin().Dot(ray.Direction());
		size_t nSize = m_strokeData.StrokeSize();
		for (unsigned int i = 0; i < nSize; ++i) {
			if ( m_strokeData.GetStrokeVertexInfo(i).pHitObject != NULL && 
				 (m_strokeData.GetStrokeVertexInfo(i).pHitObject == pHitObject || pHitObject == NULL) ) {

				// HAAAACKy mc hacken-stein...
				if ( pHackObject ) {
					Wml::Vector3f vOrigin, vDirection;
					const Wml::Vector2f & vScreen = m_strokeData.GetStrokeVertexInfo(i).vScreen;
					m_pCamera->GetPickRay((int)vScreen.X(), pViewport[3] - 1 - (int)vScreen.Y(), 
											pViewport[2], pViewport[3], vOrigin, vDirection );
					Wml::Vector3f vHit;
					bool bHit =	pHackObject->GetBlobTreeNode()->GetScalarField()->FindIntersection(vOrigin, vDirection, vHit );
					if (! bHit )
						continue;
				}
				

				float fProj = m_strokeData.GetStrokeVertexInfo(i).vWorld.Dot( ray.Direction() );
				fProj -= fInitialProj;
				if (fProj > fMaxProj) fMaxProj = fProj;
				fSumProj += fProj;
			}
		}
		
		fSumProj /= (float)nSize;
		vHits[0] = ray.Origin() + fMaxProj * ray.Direction();
#else

		size_t nSize = m_strokeData.StrokeSize();
		std::vector<float> vProjBuf;
		vProjBuf.resize(nSize);

		float fInitialProj = ray.Origin().Dot(ray.Direction());
		float fMaxProj = 0.0f;
		float fSumProj = 0.0f;

		// fill buffer of projection values
		float fFirst = 0.0f;
		for (unsigned int i = 0; i < nSize; ++i) {
			if ( m_strokeData.GetStrokeVertexInfo(i).pHitObject != NULL && 
				 (m_strokeData.GetStrokeVertexInfo(i).pHitObject == pHitObject || pHitObject == NULL) ) {
					vProjBuf[i] = m_strokeData.GetStrokeVertexInfo(i).vWorld.Dot( ray.Direction() );
					vProjBuf[i] -= fInitialProj;
			} else
				vProjBuf[i] = std::numeric_limits<float>::min();
		}

		// ok now break up buffer into continuous segments
		std::vector<unsigned int> vSegments;
		vSegments.push_back(0);
		float fPrev = vProjBuf[0];
		float fPrevDelta = fabsf(vProjBuf[1] - vProjBuf[0]);
		for ( unsigned int i = 0; i < nSize; ++i ) {
			if ( vProjBuf[i] == std::numeric_limits<float>::min() ) {
				if ( fPrev != 0.0f )
					vSegments.push_back(i);
				fPrev = 0.0f;
				fPrevDelta = -1;	// need to be able to *5 w/o overflow...
				continue;
			}
			float fDelta = fabsf(vProjBuf[i] - fPrev);
			if ( fDelta > 25.0f * fPrevDelta && i > 1 ) {  // new segment
				_RMSInfo("%d: %f %f\n", i, fDelta, fPrevDelta);
				vSegments.push_back(i);
				//fPrevDelta = 0.0f;
			} 
			fPrevDelta = fabsf(vProjBuf[i] - fPrev);
			fPrev = vProjBuf[i];
		}

		size_t nNumSegs = vSegments.size();
		_RMSInfo("num segments: %d\n", nNumSegs);

		// ok, now if 1 segment, take max of entire thing, otherwise
		// take max of first & last segmentss
		if ( nNumSegs == 1 ) {
			fMaxProj = *std::max_element( vProjBuf.begin(), vProjBuf.end() );
		} else {
			unsigned int nStart = vSegments[0];
			unsigned int nStop = vSegments[1];
			for ( unsigned int i = nStart; i < nStop; ++i ) {
				if ( vProjBuf[i] > fMaxProj )
					fMaxProj = vProjBuf[i];
			}
			nStart = vSegments[ nNumSegs - 2 ];
			nStop = vSegments[ nNumSegs - 1 ];
			for ( unsigned int i = nStart; i < nStop; ++i ) {
				if ( vProjBuf[i] > fMaxProj )
					fMaxProj = vProjBuf[i];
			}
		}

		vHits[0] = ray.Origin() + fMaxProj * ray.Direction();
#endif
	} else {

		// find intersection of pick ray with plane through origin. This is our frame origin.
		bool bResult = Wml::FindIntersection( plane, ray, nHits, vHits, true );
		ASSERT(bResult && nHits > 0);
	}

	// create plane in line with "z" axis of frame. 
	frame.Origin() = vHits[0];
	plane.SetConstant( plane.GetNormal().Dot(frame.Origin()) );

	// now project other bbox points to get width and height
	m_pCamera->GetPickRay( (int)initialBox.GetXMax(), pViewport[3] - 1 - (int)vCenter.Y(), pViewport[2], pViewport[3], ray.Origin(), ray.Direction() );
	bool bResult = Wml::FindIntersection( plane, ray, nHits, vHits, true );
	ASSERT(bResult && nHits > 0);

	fWidth = ( vHits[0] - frame.Origin() ).Length();

	m_pCamera->GetPickRay( (int)vCenter.X(), pViewport[3] - 1 - (int)initialBox.GetYMax(), pViewport[2], pViewport[3], ray.Origin(), ray.Direction() );
	bResult = Wml::FindIntersection( plane, ray, nHits, vHits, true );
	ASSERT(bResult && nHits > 0);

	fHeight = ( vHits[0] - frame.Origin() ).Length();
}



void CSketchManager::Render( bool bDrawNormals )
{
	glPushAttrib(GL_ENABLE_BIT | GL_LINE_BIT | GL_POINT_BIT );


	//glDisable(GL_LIGHTING);
	//glPointSize(5.0f);
	//size_t nHits = m_lastIntersection.size();
	//glBegin(GL_POINTS);
	//for (unsigned int i = 0; i < nHits; ++i) {
	//	glVertex3f( m_lastIntersection[i].vWorld.X(), m_lastIntersection[i].vWorld.Y(), m_lastIntersection[i].vWorld.Z() );
	//}
	//glEnd();


	//glEnable(GL_LINE_SMOOTH);
	glDisable(GL_DEPTH_TEST);
	glDisable(GL_LIGHTING);
	glDisable(GL_BLEND);
	glBlendFunc(GL_ZERO, GL_ONE);

//	if (bDrawNormals)
//		m_strokeFactory.DrawStrokeNormals( 10.0f );

	//glLineWidth(4.0f);
	//glColor3f(1.0f, 0.0f, 0.0f);
	//m_strokeFactory.DrawCurrentPolyLine();

#define SCALEFACTOR 1

	//glColor3f(0.0f, 0.0f, 1.0f);
	//m_strokeFactory.DrawStrokes();

	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glLineWidth(6.0f * SCALEFACTOR);
	glColor4f(0.0f, 0.0f, 1.0f, 0.3f);
	m_strokeFactory.DrawStrokes( );
	glDisable(GL_BLEND);

	if ( GetStrokeData().IsClosedLoop() ) {
		glEnable(GL_LINE_STIPPLE);
		glLineStipple(1 * SCALEFACTOR, 0xF0F0);
		glLineWidth(3.0f * SCALEFACTOR);
		glColor3f(0.0f, 0.0f, 0.0f);
		m_strokeFactory.DrawCurrentPolyLine( );
		glDisable(GL_LINE_STIPPLE);
	}

	glPopAttrib();	

	//DrawSketchHUD();
}


void CSketchManager::RenderCurrentStroke()
{
	glPushAttrib(GL_ENABLE_BIT | GL_LINE_BIT);
	glDisable(GL_LIGHTING);

	glColor3f(0.0f, 0.0f, 1.0f);
	glLineWidth(4.0f);
	m_strokeFactory.DrawCurrentStroke( );

	glPopAttrib();
}


void CSketchManager::DrawingAreaChanged( const unsigned int * pViewport )
{
	m_strokeFactory.SetDrawingAreaSize(pViewport);
}

void CSketchManager::MouseDownEvent( const Wml::Vector2f & vMousePoint )
{
	// start a new stroke in factory
	m_strokeFactory.BeginNewStroke();
	m_strokeFactory.AddStrokeVertex(vMousePoint);

	m_bInSketch = true;
}

void CSketchManager::MouseUpEvent( const Wml::Vector2f & vMousePoint )
{
	// end stroke in factory
	if ( m_bInSketch) {
		rmssketch::StrokeFactory::StrokeType eType = m_strokeFactory.EndStroke();

		// TODO CLEANUP
		// do something w/ type ??
		if (eType == rmssketch::StrokeFactory::SharpScribbleStroke && m_strokeFactory.GetNumStrokes() > 1 )
			m_strokeFactory.EraseStrokes(true);
		else if (eType == rmssketch::StrokeFactory::MultipleLoopsStroke && m_strokeFactory.GetNumStrokes() > 1 )
			m_strokeFactory.SmoothStrokes(true);

		TryStrokeObjectInteractions();
	
		m_bInSketch = false;

		// clear last-sketched item
		if ( eType != rmssketch::StrokeFactory::IgnoredBecauseTooSmall ) {
			m_pLastSketchObject = NULL;
			m_sketchOptionsDialog.SetSliderParams(0.0f, 1.0f, 0.5f, "No Option");
		} 

		// try HUD
		//if ( eType == rmssketch::StrokeFactory::IgnoredBecauseTooSmall ) {
		//	InteractSketchHUD_ResolutionButtons( vMousePoint );
		//}

	}

	NotifySketchChanged();
}

void CSketchManager::MouseMoveEvent( const Wml::Vector2f & vMousePoint )
{
	if (m_bInSketch)
		m_strokeFactory.AddStrokeVertex(vMousePoint);
}




void CSketchManager::ShowSketchOptionDialog( bool bShow )
{
	if (bShow)
		m_sketchOptionsDialog.ShowWindow(SW_SHOW);
	else
		m_sketchOptionsDialog.ShowWindow(SW_HIDE);
}


void CSketchManager::NotifyUserRequestedSketchOption( SketchInteractionOption option )
{
	switch(option) {
		case SketchOptionsDialogListener::LinearSweep:
			{
			m_pLastSketchObject = AddLinearSweep( false );
			}
			break;

		case SketchOptionsDialogListener::WyvillSweep:
			{
			m_pLastSketchObject = AddLinearSweep( true );
			}
			break;

		case SketchOptionsDialogListener::Revolution:
			m_pLastSketchObject = AddSurfaceRevolution();
			break;

		case SketchOptionsDialogListener::SurfaceAddBlendSweep:
			{
			m_pLastSketchObject = AddSurfaceAddBlendSweep();
			rmsui_bt::RicciBlendSceneObject * pObject = 
				dynamic_cast<rmsui_bt::RicciBlendSceneObject *>(m_pLastSketchObject);
			ASSERT(pObject != NULL);
			rmsui_bt::CappedLinearSweepSceneObject * pSweep = 
				dynamic_cast<rmsui_bt::CappedLinearSweepSceneObject *>(pObject->GetChildren().back());
			ASSERT(pSweep != NULL);
			}
			break;

		case SketchOptionsDialogListener::SurfaceAddBlendRevolution:
			{
			m_pLastSketchObject = AddSurfaceAddBlendRevolution();
			}
			break;

		case SketchOptionsDialogListener::SurfaceRemoveCut:
			{
			m_pLastSketchObject = AddSurfaceRemoveCut( );
			}
			break;

		case SketchOptionsDialogListener::SurfacePointsLine:
			{
			m_pLastSketchObject = AddPoinstsLine(true);
			}
			break;

		case SketchOptionsDialogListener::BezierSweep:
			m_pLastSketchObject = AddBezierSweep();
			break;

		case SketchOptionsDialogListener::ResizeSweep:
			m_pLastSketchObject = ResizeSweep();
			break;

		case SketchOptionsDialogListener::JoinUnion:
			m_pLastSketchObject = JoinNodes( SketchOptionsDialogListener::JoinUnion );
			break;

		case SketchOptionsDialogListener::JoinBlend: {
			m_pLastSketchObject = JoinNodes( SketchOptionsDialogListener::JoinBlend );
			}
			break;
	}

	m_eLastSketchType = option;
}



rmsui_bt::BlobTreeNodeSceneObject * CSketchManager::AddSurfaceRevolution()
{
	rmssketch::PolyLine & polyLine = m_strokeFactory.GetCurrentPolyLine();
	rmssketch::SketchField & sketchField = m_strokeFactory.GetCurrentSketchField(0.05);
//	rmsimplicit::DiscreteScalarField2D * pField = polyLine.MakeImplicitFieldBruteForce( &sketchField, 64 );
	rmsimplicit::DiscreteScalarField2D * pField = polyLine.MakeImplicitFieldVariational( &sketchField, SKETCH_FIELD_RESOLUTION );


	// get bounding box of polyline
	Wml::AxisAlignedBox2f box = polyLine.ComputeBoundingBox();

	Wml::Frame frame;
	float fWidth, fHeight;
	Compute3DFrame( box, frame, fWidth, fHeight, false );

	float fScale = std::max<float>(fWidth, fHeight);
	pField->SetFieldBounds( 
		Wml::AxisAlignedBox2f( -fScale*3.0f, fScale*3.0f, -fScale*3.0f, fScale*3.0f ) );

	// optimize bounding box
	pField->ClampBoundsToZeroContour();

	// clear sketch
	OnEditClearstrokes();

	return m_pSceneManager->SketchAddRevolution( pField, frame );
}



void ApplyScalingHack( rmsimplicit::DiscreteScalarField2D * pField )
{
	unsigned int nWidth = pField->GetWidth();
	unsigned int nHeight = pField->GetHeight();
	for ( unsigned int yi = 0; yi < nHeight; ++yi ) {
		for ( unsigned int xi = 0; xi < nWidth; ++xi ) {

			float fValue = pField->GetValue(xi,yi);

			float fValueOff = (fValue - 0.5f) / 0.5f;
			
			if ( xi > nWidth/2 && fValue > 0.5f ) {
				fValue *= 1.0f + (0.5f * fValueOff);
			}

			//float xOff = (float)xi / (float)nWidth;
			//if ( fValue > 0.5f )
			//	fValue *= 1.0f + (xOff * fValueOff);			

			pField->SetValue(xi,yi, fValue);

		}
	}

}

rmsui_bt::BlobTreeNodeSceneObject * CSketchManager::AddLinearSweep( bool bBlobbyEndCap )
{
	rmssketch::PolyLine & polyLine = m_strokeFactory.GetCurrentPolyLine();
	rmssketch::SketchField & sketchField = m_strokeFactory.GetCurrentSketchField(0.05);

	rmsimplicit::DiscreteScalarField2D * pField = polyLine.MakeImplicitFieldVariational( &sketchField, SKETCH_FIELD_RESOLUTION );
	if ( pField == NULL )
		return NULL;

	// get bounding box of polyline
	Wml::AxisAlignedBox2f box = polyLine.ComputeBoundingBox();

	Wml::Frame frame;
	float fWidth, fHeight;
	Compute3DFrame( box, frame, fWidth, fHeight, false );

	float fScale = std::max<float>(fWidth, fHeight);
	pField->SetFieldBounds( 
		Wml::AxisAlignedBox2f( -fScale*3.0f, fScale*3.0f, -fScale*3.0f, fScale*3.0f ) );

	// optimize bounding box
	pField->ClampBoundsToZeroContour();

	//ApplyScalingHack(pField);

	// this is a hacky thing. Compute length of object based on area. Say max area is 1. Then scale length based on area...
	//   (scaling factor is total hack...)
	float fArea = fWidth * fHeight;
	float fLength = std::sqrt(fArea);
	_RMSInfo( " Sweep Area: %f   (width: %f  height: %f)\n", fArea, fWidth, fHeight);

	// clear sketch
	OnEditClearstrokes();

	// for sweeps, split area half & half for normal sweep and put all area to endcap for blobby sweeps
	return m_pSceneManager->SketchAddSweep( pField, frame,	bBlobbyEndCap, 
		(bBlobbyEndCap) ? 0.001f : 0.7f * fLength,
		(bBlobbyEndCap) ? fLength : 0.3f * fLength );
}


rmsui_bt::BlobTreeNodeSceneObject * CSketchManager::AddBezierSweep()
{
	rmssketch::PolyLine & polyLine = m_strokeFactory.GetCurrentPolyLine();
	rmssketch::SketchField & sketchField = m_strokeFactory.GetCurrentSketchField(0.05);

	rmsimplicit::DiscreteScalarField2D * pField = polyLine.MakeImplicitFieldVariational( &sketchField, SKETCH_FIELD_RESOLUTION );

	// get bounding box of polyline
	Wml::AxisAlignedBox2f box = polyLine.ComputeBoundingBox();

	Wml::Frame frame;
	float fWidth, fHeight;
	Compute3DFrame( box, frame, fWidth, fHeight, false );

	float fScale = std::max<float>(fWidth, fHeight);
	pField->SetFieldBounds( 
		Wml::AxisAlignedBox2f( -fScale*3.0f, fScale*3.0f, -fScale*3.0f, fScale*3.0f ) );

	// optimize bounding box
	pField->ClampBoundsToZeroContour();

	// this is a hacky thing. Compute length of object based on area. Say max area is 1. Then scale length based on area...
	//   (scaling factor is total hack...)
	float fArea = fWidth * fHeight;
	float fLength = std::sqrt(fArea);
	_RMSInfo( " Sweep Area: %f   (width: %f  height: %f)\n", fArea, fWidth, fHeight);

	// [RMS HACK] do this for now since we suck...
	pField->SetFieldBounds(
		Wml::AxisAlignedBox2f( -0.3f, 0.3f, -0.3f, 0.3f ) );
	fLength = 0.25;


	// for sweeps, split area half & half for normal sweep and put all area to endcap for blobby sweeps
	return m_pSceneManager->SketchAddCurveSweep( pField, true, fLength );
}


rmsui_bt::BlobTreeNodeSceneObject * CSketchManager::AddSurfaceRemoveCut( )
{
	// find object we want to cut
	rmsui::SceneObject * pCutObject = m_strokeData.GetFirstHitObject();
	ASSERT(pCutObject);
	if ( ! pCutObject )
		return NULL;

	rmssketch::PolyLine & polyLine = m_strokeFactory.GetCurrentPolyLine();
	rmssketch::SketchField & sketchField = m_strokeFactory.GetCurrentSketchField(0.05);
	rmsimplicit::DiscreteScalarField2D * pField = polyLine.MakeImplicitFieldVariational( &sketchField, SKETCH_FIELD_RESOLUTION );

	// get bounding box of polyline
	Wml::AxisAlignedBox2f box = polyLine.ComputeBoundingBox();

	Wml::Frame frame;
	float fWidth, fHeight;
	Compute3DFrame( box, frame, fWidth, fHeight, true );

	float fScale = std::max<float>(fWidth, fHeight);
	pField->SetFieldBounds( 
		Wml::AxisAlignedBox2f( -fScale*3.0f, fScale*3.0f, -fScale*3.0f, fScale*3.0f ) );

	// optimize bounding box
	pField->ClampBoundsToZeroContour();

	// clear sketch
	OnEditClearstrokes();

	return m_pSceneManager->SketchAddSweepCut( pCutObject, pField, frame );
}



rmsui_bt::BlobTreeNodeSceneObject * CSketchManager::AddSurfaceAddBlendSweep( )
{
	// find object we want to cut
	rmsui::SceneObject * pBlendObject = m_strokeData.GetFirstHitObject();
	ASSERT(pBlendObject);
	if ( ! pBlendObject )
		return NULL;

	rmssketch::PolyLine & polyLine = m_strokeFactory.GetCurrentPolyLine();
	rmssketch::SketchField & sketchField = m_strokeFactory.GetCurrentSketchField(0.05);
	rmsimplicit::DiscreteScalarField2D * pField = polyLine.MakeImplicitFieldVariational( &sketchField, SKETCH_FIELD_RESOLUTION );

	// get bounding box of polyline
	Wml::AxisAlignedBox2f box = polyLine.ComputeBoundingBox();

	Wml::Frame frame;
	float fWidth, fHeight;
	Compute3DFrame( box, frame, fWidth, fHeight, true, pBlendObject );

	float fScale = std::max<float>(fWidth, fHeight);
	pField->SetFieldBounds( 
		Wml::AxisAlignedBox2f( -fScale*3.0f, fScale*3.0f, -fScale*3.0f, fScale*3.0f ) );

	// optimize bounding box
	pField->ClampBoundsToZeroContour();

	// this is a hacky thing. Compute length of object based on area. Say max area is 1. Then scale length based on area...
	//   (scaling factor is total hack...)
	float fArea = fWidth * fHeight;
	float fLength = std::sqrt(fArea);
	_RMSInfo( " Sweep Area: %f   (width: %f  height: %f)\n", fArea, fWidth, fHeight);

	// clear sketch
	OnEditClearstrokes();

	return m_pSceneManager->SketchAddSweepBlend( pBlendObject, pField, frame, true, 0.01f, fLength );
}



rmsui_bt::BlobTreeNodeSceneObject * CSketchManager::AddSurfaceAddBlendRevolution( )
{
	// find object we want to cut
	rmsui::SceneObject * pBlendObject = m_strokeData.GetFirstHitObject();
	ASSERT(pBlendObject);
	if ( ! pBlendObject )
		return NULL;

	rmssketch::PolyLine & polyLine = m_strokeFactory.GetCurrentPolyLine();
	rmssketch::SketchField & sketchField = m_strokeFactory.GetCurrentSketchField(0.05);
	rmsimplicit::DiscreteScalarField2D * pField = polyLine.MakeImplicitFieldVariational( &sketchField, SKETCH_FIELD_RESOLUTION );

	// get bounding box of polyline
	Wml::AxisAlignedBox2f box = polyLine.ComputeBoundingBox();

	Wml::Frame frame;
	float fWidth, fHeight;
	Compute3DFrame( box, frame, fWidth, fHeight, false );

	float fScale = std::max<float>(fWidth, fHeight);
	pField->SetFieldBounds( 
		Wml::AxisAlignedBox2f( -fScale*3.0f, fScale*3.0f, -fScale*3.0f, fScale*3.0f ) );

	// optimize bounding box
	pField->ClampBoundsToZeroContour();

	return m_pSceneManager->SketchAddRevolutionBlend( pBlendObject, pField, frame );
}



rmsui_bt::BlobTreeNodeSceneObject * CSketchManager::AddPoinstsLine( bool bSurfacePoints )
{
	std::vector<Wml::Vector3f> vPoints;

	if (bSurfacePoints) {

		// find object we want to blend to
		size_t nHits = m_strokeData.StrokeSize();
		for (unsigned int i = 0; i < nHits; ++i)
			if (m_strokeData.GetStrokeVertexInfo(i).pHitObject != NULL)
				vPoints.push_back( m_strokeData.GetStrokeVertexInfo(i).vWorld );
	} else {
		ASSERT(0);
	}

	// clear sketch
	OnEditClearstrokes();

	return m_pSceneManager->SketchAddPointLine( vPoints, 0.08f );
}


rmsui_bt::BlobTreeNodeSceneObject * CSketchManager::ResizeSweep( )
{
	OnEditClearstrokes();

	return m_pSceneManager->SketchResizeSelectedSweep( m_vLastPoint1, m_vLastPoint2 );	
}

rmsui_bt::BlobTreeNodeSceneObject * CSketchManager::JoinNodes( SketchOptionsDialogListener::SketchInteractionOption eType )
{
	// find first selection object
	rmsui::SceneObject * pFirst = m_strokeData.GetFirstHitObject();
	rmsui::SceneObject * pLast = m_strokeData.GetLastHitObject();
	if ( pFirst == NULL || pFirst == pLast ) {
		ASSERT(0);
		return NULL;
	}

	OnEditClearstrokes();
	
	if ( eType == SketchOptionsDialogListener::JoinUnion )
		return m_pSceneManager->SketchJoinNodes( pFirst, pLast, false );
	else if ( eType == SketchOptionsDialogListener::JoinBlend )
		return m_pSceneManager->SketchJoinNodes( pFirst, pLast, true );
	else
		return NULL;
}



void CSketchManager::SetCuttingPlane()
{
	// take average intersection point as origin
	Wml::Vector3f vOrigin(Wml::Vector3f::ZERO);
	Wml::Vector3f vFirst, vLast;
	Wml::Vector2f vFirstScreen, vLastScreen;
	bool bFoundFirst = false;
	int nCount = 0;
	size_t nHits = m_strokeData.StrokeSize();
	for (unsigned int i = 0; i < nHits; ++i) {
		if (m_strokeData.GetStrokeVertexInfo(i).pHitObject != NULL) {
			vOrigin += m_strokeData.GetStrokeVertexInfo(i).vWorld;
			if (!bFoundFirst) {
				vFirst = m_strokeData.GetStrokeVertexInfo(i).vWorld;
				vFirstScreen = m_strokeData.GetStrokeVertexInfo(i).vScreen;
				bFoundFirst = true;
			} else {
				vLast = m_strokeData.GetStrokeVertexInfo(i).vWorld;
				vLastScreen = m_strokeData.GetStrokeVertexInfo(i).vScreen;
			}

			++nCount;
		}
	}

	vOrigin *= 1.0f / (float)nCount;

	// ok, now compute pick rays for vFirstScreen and vLastScreen;
	const unsigned int * pViewport = m_pCamera->GetViewport();
	Wml::Ray3f rayFirst, rayLast;
	m_pCamera->GetPickRay( (int)vFirstScreen.X(), pViewport[3] - 1 - (int)vFirstScreen.Y(), 
							pViewport[2], pViewport[3], rayFirst.Origin(), rayFirst.Direction() );
	m_pCamera->GetPickRay( (int)vLastScreen.X(), pViewport[3] - 1 - (int)vLastScreen.Y(), 
							pViewport[2], pViewport[3], rayLast.Origin(), rayLast.Direction() );
	Wml::Vector3f vNormal( rayFirst.Direction().Cross( rayLast.Direction() ) );
	vNormal.Normalize();

	// ok, the cross-product of these two rays gives us the normal, but not
	// the direction. Use the ray through the *last* screen point (the final point in the 'L' used
	// to specify the clip plane) to determine the direction
	Wml::Ray3f rayUp;
	Wml::Vector2f vFinalScreen( m_strokeData.GetStrokeVertexInfo(m_strokeData.StrokeSize()-1).vScreen );
	m_pCamera->GetPickRay( (int)vFinalScreen.X(), pViewport[3] - 1 - (int)vFinalScreen.Y(), 
							pViewport[2], pViewport[3], rayUp.Origin(), rayUp.Direction() );
	Wml::Vector3f vUpDir( rayUp.Direction() - rayLast.Direction() );
	if ( vUpDir.Dot( vNormal ) > 0 )		// this is opposite because the screen coords are backwards (??)
		vNormal *= -1.0f;

	Wml::Frame3f vFrame( vOrigin, vNormal );

	m_pSceneManager->SetTemporaryCuttingPlane( vFrame );

	OnEditClearstrokes();
}



void CSketchManager::DrawSketchHUD()
{
	glPushMatrix();
	glLoadIdentity();

	const unsigned int * pViewport = m_pCamera->GetViewport();
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	glOrtho(pViewport[0], pViewport[0] + pViewport[2], 
			pViewport[1], pViewport[1] + pViewport[3], -1.0f, 1.0f );

	glPushAttrib(GL_ENABLE_BIT);
	glDisable(GL_LIGHTING);
	glDisable(GL_DEPTH_TEST);

	DrawSketchHUDResolutionButtons();

	glPopAttrib();

	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();
}

void CSketchManager::DrawSketchHUDResolutionButtons()
{
	const unsigned int * pViewport = m_pCamera->GetViewport();
	float nRight = (float)(pViewport[0] + pViewport[2]);
	float nTop = (float)(pViewport[1] + pViewport[3]);

	glColor3f(1.0f, 1.0f, 0.0f);
	glBegin(GL_QUADS);
	glVertex2f( nRight - 25.0f, nTop - 25.0f );
	glVertex2f( nRight - 5.0f, nTop - 25.0f );
	glVertex2f( nRight - 5.0f, nTop - 5.0f );
	glVertex2f( nRight - 25.0f, nTop - 5.0f );
	glEnd();
}

void CSketchManager::InteractSketchHUD_ResolutionButtons( const Wml::Vector2f & vMousePoint )
{
	const unsigned int * pViewport = m_pCamera->GetViewport();
	float nRight = (float)(pViewport[0] + pViewport[2]);
	float nTop = (float)(pViewport[1] + pViewport[3]);	

	if ( vMousePoint.X() > nRight - 25.0f && vMousePoint.X() < nRight - 5.0f &&
		 vMousePoint.Y() > nTop - 25.0f && vMousePoint.Y() < nTop - 5.0f )
	{

		CMainFrame * pMainFrame = (CMainFrame *)AfxGetMainWnd();
		CChildFrame * pFrame = (CChildFrame *)pMainFrame->GetActiveFrame();
		CView * pView = dynamic_cast<CView *>(pFrame->GetActiveView());
		if ( pView )
			pView->PostMessage(WM_CHAR, (WPARAM)'+', (LPARAM)0);
			
	}

}


void CSketchManager::InitializeSuggestionListToolbar()
{
	static const float fButtonSize = 0.125f;
//	static const float fButtonSize = 0.25f;

	m_pSuggestionListToolbar = new rmsui::HUDToolbar("SuggestionListToolbar", false);

	rmsui::HUDFileIcon * pIcon = new rmsui::HUDFileIcon(CSystemConstants::GetIconPath() + "cancel.png");
	m_pClearSketchButton = new rmsui::HUDButton("ClearSketchButton", pIcon, false, true);
	m_pClearSketchButton->Bounds() = Wml::AxisAlignedBox2f( 0.0f, fButtonSize, 0.0f, fButtonSize );

	pIcon = new rmsui::HUDFileIcon(CSystemConstants::GetIconPath() + "blobbySweep.png");
	m_pBlobbySweepItemButton = new rmsui::HUDButton("BlobbySweepButton", pIcon, false, true);
	m_pBlobbySweepItemButton->Bounds() = Wml::AxisAlignedBox2f( 0.0f, fButtonSize, 0.0f, fButtonSize );

	pIcon = new rmsui::HUDFileIcon(CSystemConstants::GetIconPath() + "linearSweep.png");
	m_pLinearSweepItemButton = new rmsui::HUDButton("LinearSweepButton", pIcon, false, true);
	m_pLinearSweepItemButton->Bounds() = Wml::AxisAlignedBox2f( 0.0f, fButtonSize, 0.0f, fButtonSize );

	pIcon = new rmsui::HUDFileIcon(CSystemConstants::GetIconPath() + "blendBlobbySweep.png");
	m_pBlendBlobbySweepItemButton = new rmsui::HUDButton("BlendBlobbySweepButton", pIcon, false, true);
	m_pBlendBlobbySweepItemButton->Bounds() = Wml::AxisAlignedBox2f( 0.0f, fButtonSize, 0.0f, fButtonSize );

	pIcon = new rmsui::HUDFileIcon(CSystemConstants::GetIconPath() + "cutSurface.png");
	m_pCutItemButton = new rmsui::HUDButton("CutSurfaceButton", pIcon, false, true);
	m_pCutItemButton->Bounds() = Wml::AxisAlignedBox2f( 0.0f, fButtonSize, 0.0f, fButtonSize );

	pIcon = new rmsui::HUDFileIcon(CSystemConstants::GetIconPath() + "revolution.png");
	m_pRevolutionItemButton = new rmsui::HUDButton("RevolutionButton", pIcon, false, true);
	m_pRevolutionItemButton->Bounds() = Wml::AxisAlignedBox2f( 0.0f, fButtonSize, 0.0f, fButtonSize );

	pIcon = new rmsui::HUDFileIcon(CSystemConstants::GetIconPath() + "surfacePointsLine.png");
	m_pSurfacePointsLineItemButton = new rmsui::HUDButton("SurfacePointsLineButton", pIcon, false, true);
	m_pSurfacePointsLineItemButton->Bounds() = Wml::AxisAlignedBox2f( 0.0f, fButtonSize, 0.0f, fButtonSize );

	pIcon = new rmsui::HUDFileIcon(CSystemConstants::GetIconPath() + "sweepResize.png");
	m_pResizeSweepItemButton = new rmsui::HUDButton("ResizeSweepButton", pIcon, false, true);
	m_pResizeSweepItemButton->Bounds() = Wml::AxisAlignedBox2f( 0.0f, fButtonSize, 0.0f, fButtonSize );

	pIcon = new rmsui::HUDFileIcon(CSystemConstants::GetIconPath() + "joinUnion.png");
	m_pJoinUnionItemButton = new rmsui::HUDButton("JoinUnionButton", pIcon, false, true);
	m_pJoinUnionItemButton->Bounds() = Wml::AxisAlignedBox2f( 0.0f, fButtonSize, 0.0f, fButtonSize );

	pIcon = new rmsui::HUDFileIcon(CSystemConstants::GetIconPath() + "joinBlend.png");
	m_pJoinBlendItemButton = new rmsui::HUDButton("JoinBlendButton", pIcon, false, true);
	m_pJoinBlendItemButton->Bounds() = Wml::AxisAlignedBox2f( 0.0f, fButtonSize, 0.0f, fButtonSize );


	// add handlers for each button
	m_pClearSketchButtonHandler = new SketchActionButtonHandler( this, SketchActionButtonHandler::ClearSketch );
	m_pClearSketchButton->AddListener( m_pClearSketchButtonHandler );

	m_pBlobbySweepItemButtonHandler = new SketchActionButtonHandler( this, SketchActionButtonHandler::BlobbySweep );
	m_pBlobbySweepItemButton->AddListener( m_pBlobbySweepItemButtonHandler );

	m_pLinearSweepItemButtonHandler = new SketchActionButtonHandler( this, SketchActionButtonHandler::LinearSweep );
	m_pLinearSweepItemButton->AddListener( m_pLinearSweepItemButtonHandler );

	m_pBlendBlobbySweepItemButtonHandler = new SketchActionButtonHandler( this, SketchActionButtonHandler::BlendBlobbySweep );
	m_pBlendBlobbySweepItemButton->AddListener( m_pBlendBlobbySweepItemButtonHandler );

	m_pCutItemButtonHandler = new SketchActionButtonHandler( this, SketchActionButtonHandler::CutSurface );
	m_pCutItemButton->AddListener( m_pCutItemButtonHandler );

	m_pRevolutionItemButtonHandler = new SketchActionButtonHandler( this, SketchActionButtonHandler::Revolution );
	m_pRevolutionItemButton->AddListener( m_pRevolutionItemButtonHandler );

	m_pSurfacePointsLineItemButtonHandler = new SketchActionButtonHandler( this, SketchActionButtonHandler::SurfacePointsLine );
	m_pSurfacePointsLineItemButton->AddListener( m_pSurfacePointsLineItemButtonHandler );

	m_pResizeSweepItemButtonHandler = new SketchActionButtonHandler( this, SketchActionButtonHandler::ResizeSweep );
	m_pResizeSweepItemButton->AddListener( m_pResizeSweepItemButtonHandler );

	m_pJoinUnionItemButtonHandler = new SketchActionButtonHandler( this, SketchActionButtonHandler::JoinUnion );
	m_pJoinUnionItemButton->AddListener( m_pJoinUnionItemButtonHandler );

	m_pJoinBlendItemButtonHandler = new SketchActionButtonHandler( this, SketchActionButtonHandler::JoinBlend );
	m_pJoinBlendItemButton->AddListener( m_pJoinBlendItemButtonHandler );

	// set toolbar position
	m_pSuggestionListToolbar->SetMinButtonSize( fButtonSize );
	m_pSuggestionListToolbar->Translate( -0.2f, -1.0f, false );

	// update toolbar
	UpdateSuggestionListToolbar();
}


void CSketchManager::AddSuggestionToolbarButton( rmsui::HUDButton * pButton )
{
	m_pSuggestionListToolbar->AppendButton( pButton );
}

void CSketchManager::UpdateSuggestionListToolbar( )
{
	m_pSuggestionListToolbar->ClearButtons();

	// always add clear button (?)
	m_pSuggestionListToolbar->AppendButton( m_pClearSketchButton );

	// try to find an interface component that wants exclusive capture
	size_t nComponents = m_vSketchComponents.size();
	ISketchInterfaceComponent * pExclusive = NULL;
	for ( unsigned int i = 0; i < nComponents && pExclusive == NULL; ++i ) {
		if ( m_vSketchComponents[i]->WantExclusiveCapture() )
			pExclusive = m_vSketchComponents[i];
	}

	if ( pExclusive ) {
		pExclusive->OnUpdateSuggestionList( m_strokeData );
	} else {
		for ( unsigned int i = 0; i < nComponents; ++i )
			m_vSketchComponents[i]->OnUpdateSuggestionList( m_strokeData );
	}

	NotifySketchChanged();
}


bool CSketchManager::WantExclusiveCapture()
{
	return false;   // never want this
}

void CSketchManager::OnUpdateSuggestionList( StrokeData & strokeData )
{
	// compute sketch options
	CSketchOptionsDialog::SketchOptions options = GetSketchOptions( strokeData );

	if ( options.find( SketchOptionsDialogListener::SurfaceAddBlendSweep ) != options.end () )
		AddSuggestionListItem( m_pBlendBlobbySweepItemButton );

	if ( options.find( SketchOptionsDialogListener::SurfaceRemoveCut ) != options.end() )
		AddSuggestionListItem( m_pCutItemButton );

	if ( options.find( SketchOptionsDialogListener::SurfacePointsLine ) != options.end() )
		AddSuggestionListItem( m_pSurfacePointsLineItemButton );

	if ( options.find( SketchOptionsDialogListener::WyvillSweep ) != options.end() )
		AddSuggestionListItem( m_pBlobbySweepItemButton );

	if ( options.find( SketchOptionsDialogListener::LinearSweep ) != options.end() )
		AddSuggestionListItem( m_pLinearSweepItemButton );

	if ( options.find( SketchOptionsDialogListener::Revolution ) != options.end() )
		AddSuggestionListItem( m_pRevolutionItemButton );

	if ( options.find( SketchOptionsDialogListener::ResizeSweep ) != options.end() )
		AddSuggestionListItem( m_pResizeSweepItemButton );

	if ( options.find( SketchOptionsDialogListener::JoinUnion ) != options.end() )
		AddSuggestionListItem( m_pJoinUnionItemButton );

	if ( options.find( SketchOptionsDialogListener::JoinBlend ) != options.end() )
		AddSuggestionListItem( m_pJoinBlendItemButton );
}




void CSketchManager::OnEditClearstrokes()
{
	m_strokeFactory.ClearStrokes();
	m_strokeData.Clear();

	UpdateSuggestionListToolbar();

	NotifySketchChanged();
}

void CSketchManager::OnUpdateEditClearstrokes(CCmdUI *pCmdUI)
{
	pCmdUI->Enable();
}

void CSketchManager::OnEditPopstroke()
{
	m_strokeFactory.RemoveLastStroke();

	TryStrokeObjectInteractions();

	NotifySketchChanged();
}
