// Ryan Schmidt - rms@unknownroad.com
// Copyright (c) 2004-2010
// Distributed under the Boost Software License, Version 1.0.
// http://www.boost.org/LICENSE_1_0.txt


#ifndef __RMS_VF_TRIANGLE_MESH_H__
#define __RMS_VF_TRIANGLE_MESH_H__

#include "rmsimplicit.h"
#include "TriangleMesh.h"

#include <WmlVector3.h>
#include <set>
#include <map>
//#include <dense_hash_map>
#include <hash_map>

#define USE_WMLEXT_MEMPOOL_EXTERNAL_DLL RMSIMPLICIT_API		// need this so WmlMemoryPool.h will leave out dll stuff...
#include <WmlMemoryPool.h>

namespace rmsimplicit {


class RMSIMPLICIT_API VFTriangleMesh
{
public:
	typedef unsigned int VertexID;
	typedef unsigned int NormalID;
	typedef unsigned int TriangleID;
	static const VertexID INVALID_VERTEX;
	static const NormalID INVALID_NORMAL;
	static const TriangleID INVALID_TRIANGLE;

	VFTriangleMesh();
	virtual ~VFTriangleMesh();

	// add functions

	VertexID AddVertex( const Wml::Vector3f & vVertex );
	NormalID AddNormal( const Wml::Vector3f & vNormal );
	TriangleID AddTriangle( VertexID v1, VertexID v2, VertexID v3, 
							NormalID n1 = INVALID_NORMAL, NormalID n2 = INVALID_NORMAL, NormalID n3 = INVALID_NORMAL );

	// safe remove functions  ( will remove invalid related elements... )

	bool RemoveVertex( VertexID id );		// removes attached triangles and normals
	bool RemoveTriangle( TriangleID id );	// just removes triangle...

	// drastic remove functions
	void Clear();


	// bookkeeping functions
	unsigned int GetVertexCount();
	unsigned int GetNormalCount();
	unsigned int GetTriangleCount();

	unsigned int GetTriangleCount( VertexID id );


	// vertex and normal access functions

	Wml::Vector3f GetVertex( VertexID id );
	void GetVertex( VertexID id, Wml::Vector3f & vVertex );

	Wml::Vector3f GetNormal( NormalID id );
	void GetNormal( NormalID id, Wml::Vector3f & vNormal );
	NormalID GetAnyNormal( VertexID id, Wml::Vector3f & vNormal );

	// triangle vertex and normal access functions

	VertexID GetVertexID( TriangleID id, unsigned int nVertex );
	NormalID GetNormalID( TriangleID id, unsigned int nVertex );

	Wml::Vector3f GetVertex( TriangleID id, unsigned int nVertex );
	VertexID GetVertex( TriangleID id, unsigned int nVertex, Wml::Vector3f & vVertex );

	Wml::Vector3f GetNormal( TriangleID id, unsigned int nVertex );
	void GetNormal( TriangleID id, unsigned int nVertex, Wml::Vector3f & vNormal );

	VertexID GetVertex( TriangleID id, unsigned int nVertex, Wml::Vector3f & vVertex, Wml::Vector3f & vNormal );
	void GetTriangle( TriangleID id, Wml::Vector3f * pStore );
	void GetTriangle( TriangleID id, VertexID * pStore );
	void GetTriangleNormals( TriangleID id, Wml::Vector3f * pStore );
	void GetTriangleNormals( TriangleID id, NormalID * pStore );

	// v1 and v2 are indices in vertex list for triangle
	TriangleID GetEdgeNeighbour( TriangleID id, int v1, int v2 );

	// set functions

	void SetVertex( VertexID id, const Wml::Vector3f & vVertex );
	void SetNormal( NormalID id, const Wml::Vector3f & vNormal );


	// returns false if overflow would occur (does not resize vector!)
	bool GetUnordered1Ring( VertexID id, std::vector<VertexID> & vStore );

	// simple triangle iteration
	bool GetNextTriangle( VertexID id, TriangleID & lastTriangle );

	void DrawGL( bool bNormals );

	// conversion
	bool InitializeFromSimpleMesh( rmsimplicit::TriangleMesh * pMesh );

protected:


	class SmallTriangleList {
	public:
		SmallTriangleList() { m_vTriangles.reserve(8); m_vTriangles.push_back(VFTriangleMesh::INVALID_TRIANGLE);  }
		~SmallTriangleList() {}

		void insert( TriangleID id );
		bool find( TriangleID id );
		void erase( TriangleID id );
		unsigned int size() { return (unsigned int)m_vTriangles.size() - 1; }
		void clear() {m_vTriangles.resize(0); m_vTriangles.push_back(VFTriangleMesh::INVALID_TRIANGLE);}
		bool empty() { return m_vTriangles.size() == 1; }
		TriangleID begin() { return m_vTriangles[0]; }
		TriangleID get(int i) { return m_vTriangles[i]; }

	protected:
		std::vector< TriangleID > m_vTriangles;
	};


	struct VFVertex {
		Wml::Vector3f v;
		unsigned int nRefCount;

		SmallTriangleList vTriangles;
	};

	struct VFNormal {
		Wml::Vector3f n;
		unsigned int nRefCount;		// for knowing when to remove...
	};

	struct VFTriangle {
		VertexID vVertices[3];
		NormalID vNormals[3];

		// could have separate normals here!
	};

	// maps for item lookups
	std::map< VertexID, VFVertex *> m_vVertices;
	std::map< NormalID, VFNormal *> m_vNormals;
	std::map< TriangleID, VFTriangle *> m_vTriangles;

	// memory pools for allocating stuff
	Wml::MemoryPool< VFVertex > m_VertexPool;
	Wml::MemoryPool< VFNormal > m_NormalPool;
	Wml::MemoryPool< VFTriangle > m_TrianglePool;

	// internal remove functions that are not safe (but do check that it exists)
	bool DeleteVertex( VertexID id );
	bool DeleteNormal( NormalID id );
	bool DeleteTriangle( TriangleID id );

	// internal remove functions that do not check existence
	void DeleteTriangle( TriangleID id, VFTriangle * pTriangle );


	// internal cache for RemoveVertex
	std::vector< TriangleID > m_vTempTriangles;

public:

	class TriangleIterator {
	public:
		TriangleIterator( std::map<TriangleID, VFTriangle *>::iterator & start ) 
			{ m_cur = start; }
		~TriangleIterator() {}

		TriangleID operator*() 
			{ return (*m_cur).first; }
		void operator++() 
			{ m_cur++; }
		void operator++(int nPostfix) 
			{ ++m_cur; }
		bool operator==( const TriangleIterator & p2 ) const 
			{ return m_cur == p2.m_cur; }
		bool operator!=( const TriangleIterator & p2 ) const
			{ return m_cur != p2.m_cur; }		

	protected:
		std::map<TriangleID, VFTriangle *>::iterator m_cur;
	};

	TriangleIterator BeginTriangle() 
		{ return TriangleIterator( m_vTriangles.begin() ); }
	TriangleIterator EndTriangle() 
		{ return TriangleIterator( m_vTriangles.end() ); }



	class VertexIterator {
	public:
		VertexIterator( std::map<VertexID, VFVertex *>::iterator & start ) 
			{ m_cur = start; }
		~VertexIterator() {}

		VertexID operator*() 
			{ return (*m_cur).first; }
		void operator++() 
			{ m_cur++; }
		void operator++(int nPostfix) 
			{ ++m_cur; }
		bool operator==( const VertexIterator & p2 ) const 
			{ return m_cur == p2.m_cur; }
		bool operator!=( const VertexIterator & p2 ) const
			{ return m_cur != p2.m_cur; }		

	protected:
		std::map<VertexID, VFVertex *>::iterator m_cur;
	};

	VertexIterator BeginVertex() 
		{ return VertexIterator( m_vVertices.begin() ); }
	VertexIterator EndVertex() 
		{ return VertexIterator( m_vVertices.end() ); }
};



} // end namespace rmsimplicit


#endif   // __RMS_VF_TRIANGLE_MESH_H__

