#ifndef __RMRSUI_CONSTRAINT_H__
#define __RMRSUI_CONSTRAINT_H__

#include <WmlVector3.h>
#include <WmlRay3.h>
#include <WmlCircle3.h>
#include <WmlColors.h>
#include <Frame.h>

#include <WmlTriangle3.h>
#include <WmlSegment3.h>

#include "Geometry.h"
#include "GeometrySet.h"
#include "Tool.h"
#include "SceneObject.h"

namespace rmsui 
{


class ConstraintTarget : public SceneObjectListener
{
public:
	ConstraintTarget( SceneObject * pSO ) {
		m_pSceneObject = pSO;
		m_pSceneObject->AddListener(this);
	}
	virtual ~ConstraintTarget() {
		m_pSceneObject->RemoveListener(this);
	}

	SceneObject * SceneObj() const { return m_pSceneObject; }

	virtual void TranslateFrame( const Wml::Vector3f & vTranslate, bool bRelative = true )
		{ m_pSceneObject->Translate(vTranslate, bRelative); }
	virtual void RotateFrame( const Wml::Matrix3f & rotation ) 
		{ m_pSceneObject->Rotate(rotation); }
	virtual void Scale( float fScaleX, float fScaleY, float fScaleZ, bool bRelative, bool bUniform )
		{ m_pSceneObject->Scale(fScaleX, fScaleY, fScaleZ, bRelative, bUniform); }

	virtual Wml::Frame3f GetFrame() { return m_pSceneObject->GetFrame(); }

	virtual bool operator==(const ConstraintTarget & target2) const 
		{ return m_pSceneObject == target2.SceneObj(); }

	/* Event Listener */

	class Listener {
	public:
		virtual void OnTargetChanged(ConstraintTarget * pTarget) = 0;
	};
	void AddListener( Listener * pListener ) 
		{ m_vListeners.insert(pListener); }
	void RemoveListener( Listener * pListener ) 
		{ m_vListeners.erase(pListener); }
	virtual void OnSceneObjectChanged( SceneObject * pChanged, void * pChangeData )
		{ if ( pChanged == m_pSceneObject ) Notify(); }

protected:
	SceneObject * m_pSceneObject;

	std::set<Listener *> m_vListeners;
	void Notify() {
		std::set<Listener *>::iterator cur(m_vListeners.begin()), end(m_vListeners.end());
		while ( cur != end )  (*cur++)->OnTargetChanged(this);
	}
};





class Constraint : public ConstraintTarget::Listener
{
public:
	Constraint( ConstraintTarget * pTarget );
	~Constraint();

	enum Type {
		Unknown,
		AxisTranslate,
		AxisRotate,
		AxisScale,
		PlaneTranslate,
		UniformScale
	};
	virtual Type GetType() { return Unknown; }


	ConstraintTarget * Target() { return m_pTarget; }

	virtual void Render(const ViewState & eViewState) = 0;
	
	//! returns true on intersection, or if point was suitably close. nHitID tells us what part was hit (if applicable)
	virtual bool Intersect( const Wml::Ray3f & ray, Wml::Vector3f & ptHit, 
							float & fDistanceFromRay, float & fRayT, int & nHitID ) = 0;

	//! initialize constraint
	virtual void InitializeConstraint( const Wml::Ray3f & ray, const Wml::Frame3f & vInitialFrame ) = 0;

	//! returns translation to apply to frame origin
	virtual bool GetFrameTranslation( const Wml::Ray3f & ray, const Wml::Frame3f & vOldFrame,
									  Wml::Vector3f & vTranslate, bool & bRelative ) { return false; }

	//! returns rotation to apply to frame
	virtual bool GetFrameRotation( const Wml::Ray3f & ray, const Wml::Frame3f & vOldFrame,
								   Wml::Matrix3f & rotation ) { return false; }

	//! returns uniform scale to apply to target
	virtual bool GetUniformScaleDelta( const Wml::Ray3f & ray, const Wml::Frame3f & vOldFrame,
									   float & fScale ) { return false; }

	//! returns scale to apply to target
	virtual bool GetScaleDelta( const Wml::Ray3f & ray, const Wml::Frame3f & vOldFrame,
								Wml::Vector3f & vScale ) { return false; }


	//! crossing test
	virtual bool TestCross( const ViewState & eViewState, const Wml::Vector2f & v1, 
							const Wml::Vector2f & v2, Wml::Vector3f & vHit3D ) { return false; }

	//! called by DynamicTool3D to tell this constraint that it's target has changed (so update your frame!)
	//! [RMS TODO] can we handle this automatically? Callback from individual constraint is not sufficient...
	//virtual void OnTargetChanged() = 0;
	virtual void OnTargetChanged(ConstraintTarget * pTarget) = 0;

	const Wml::Frame3f & Frame() const { return m_vFrame; }

	//! somebody may want to overload this...
	virtual void SetFrame( const Wml::Frame3f & vFrame );

	const Wml::Vector3f & Origin() const { return m_vFrame.Origin(); }
	Wml::Vector3f Axis() const { return m_vFrame.Z(); }
	void UpdateFrame( const Wml::Vector3f & vNewOrigin, const Wml::Vector3f & vNewAxis );

	bool Pinned() const { return m_bPinned; }
	bool & Pinned() { return m_bPinned; }

	enum AxisFrameType {
		World, 
		Object, 
		Pivot,
		Surface, 
		Screen 
	};

	AxisFrameType & AxisType() { return m_eType; }

	const Wml::ColorRGBA & Color() const { return m_cColor; }
	Wml::ColorRGBA & Color() { return m_cColor; }

protected:
	ConstraintTarget * m_pTarget;

	Wml::Frame3f m_vFrame;
	Wml::Frame3f m_vRelativeFrame;
	void UpdateRelative( const Wml::Frame3f & vRelativeToFrame );
	void FromRelative( const Wml::Frame3f & vRelativeFromFrame );

	AxisFrameType m_eType;
	bool m_bPinned;

	Wml::ColorRGBA m_cColor;
};















class UniformScaleConstraint  : public Constraint
{
public:
	UniformScaleConstraint( ConstraintTarget * pTarget, const Wml::Vector3f & vOrigin );
	UniformScaleConstraint( ConstraintTarget * pTarget, const Wml::Frame3f & vFrame );

	virtual Type GetType() { return Constraint::UniformScale; }

	virtual void Render(const ViewState & eViewState);
	
	//! returns true on intersection, or if point was suitably close
	virtual bool Intersect( const Wml::Ray3f & ray, Wml::Vector3f & ptHit, 
							float & fDistanceFromRay, float & fRayT, int & nHitID );

	//! initialize constraint
	virtual void InitializeConstraint( const Wml::Ray3f & ray, const Wml::Frame3f & vInitialFrame );

	//! returns translation to apply to frame origin
	virtual bool GetUniformScaleDelta( const Wml::Ray3f & ray, const Wml::Frame3f & vOldFrame,
									   float & fScale );

	virtual bool TestCross( const ViewState & eViewState, const Wml::Vector2f & v1, 
							const Wml::Vector2f & v2, Wml::Vector3f & vHit3D );

	const Wml::Vector3f & Origin() { return m_vOrigin; }

	virtual void OnTargetChanged(ConstraintTarget * pTarget);

protected:
	float m_fDrawLength;
	float m_fDrawLineWidth;

	Wml::Vector3f m_vOrigin;
	Wml::Vector3f m_vPlaneNormal;

	Wml::Frame3f m_vPlaneFrame;
	bool m_bPlaneFrameSet;

	Wml::Vector3f m_vLastHitPoint;

	void GetHitPoint( const Wml::Ray3f & ray, Wml::Vector3f & ptHit, 
					  float & fDistanceFromRay, float & fRayT );

	GeometrySet m_geom;
	Wml::Frame3f m_geomFrame;
	Wml::Vector3f m_vLookDir;
		
	void UpdateGeometry( const ViewState & eViewState );
};








class AxisScaleConstraint  : public Constraint
{
public:
	AxisScaleConstraint( ConstraintTarget * pTarget, const Wml::Vector3f & vOrigin, const Wml::Vector3f & vDirection );

	virtual Type GetType() { return Constraint::AxisScale; }

	virtual void Render(const ViewState & eViewState);
	
	//! returns true on intersection, or if point was suitably close
	virtual bool Intersect( const Wml::Ray3f & ray, Wml::Vector3f & ptHit, 
							float & fDistanceFromRay, float & fRayT, int & nHitID );

	//! initialize constraint
	virtual void InitializeConstraint( const Wml::Ray3f & ray, const Wml::Frame3f & vInitialFrame );

	//! returns translation to apply to frame origin
	virtual bool GetScaleDelta( const Wml::Ray3f & ray, const Wml::Frame3f & vOldFrame,
								Wml::Vector3f & vScale );

	virtual bool TestCross( const ViewState & eViewState, const Wml::Vector2f & v1, 
							const Wml::Vector2f & v2, Wml::Vector3f & vHit3D );

	const Wml::Vector3f & Origin() { return m_vOrigin; }
	const Wml::Vector3f & Direction() { return m_vDirection; }

	virtual void OnTargetChanged(ConstraintTarget * pTarget);

protected:
	float m_fDrawLength;
	float m_fDrawLineWidth;

	Wml::Vector3f m_vOrigin;
	Wml::Vector3f m_vDirection;
	Wml::Frame3f m_vFrame;
	Wml::Vector3f m_vOriginRelative;
	Wml::Vector3f m_vDirectionRelative;

	Wml::Vector3f m_vLastHitPoint;

	void GetNearestPoint( const Wml::Ray3f & ray, Wml::Vector3f & ptHit, 
							float & fDistanceFromRay, float & fRayT );

	GeometrySet m_geom;

	Wml::Frame3f m_geomFrame;
	float m_fVisibility;
	Wml::Segment3f m_vCrossingSegment;
		
	void UpdateGeometry( const ViewState & eViewState );
};





class ViewStateProjector : public GeometrySet::Projector
{
public:
	const ViewState * pViewState;
	virtual Wml::Vector2f Get2D(const Wml::Vector3f & v3D) {
		return pViewState->GLProject(v3D);
	}
};




}   // end namespace rmsui

#endif // __RMRSUI_CONSTRAINT_H__