#ifndef __RMRSUI_CONSTRAINT_SET_H__
#define __RMRSUI_CONSTRAINT_SET_H__

#include <WmlVector3.h>
#include <WmlRay3.h>
#include <WmlCircle3.h>
#include <WmlColors.h>
#include <Frame.h>

#include <WmlTriangle3.h>

#include "Geometry.h"
#include "Tool.h"
#include "SceneObject.h"
#include "Constraint.h"
#include "TranslateConstraints.h"
#include "RotateConstraints.h"
#include "PivotSO.h"

namespace rmsui 
{

class ConstraintAxis
{
public:
	Constraint::AxisFrameType eType;
	Wml::Vector3f vOrigin;
	Wml::Vector3f vDirection;

	bool m_bFrameSpecified;
	Wml::Frame3f vFrame;

	//! only used for history
	Constraint::Type m_eRestoreType;
	bool m_bPinned;
	Wml::ColorRGBA m_vColor;
	float m_fRadius;

	ConstraintAxis() {
		m_eRestoreType = Constraint::Unknown;
		m_bPinned = false;
	}
	ConstraintAxis(Constraint::AxisFrameType type, Wml::Vector3f origin, Wml::Vector3f direction) { 
		m_eRestoreType = Constraint::Unknown;
		m_bPinned = false;
		eType = type; 
		vOrigin = origin; 
		vDirection = direction; 
		m_bFrameSpecified = false; 
	}

	void SetFrame( const Wml::Frame3f & frame, bool bAuto = true ) { 
		m_eRestoreType = Constraint::Unknown;
		m_bPinned = false;
		vFrame = frame; 
		m_bFrameSpecified = true; 
	}

};






class ConstraintSet 
{
public:
	ConstraintSet();
	~ConstraintSet();

	void Render(const ViewState & viewState);
	Constraint * Intersect( const Wml::Ray3f & ray, Wml::Vector3f & ptPosition, int & nHitID );
	Constraint * FindBestCross( const ViewState & eViewState, const Wml::Vector2f & v1, const Wml::Vector2f & v2, Wml::Vector3f & vHit3D );

	VectorConstraint * MakeVectorConstraint( ConstraintTarget * pTarget, const ConstraintAxis & axis, const Wml::ColorRGBA & cColor );
	PlaneTranslateConstraint * MakePlaneTranslateConstraint( ConstraintTarget * pTarget, const ConstraintAxis & axis, const Wml::ColorRGBA & cColor );
	AxisRotationConstraint * MakeAxisRotationConstraint( ConstraintTarget * pTarget, const ConstraintAxis & axis, float fRadius, const Wml::ColorRGBA & cColor );
	UniformScaleConstraint * MakeUniformScaleConstraint( ConstraintTarget * pTarget, const ConstraintAxis & axis, const Wml::ColorRGBA & cColor );
	AxisScaleConstraint * MakeAxisScaleConstraint( ConstraintTarget * pTarget, const ConstraintAxis & axis, const Wml::ColorRGBA & cColor );

	void Clear( bool bClearPinned = true, bool bDeleteConstraints = true );

	const std::set< UniformScaleConstraint *> & GetUniformScaleConstraints() { return m_vUniformScaleConstraints; }

	void RemoveConstraint( Constraint * pConstraint );

	void FindConstraintsByTarget( ConstraintTarget * pTarget, std::vector<Constraint *> & vConstraints );

	VectorConstraint * FindVectorConstraint( const Wml::Vector3f & vOrigin, const Wml::Vector3f & vDirection, bool bPinnedOnly );
	AxisRotationConstraint * FindAxisRotationConstraint( const Wml::Vector3f & vOrigin, const Wml::Vector3f & vDirection, bool bPinnedOnly );
	PlaneTranslateConstraint * FindPlaneTranslateConstraint( const Wml::Vector3f & vOrigin, const Wml::Vector3f & vDirection, bool bPinnedOnly );
	UniformScaleConstraint * FindUniformScaleConstraint( ConstraintTarget * pTarget );

	bool FindConstraints( const Wml::Vector3f & vOrigin, const Wml::Vector3f & vDirection, bool bSkipPinned, std::vector<Constraint *> & vFound );

	void TogglePinConstraint( Constraint * pCons );
	bool IsPinned( Constraint * pCons );

	void AppendHistoryCheckpoint();
	unsigned int GetHistoryDepth() { return (unsigned int)m_vHistory.size(); }
	bool SetHistoryState( unsigned int nState, ConstraintTarget * pTarget );



	const ViewState & CachedViewState( ) const { return m_viewStateCache; }

protected:

	ViewState m_viewStateCache;

	struct ConstraintInfo {
		Constraint * pConstraint;

		bool operator<(const ConstraintInfo & info2) const {
			return pConstraint < info2.pConstraint;
		}
	};


	typedef std::set<ConstraintInfo> AnyConstraintSet;
	typedef std::set<VectorConstraint *> VectorConstraintSet;
	typedef std::set<PlaneTranslateConstraint *> PlaneTranslateConstraintSet;
	typedef std::set<AxisRotationConstraint *> AxisRotationConstraintSet;
	typedef std::set<UniformScaleConstraint *> UniformScaleConstraintSet;
	typedef std::set<AxisScaleConstraint *> AxisScaleConstraintSet;

	AnyConstraintSet m_vConstraints;
	VectorConstraintSet m_vVectorConstraints;
	PlaneTranslateConstraintSet m_vPlaneTranslateConstraints;
	AxisRotationConstraintSet m_vAxisRotationConstraints;
	UniformScaleConstraintSet m_vUniformScaleConstraints;
	AxisScaleConstraintSet m_vAxisScaleConstraints;

	void AddConstraint( VectorConstraint * pVec );
	void AddConstraint( PlaneTranslateConstraint * pPlaneTrans );
	void AddConstraint( AxisRotationConstraint * pAxisRot );
	void AddConstraint( UniformScaleConstraint * pUniScale );
	void AddConstraint( AxisScaleConstraint * pAxisScale );


	struct HistoryEvent {
		std::vector<ConstraintAxis> vConstraints;
	};
	std::vector<HistoryEvent> m_vHistory;
};






class PivotSet
{
public:
	PivotSet() {}
	~PivotSet() {}

	void AddPivot( PivotSO * pPivot );

	void DeletePivot( PivotSO * pPivot );
	void ClearPivots();

	size_t Count() { return m_vPivots.size(); }
	bool IsEmpty() { return m_vPivots.empty(); }
	const std::vector<PivotSO *> & Pivots() { return m_vPivots; }

	void Render(const ViewState & viewState);

	PivotSO * Intersect( const Wml::Ray3f & ray, Wml::Vector3f & ptPosition, float & fDepth );
	void ClearTargets();

protected:
	std::vector<PivotSO *> m_vPivots;
};




}   // end namespace rmsui

#endif // __RMRSUI_CONSTRAINT_SET_H__