// Ryan Schmidt - rms@unknownroad.com
// Copyright (c) 2005-2010
// Distributed under the Boost Software License, Version 1.0.
// http://www.boost.org/LICENSE_1_0.txt


#ifndef RMSUI_GEOMETRY_H
#define RMSUI_GEOMETRY_H

#include <WmlVector3.h>
#include <WmlMatrix3.h>
#include <WmlSphere3.h>
#include <WmlCircle3.h>
#include <WmlCone3.h>
#include <WmlRay3.h>
#include <WmlLine3.h>
#include <WmlColorRGB.h>
#include <WmlColors.h>

#include <map>


namespace rmsui
{

class IShape
{
public:
	virtual bool Intersect( const Wml::Ray3f & ray, 
							Wml::Vector3f & ptPosition, float & fDepth ) = 0;

	virtual void Render( const Wml::ColorRGBA & color ) = 0;
};





class Sphere : public IShape
{
public:
	Sphere( const Wml::Vector3f & ptCenter = Wml::Vector3f::ZERO,
			float fRadius = 1.0 )
	{
		m_sphere.Center() = ptCenter;
		m_sphere.Radius() = fRadius;
		m_sphereDisplayList = 0;
	}

	virtual ~Sphere() {}

	Wml::Sphere3f & Geometry()
		{ return m_sphere; }

	virtual bool Intersect( const Wml::Ray3f & ray, 
							Wml::Vector3f & ptPosition, float & fDepth );

	virtual void Render( const Wml::ColorRGBA & color = Wml::ColorRGBA::RED );

	void ClosestPoint( const Wml::Line3f & line, Wml::Vector3f & ptPosition );
	

protected:
	Wml::Sphere3f m_sphere;

	//! OpenGL display list
	unsigned int m_sphereDisplayList;
};




class Circle : public IShape
{
public:
	Circle( const Wml::Vector3f & ptCenter = Wml::Vector3f::ZERO,
			const Wml::Vector3f & vNormal = Wml::Vector3f::UNIT_Z,
			float fRadius = 1.0,
			float fLineWidth = 2.0f)
	{
		InitializeCircle( ptCenter, vNormal, fRadius);
		m_circleDisplayList = 0;
		m_fLineWidth = fLineWidth;
	}

	virtual ~Circle() {}

	Wml::Circle3f & Geometry()
		{ return m_circle; }

	virtual bool Intersect( const Wml::Ray3f & ray, 
							Wml::Vector3f & ptPosition, float & fDepth );

	virtual void Render( const Wml::ColorRGBA & color = Wml::ColorRGBA::RED );

protected:
	Wml::Circle3f m_circle;
	float m_glRotationMatrix[4][4];
	float m_fLineWidth;

	//! OpenGL display list
	unsigned int m_circleDisplayList;

	const static float s_IntersectionTolerance;

	//! initialize Wml circle from center point, normal, and radius
	void InitializeCircle( const Wml::Vector3f & ptCenter,
						   const Wml::Vector3f & vNormal, float fRadius );
};




class Chevron : public IShape
{
public:
	Chevron( const Wml::Vector3f & ptCenter = Wml::Vector3f::ZERO,
			 const Wml::Vector3f & vNormal = Wml::Vector3f::UNIT_Z,
			 float fAngle = 0.0f,
			 float fScale = 1.0f,
			 float fLineWidth = 2.0f)
	{
		InitializeChevron( ptCenter, vNormal, fAngle, fScale );
		m_chevronDisplayList = 0;
		m_fLineWidth = fLineWidth;
	}

	virtual ~Chevron() {}

	Wml::Circle3f & PlaneGeometry()
		{ return m_circle; }

	virtual bool Intersect( const Wml::Ray3f & ray, 
							Wml::Vector3f & ptPosition, float & fDepth );

	virtual void Render( const Wml::ColorRGBA & color = Wml::ColorRGBA::RED );

protected:
	Wml::Circle3f m_circle;
	float m_fAngle;
	float m_glRotationMatrix[4][4];
	float m_fLineWidth;

	//! OpenGL display list
	unsigned int m_chevronDisplayList;

	const static float s_IntersectionTolerance;

	//! initialize Wml circle from center point, normal, and radius
	void InitializeChevron( const Wml::Vector3f & ptCenter,
						    const Wml::Vector3f & vNormal, float fAngle, float fScale );
};




class Cone
{
public:
	Cone( const Wml::Vector3f & ptVertex = Wml::Vector3f::ZERO,
		  const Wml::Vector3f & vAxis = Wml::Vector3f::UNIT_Z)
	{
		m_cone.Axis() = vAxis;
		m_cone.Axis().Normalize();

		m_cone.Vertex() = ptVertex + (m_cone.Axis() * s_fConeHeight);
	
		// cone angle is hardcoded to 45 degrees here
		float fRadians = 45.0f * Wml::Mathf::DEG_TO_RAD;
		m_cone.CosAngle() = Wml::Mathf::Cos(fRadians);
		m_cone.SinAngle() = Wml::Mathf::Sin(fRadians);
	}

	Wml::Vector3f BasePoint()
	{
		return m_cone.Vertex() - (m_cone.Axis() * s_fConeHeight);
	}

	float Height()
	{
		return s_fConeHeight;
	}

	float Width()
	{
		return s_fConeWidth;
	}

	void Render();

	Wml::Cone3f & Geometry()
	{
		return m_cone;
	}

protected:
	Wml::Cone3f m_cone;
	float m_fWidth;
	float m_fHeight;

	static unsigned int GetConeDisplayList();

	// shared members
	static float s_fConeWidth;
	static float s_fConeHeight;
	static unsigned int s_coneResolution;
	static unsigned int s_coneDisplayList;
};


}	// end namespace rmsui



#endif  // RMSUI_GEOMETRY_H