#include "StdAfx.h"
#include "GeometrySet.h"
#include "UIGlobals.h"
#include "rmsdebug.h"

#include <GL/gl.h>
#include <GL/glu.h>
#include <limits>

#include <WmlIntrLin3Tri3.h>
#include <WmlIntrLin2Lin2.h>


using namespace rmsui;


GeometrySet::GeometrySet()
{
	m_vFrame = Wml::Frame();
	m_fRenderAlpha = 1.0f;
	m_cCurrentColor = Wml::ColorRGBA::RED;
	m_cRenderGeomColor = Wml::ColorRGBA::WHITE;
	m_cRenderLineColor = Wml::ColorRGBA::BLACK;

	m_nCurrentTag = 0;
}

GeometrySet::~GeometrySet()
{
}

void GeometrySet::Clear(bool bClearFrame)
{
	m_vTriangles.resize(0);
	m_vQuads.resize(0);
	m_vLines.resize(0);
	m_vLoops.resize(0);
	m_vLoopVerts.resize(0);
	m_vTriFans.resize(0);
	m_vTriFanVerts.resize(0);
	m_vQuadStrips.resize(0);
	m_vQuadStripElems.resize(0);
	if (bClearFrame)
		m_vFrame = Wml::Frame();
	m_nCurrentTag = 0;
}



void GeometrySet::AddTriangle( const Wml::Vector3f & v1, const Wml::Vector3f & v2, const Wml::Vector3f & v3, const Wml::ColorRGBA & cColor )
{
	m_vTriangles.resize( m_vTriangles.size() + 1 );
	Tri & t = m_vTriangles.back();
	t.vVerts[0] = m_vFrame.Origin() + m_vFrame.FrameMatrix() * v1;
	t.vVerts[1] = m_vFrame.Origin() + m_vFrame.FrameMatrix() * v2;   
	t.vVerts[2] = m_vFrame.Origin() + m_vFrame.FrameMatrix() * v3;
	t.cColor = cColor;
	t.tag = m_nCurrentTag;
}
void GeometrySet::AddQuad( const Wml::Vector3f & v1, const Wml::Vector3f & v2, const Wml::Vector3f & v3, const Wml::Vector3f & v4, const Wml::ColorRGBA & cColor )
{
	m_vQuads.resize( m_vQuads.size() + 1 );
	Quad & q = m_vQuads.back();
	q.vVerts[0] = m_vFrame.Origin() + m_vFrame.FrameMatrix() * v1;	
	q.vVerts[1] = m_vFrame.Origin() + m_vFrame.FrameMatrix() * v2;   
	q.vVerts[2] = m_vFrame.Origin() + m_vFrame.FrameMatrix() * v3;   
	q.vVerts[3] = m_vFrame.Origin() + m_vFrame.FrameMatrix() * v4;
	q.cColor = cColor;
	q.tag = m_nCurrentTag;
}

void GeometrySet::AddQuad( const Wml::AxisAlignedBox2f & box, const Wml::ColorRGBA & cColor )
{
	AddQuad(
		Wml::Vector3f(box.GetXMin(), box.GetYMin(), 0),
		Wml::Vector3f(box.GetXMax(), box.GetYMin(), 0),
		Wml::Vector3f(box.GetXMax(), box.GetYMax(), 0),
		Wml::Vector3f(box.GetXMin(), box.GetYMax(), 0), cColor );
}


void GeometrySet::BeginTriFan(const Wml::ColorRGBA & cColor)
{
	m_vTriFans.resize(m_vTriFans.size() + 1);
	m_vTriFans.back().cColor = cColor;
	m_vTriFans.back().nStart = (int)m_vTriFanVerts.size();
	m_vTriFans.back().nCount = 0;
	m_vTriFans.back().tag = m_nCurrentTag;
}
void GeometrySet::AppendTriFanVert( const Wml::Vector3f & v1 )
{
	m_vTriFanVerts.resize( m_vTriFanVerts.size() + 1 );
	m_vTriFanVerts.back() = m_vFrame.Origin() + m_vFrame.FrameMatrix() * v1;
	m_vTriFans.back().nCount++;
}


void GeometrySet::BeginQuadStrip(const Wml::ColorRGBA & cColor, bool bCullBackFaces)
{
	m_vQuadStrips.resize(m_vQuadStrips.size() + 1);
	m_vQuadStrips.back().cColor = cColor;
	m_vQuadStrips.back().nStart = (int)m_vQuadStripElems.size();
	m_vQuadStrips.back().nCount = 0;
	m_vQuadStrips.back().bCullBackFaces = bCullBackFaces;
	m_vQuadStrips.back().tag = m_nCurrentTag;
}
void GeometrySet::AddQuadStripElem( const Wml::Vector3f & v1, const Wml::Vector3f & v2 )
{
	m_vQuadStripElems.resize( m_vQuadStripElems.size() + 1 );
	m_vQuadStripElems.back().vVerts[0] = m_vFrame.Origin() + m_vFrame.FrameMatrix() * v1;
	m_vQuadStripElems.back().vVerts[1] = m_vFrame.Origin() + m_vFrame.FrameMatrix() * v2;
	m_vQuadStrips.back().nCount++;
}
void GeometrySet::QuadStripToLoop()
{
	QuadStrip & cur = m_vQuadStrips.back();
	m_vQuadStripElems.push_back( m_vQuadStripElems[cur.nStart] );
	cur.nCount++;
}
void GeometrySet::AppendQuadStripBorderLoops(const Wml::ColorRGBA & cColor)
{
	QuadStrip & strip = m_vQuadStrips.back();

	// store frame
	Wml::Frame3f vFrame = Frame();
	Frame() = Wml::Frame3f();

	BeginLoop(cColor);
	for ( int i = 0; i < strip.nCount-1; ++i ) {
		QuadStripElem & elem = m_vQuadStripElems[strip.nStart+i];
		AppendLoopVert( elem.vVerts[0] );
	}
	BeginLoop(cColor);
	for ( int i = 0; i < strip.nCount-1; ++i ) {
		QuadStripElem & elem = m_vQuadStripElems[strip.nStart+i];
		AppendLoopVert( elem.vVerts[1] );
	}

	Frame() = vFrame;
}
void GeometrySet::AppendQuadStripInteriorEdges(const Wml::ColorRGBA & cColor, int nSkip )
{
	QuadStrip & strip = m_vQuadStrips.back();

	// store frame
	Wml::Frame3f vFrame = Frame();
	Frame() = Wml::Frame3f();

	if ( nSkip == 0 )
		nSkip = 1;
	for ( int i = 0; i < strip.nCount; i += nSkip ) {
		QuadStripElem & elem = m_vQuadStripElems[strip.nStart+i];
		AddLine( elem.vVerts[0], elem.vVerts[1], cColor );
	}

	Frame() = vFrame;
}




void GeometrySet::AddSphere( const Wml::ColorRGBA & cColor, float fRadius, int nSlices, int nStacks )
{
	float fYDelta = Wml::Mathf::PI / (int)nStacks;
	float fXDelta = Wml::Mathf::TWO_PI / (int)nSlices;

	for ( int yi = 1; yi < nStacks-1; ++yi ) {
		float fYAngle1 = (float)yi * fYDelta;
		float fYAngle2 = (float)(yi+1) * fYDelta;
		float fSinY1 = sin(fYAngle1);   float fSinY2 = sin(fYAngle2);
		float fCosY1 = cos(fYAngle1);   float fCosY2 = cos(fYAngle2);

		BeginQuadStrip( cColor );
		for ( int i = 0; i < nSlices; ++i ) {
			float fAngle = (float)i * fXDelta;
			Wml::Vector3f vVert1( 
				fRadius * fSinY1 * cos(fAngle), fRadius * fSinY1 * sin(fAngle), fRadius * fCosY1 );
			Wml::Vector3f vVert2( 
				fRadius * fSinY2 * cos(fAngle), fRadius * fSinY2 * sin(fAngle), fRadius * fCosY2 );
			AddQuadStripElem( vVert1, vVert2 );
		}
		QuadStripToLoop();
	}

	// do endcaps
	BeginTriFan(cColor);
	AppendTriFanVert( Wml::Vector3f(0,fRadius,0) );
	float fYAngle = (float)1 * fYDelta;
	float fSinY = sin(fYAngle);   
	float fCosY = cos(fYAngle);
	for ( int i = 0; i < nSlices; ++i ) {
		float fAngle = (float)i * fXDelta;
		AppendTriFanVert( Wml::Vector3f( 
			fRadius * fSinY * cos(fAngle), fRadius * fSinY * sin(fAngle), fRadius * fCosY ) );
	}
	BeginTriFan(cColor);
	AppendTriFanVert( Wml::Vector3f(0,-fRadius,0) );
	fYAngle = (float)(nStacks-1) * fYDelta;
	fSinY = sin(fYAngle);   
	fCosY = cos(fYAngle);
	for ( int i = 0; i < nSlices; ++i ) {
		float fAngle = (float)i * fXDelta;
		AppendTriFanVert( Wml::Vector3f( 
			fRadius * fSinY * cos(fAngle), fRadius * fSinY * sin(fAngle), fRadius * fCosY ) );
	}

}



void GeometrySet::AddCube( const Wml::ColorRGBA & cFaceColor, float fEdgeLength, bool bEdges, const Wml::ColorRGBA & cEdgeColor )
{
	float fR = fEdgeLength / 2;
	Wml::Vector3f v1(-fR, -fR, fR);
	Wml::Vector3f v2(fR, -fR, fR);
	Wml::Vector3f v3(fR, fR, fR);
	Wml::Vector3f v4(-fR, fR, fR);

	Wml::Vector3f v5(-fR, -fR, -fR);
	Wml::Vector3f v6(fR, -fR, -fR);
	Wml::Vector3f v7(fR, fR, -fR);
	Wml::Vector3f v8(-fR, fR, -fR);

	AddQuad(v1, v2, v3, v4, cFaceColor);
	AddQuad(v5, v6, v7, v8, cFaceColor);

	AddQuad(v1, v2, v6, v5, cFaceColor);
	AddQuad(v3, v4, v8, v7, cFaceColor);

	AddQuad(v2, v3, v7, v6, cFaceColor);
	AddQuad(v4, v1, v5, v8, cFaceColor);

	if ( bEdges ) {
		BeginLoop(cEdgeColor);
		AppendLoopVert(v1);   AppendLoopVert(v2);  AppendLoopVert(v3);  AppendLoopVert(v4);
		BeginLoop(cEdgeColor);
		AppendLoopVert(v5);   AppendLoopVert(v6);  AppendLoopVert(v7);  AppendLoopVert(v8);
		AddLine(v1, v5, cEdgeColor);
		AddLine(v2, v6, cEdgeColor);
		AddLine(v3, v7, cEdgeColor);
		AddLine(v4, v8, cEdgeColor);
	}

}




void GeometrySet::AddLine( const Wml::Vector3f & v1, const Wml::Vector3f & v2, const Wml::ColorRGBA & cColor )
{
	m_vLines.resize( m_vLines.size() + 1 );
	Line & t = m_vLines.back();
	t.vVerts[0] = m_vFrame.Origin() + m_vFrame.FrameMatrix() * v1;
	t.vVerts[1] = m_vFrame.Origin() + m_vFrame.FrameMatrix() * v2;   
	t.cColor = cColor;
	t.tag = m_nCurrentTag;
}

void GeometrySet::BeginLoop(const Wml::ColorRGBA & cColor)
{
	m_vLoops.resize(m_vLoops.size() + 1);
	m_vLoops.back().cColor = cColor;
	m_vLoops.back().nStart = (int)m_vLoopVerts.size();
	m_vLoops.back().nCount = 0;
	m_vLoops.back().tag = m_nCurrentTag;
}

void GeometrySet::AppendLoopVert( const Wml::Vector3f & vVert )
{
	m_vLoopVerts.push_back( m_vFrame.Origin() + m_vFrame.FrameMatrix() * vVert );
	m_vLoops.back().nCount++;
}




void GeometrySet::AddCircle( const Wml::ColorRGBA & cColor, const Wml::Frame3f & vFrame, float fRadius, int nSteps )
{
	float fDelta = Wml::Mathf::TWO_PI / (float)(nSteps);
	BeginLoop(cColor);
	for ( int i = 0; i < nSteps; ++i ) {
		float fAngle = (float)i * fDelta;
		Wml::Vector3f vVert( 
			vFrame.Origin() + 
			(fRadius * cos(fAngle)) * vFrame.X() + 
			(fRadius * sin(fAngle)) * vFrame.Y() );
		AppendLoopVert( vVert );
	}
}


void GeometrySet::AddDisc( const Wml::ColorRGBA & cColor, const Wml::Frame3f & vFrame, float fRadius, int nSteps )
{
	float fDelta = Wml::Mathf::TWO_PI / (float)(nSteps);
	BeginTriFan(cColor);
	AppendTriFanVert(vFrame.Origin());
	for ( int i = 0; i < nSteps+1; ++i ) {
		float fAngle = (float)i * fDelta;
		Wml::Vector3f vVert( 
			vFrame.Origin() + 
			(fRadius * cos(fAngle)) * vFrame.X() + 
			(fRadius * sin(fAngle)) * vFrame.Y() );
		AppendTriFanVert( vVert );
	}
}



void GeometrySet::Render( bool bUseGeomColors, bool bUseLineColors )
{
	size_t nCount;

	glPushAttrib(GL_ENABLE_BIT);

	glDepthFunc(GL_LEQUAL);

	glColor4f( m_cRenderLineColor.R(), m_cRenderLineColor.G(), m_cRenderLineColor.B(), m_cRenderLineColor.A() * m_fRenderAlpha );

	nCount = m_vLoops.size();
	for ( unsigned int i = 0; i < nCount; ++i ) {
		LineLoop & loop = m_vLoops[i];
		if ( bUseLineColors ) 
			glColor4f( loop.cColor.R(), loop.cColor.G(), loop.cColor.B(), loop.cColor.A() * m_fRenderAlpha );
		glBegin(GL_LINE_LOOP);
		for ( int j = 0; j < loop.nCount; ++j ) {
			glVertex3fv( m_vLoopVerts[ loop.nStart + j ] );
		}
		glEnd();
	}

	nCount = m_vLines.size();
	glBegin(GL_LINES);
	for ( unsigned int i = 0; i < nCount; ++i ) {
		Line & q = m_vLines[i];
		if ( bUseLineColors ) 
			glColor4f( q.cColor.R(), q.cColor.G(), q.cColor.B(), q.cColor.A() * m_fRenderAlpha );
		glVertex3fv(q.vVerts[0]);
		glVertex3fv(q.vVerts[1]);
	}
	glEnd();


	glDepthFunc(GL_LESS);

	// now draw polys w/ polygon offset

	glColor4f( m_cRenderGeomColor.R(), m_cRenderGeomColor.G(), m_cRenderGeomColor.B(), m_cRenderGeomColor.A() * m_fRenderAlpha );


	glEnable(GL_POLYGON_OFFSET_FILL);
	glPolygonOffset(1.0f, 1.0f);

	nCount = m_vTriangles.size();
	glBegin(GL_TRIANGLES);
	for ( unsigned int i = 0; i < nCount; ++i ) {
		Tri & t = m_vTriangles[i];
		if ( bUseGeomColors )
			glColor4f( t.cColor.R(), t.cColor.G(), t.cColor.B(), t.cColor.A() * m_fRenderAlpha );
		for ( int j = 0; j < 3; ++j )
			glVertex3fv( t.vVerts[j] );
	}
	glEnd();

	nCount = m_vQuads.size();
	glBegin(GL_TRIANGLES);
	for ( unsigned int i = 0; i < nCount; ++i ) {
		Quad & q = m_vQuads[i];
		if ( bUseGeomColors ) 
			glColor4f( q.cColor.R(), q.cColor.G(), q.cColor.B(), q.cColor.A() * m_fRenderAlpha );
		for ( int j = 0; j < 3; ++j )
			glVertex3fv( q.vVerts[j] );
		for ( int j = 0; j < 3; ++j )
			glVertex3fv( q.vVerts[ (j==0) ? j : j+1] );
	}
	glEnd();


	nCount = m_vTriFans.size();
	for ( unsigned int i = 0; i < nCount; ++i ) {
		TriFan & fan = m_vTriFans[i];
		if ( bUseGeomColors ) 
			glColor4f( fan.cColor.R(), fan.cColor.G(), fan.cColor.B(), fan.cColor.A() * m_fRenderAlpha );
		glBegin(GL_TRIANGLE_FAN);
		for ( int j = 0; j < fan.nCount; ++j ) {
			glVertex3fv( m_vTriFanVerts[ fan.nStart + j ] );
		}
		glEnd();
	}


	glPushAttrib(GL_ENABLE_BIT);

	nCount = m_vQuadStrips.size();
	for ( unsigned int i = 0; i < nCount; ++i ) {
		QuadStrip & strip = m_vQuadStrips[i];
		if ( bUseGeomColors ) 
			glColor4f( strip.cColor.R(), strip.cColor.G(), strip.cColor.B(), strip.cColor.A() * m_fRenderAlpha );

		if ( strip.bCullBackFaces ) {
			glEnable(GL_CULL_FACE);
			glCullFace(GL_BACK);
		} else
			glDisable(GL_CULL_FACE);

		glBegin(GL_TRIANGLE_STRIP);
		for ( int j = 0; j < strip.nCount; ++j ) {
			QuadStripElem & q = m_vQuadStripElems[ strip.nStart + j ];
			glVertex3fv(q.vVerts[0]);
			glVertex3fv(q.vVerts[1]);
		}
		glEnd();
	}

	glPopAttrib();


	glDisable(GL_POLYGON_OFFSET_FILL);



	glPopAttrib();
}


bool GeometrySet::Intersect( const Wml::Ray3f & ray, Wml::Vector3f & ptHit, GeomTag & nHitTag )
{
	Wml::Triangle3f tri;

	size_t nCount = m_vTriangles.size();
	for ( unsigned int i = 0; i < nCount; ++i ) {
		Tri & t = m_vTriangles[i];
		
		tri.Origin() = t.vVerts[0];
		tri.Edge0() = t.vVerts[1] - tri.Origin(); 
		tri.Edge1() = t.vVerts[2] - tri.Origin();
		if ( Wml::FindIntersection(ray, tri, ptHit) ) {
			nHitTag = t.tag;
			return true;
		}
	}

	nCount = m_vQuads.size();
	for ( unsigned int i = 0; i < nCount; ++i ) {
		Quad & q = m_vQuads[i];

		tri.Origin() = q.vVerts[0];
		tri.Edge0() = q.vVerts[1] - tri.Origin(); 
		tri.Edge1() = q.vVerts[2] - tri.Origin();
		if ( Wml::FindIntersection(ray, tri, ptHit) ) {
			nHitTag = q.tag;
			return true;
		}

		tri.Edge0() = q.vVerts[3] - tri.Origin();
		if ( Wml::FindIntersection(ray, tri, ptHit) ) {
			nHitTag = q.tag;
			return true;
		}
	}


	nCount = m_vQuadStrips.size();
	for ( unsigned int i = 0; i < nCount; ++i ) {
		QuadStrip & strip = m_vQuadStrips[i];
		for ( int j = 0; j < strip.nCount-1; ++j ) {
			QuadStripElem & q1 = m_vQuadStripElems[strip.nStart + j];
			QuadStripElem & q2 = m_vQuadStripElems[strip.nStart + j + 1];

			tri.Origin() = q1.vVerts[0];
			tri.Edge0() = q1.vVerts[1] - tri.Origin(); 
			tri.Edge1() = q2.vVerts[1] - tri.Origin();

			bool bHit = Wml::FindIntersection(ray, tri, ptHit);
			if ( ! bHit ) {
				tri.Edge0() = tri.Edge1();
				tri.Edge1() = q2.vVerts[0] - tri.Origin();
				bHit = Wml::FindIntersection(ray, tri, ptHit);
			}

			// [RMS HACK] cull back-facing quads
			if ( bHit ) {
				Wml::Vector3f vNormal = tri.Edge0().Cross(tri.Edge1());
				vNormal.Normalize();
				if ( ray.Direction().Dot(vNormal) > 0 )
					bHit = false;
			}
			if ( bHit ) {
				nHitTag = strip.tag;
				return true;
			}
		}
	}


	nCount = m_vTriFans.size();
	for ( unsigned int i = 0; i < nCount; ++i ) {
		TriFan & fan = m_vTriFans[i];
		for ( int j = 0; j < fan.nCount-1; ++j ) {
			Wml::Vector3f & v1 = m_vTriFanVerts[fan.nStart ];
			Wml::Vector3f & v2 = m_vTriFanVerts[fan.nStart + j ];
			Wml::Vector3f & v3 = m_vTriFanVerts[fan.nStart + j + 1 ];

			tri.Origin() = v1;
			tri.Edge0() = v2 - tri.Origin(); 
			tri.Edge1() = v3 - tri.Origin();
			bool bHit = Wml::FindIntersection(ray, tri, ptHit);
			if ( bHit ) {
				nHitTag = fan.tag;
				return true;
			}
		}
	}

	return false;

}


bool GeometrySet::TestLinesCross( const Wml::Segment2f & seg, Projector & proj, Wml::Vector3f & vHit3D )
{
	size_t nCount;

	int nQuantity;
	float fT[2];

	nCount = m_vLoops.size();
	for ( unsigned int i = 0; i < nCount; ++i ) {
		LineLoop & loop = m_vLoops[i];
		for ( int j = 0; j < loop.nCount; ++j ) {
			Wml::Vector2f v1 = proj.Get2D( m_vLoopVerts[ loop.nStart + j ] );
			Wml::Vector2f v2 = proj.Get2D( m_vLoopVerts[ loop.nStart + ( (j+1)%loop.nCount) ] );
			Wml::Segment2f seg2;
			seg2.Origin() = v1;
			seg2.Direction() = v2 - v1;
			if ( Wml::FindIntersection(seg2, seg, nQuantity, fT) ) {
				Wml::Vector3f vA( m_vLoopVerts[ loop.nStart + j ] );
				Wml::Vector3f vB(  m_vLoopVerts[ loop.nStart + ( (j+1)%loop.nCount) ]  );
				vHit3D = vA + fT[0] * (vB - vA);
				return true;
			}
		}
	}

	nCount = m_vLines.size();
	for ( unsigned int i = 0; i < nCount; ++i ) {
		Line & q = m_vLines[i];
		Wml::Vector2f v1 = proj.Get2D( q.vVerts[0] );
		Wml::Vector2f v2 = proj.Get2D( q.vVerts[1] );
		Wml::Segment2f seg2;
		seg2.Origin() = v1;
		seg2.Direction() = v2 - v1;
		if ( Wml::FindIntersection(seg2, seg, nQuantity, fT) ) {
			vHit3D = q.vVerts[0] + fT[0] * (q.vVerts[1] - q.vVerts[0]);
			return true;
		}
	
	}

	return false;
}





