#ifndef __RMRSUI_GEOMETRY_SET_H__
#define __RMRSUI_GEOMETRY_SET_H__

#include <WmlVector3.h>
#include <WmlRay3.h>
#include <WmlCircle3.h>
#include <WmlVector2.h>
#include <WmlSegment2.h>
#include <WmlColors.h>
#include <Frame.h>
#include <WmlTriangle3.h>

#include <vector>

namespace rmsui 
{

class GeometrySet 
{
public:
	//! type for tagging geometry, so we know what was clicked...
	typedef int GeomTag;

	GeometrySet();
	~GeometrySet();

	//! alpha on geometry color is multiplied by this value during rendering
	float & RenderAlpha() { return m_fRenderAlpha; }

	//! fixed rendering colors
	Wml::ColorRGBA & RenderGeomColor() { return m_cRenderGeomColor; }
	Wml::ColorRGBA & RenderLineColor() { return m_cRenderLineColor; }


	void Render( bool bUseGeomColors = true, bool bUseLineColors = true );
	bool Intersect( const Wml::Ray3f & ray, Wml::Vector3f & ptHit, GeomTag & nHitTag );

	void Clear(bool bClearFrame = true);


	//! default tag for incoming geometry
	GeomTag & CurrentTag() { return m_nCurrentTag; }

	//! default incoming geometry color
	Wml::ColorRGBA & CurrentColor() { return m_cCurrentColor; }

	//! origin is added to all incoming geometry (default is zero)
	Wml::Vector3f & Origin() { return m_vFrame.Origin(); }

	//! all incoming geometry is projected to this frame (default is canonical x/y/z)
	Wml::Frame3f & Frame() { return m_vFrame; }

	void AddTriangle( const Wml::Vector3f & v1, const Wml::Vector3f & v2, const Wml::Vector3f & v3 )
		{ AddTriangle(v1,v2,v3, m_cCurrentColor); }
	void AddTriangle( const Wml::Vector3f & v1, const Wml::Vector3f & v2, const Wml::Vector3f & v3, const Wml::ColorRGBA & cColor );

	void AddQuad( const Wml::Vector3f & v1, const Wml::Vector3f & v2, const Wml::Vector3f & v3, const Wml::Vector3f & v4 )
		{ AddQuad(v1,v2,v3,v4, m_cCurrentColor); }
	void AddQuad( const Wml::Vector3f & v1, const Wml::Vector3f & v2, const Wml::Vector3f & v3, const Wml::Vector3f & v4, const Wml::ColorRGBA & cColor );

	void AddQuad( const Wml::AxisAlignedBox2f & box )
		{ AddQuad(box, m_cCurrentColor); }
	void AddQuad( const Wml::AxisAlignedBox2f & box, const Wml::ColorRGBA & cColor );

	void BeginTriFan()
		{ BeginTriFan(m_cCurrentColor); }
	void BeginTriFan(const Wml::ColorRGBA & cColor);
	void AppendTriFanVert( const Wml::Vector3f & v1);


	void BeginQuadStrip(bool bCullBackFaces = false)
		{ BeginQuadStrip(m_cCurrentColor, bCullBackFaces); }
	void BeginQuadStrip(const Wml::ColorRGBA & cColor, bool bCullBackFaces = false);
	void AddQuadStripElem( const Wml::Vector3f & v1, const Wml::Vector3f & v2 );
	void QuadStripToLoop();
	void AppendQuadStripBorderLoops(const Wml::ColorRGBA & cColor );
	void AppendQuadStripInteriorEdges(const Wml::ColorRGBA & cColor, int nSkip = 0 );

	void AddLine( const Wml::Vector3f & v1, const Wml::Vector3f & v2 )
		{ AddLine(v1,v2, m_cCurrentColor); }
	void AddLine( const Wml::Vector3f & v1, const Wml::Vector3f & v2, const Wml::ColorRGBA & cColor );

	void BeginLoop() 
		{ BeginLoop(m_cCurrentColor); }
	void BeginLoop(const Wml::ColorRGBA & cColor);
	void AppendLoopVert( const Wml::Vector3f & vVert );


	void AddSphere( const Wml::ColorRGBA & cColor, float fRadius, int nSlices, int nStacks );
	void AddCube( const Wml::ColorRGBA & cFaceColor, float fEdgeLength, bool bEdges, const Wml::ColorRGBA & cEdgeColor = Wml::ColorRGBA::BLACK);

	//! circle is created wrt X/Y vectors of frame
	void AddCircle( const Wml::ColorRGBA & cColor, const Wml::Frame3f & vFrame, float fRadius, int nSteps = 32 );
	void AddDisc( const Wml::ColorRGBA & cColor, const Wml::Frame3f & vFrame, float fRadius, int nSteps = 32 );


	//! functions for screen intersection
	class Projector {
		public:
		virtual Wml::Vector2f Get2D(const Wml::Vector3f & v3D) = 0;
	};
	bool TestLinesCross(const Wml::Segment2f & seg, Projector & proj, Wml::Vector3f & vHit3D );

protected:
	GeomTag m_nCurrentTag;
	Wml::ColorRGBA m_cCurrentColor;
	Wml::Frame3f m_vFrame;

	float m_fRenderAlpha;
	Wml::ColorRGBA m_cRenderGeomColor;
	Wml::ColorRGBA m_cRenderLineColor;

	struct Tri {
		GeomTag tag;
		Wml::Vector3f vVerts[3];
		Wml::ColorRGBA cColor;
	};
	struct Quad {
		GeomTag tag;
		Wml::Vector3f vVerts[4];
		Wml::ColorRGBA cColor;
	};


	struct TriFan {
		GeomTag tag;
		int nStart;
		int nCount;
		Wml::ColorRGBA cColor;
	};

	struct QuadStripElem {
		Wml::Vector3f vVerts[2];
	};
	struct QuadStrip {
		GeomTag tag;
		bool bCullBackFaces;
		int nStart;
		int nCount;
		Wml::ColorRGBA cColor;
	};

	std::vector<Tri> m_vTriangles;
	std::vector<Quad> m_vQuads;

	std::vector<TriFan> m_vTriFans;
	std::vector<Wml::Vector3f> m_vTriFanVerts;

	std::vector<QuadStrip> m_vQuadStrips;
	std::vector<QuadStripElem> m_vQuadStripElems;



	struct Line {
		GeomTag tag;
		Wml::Vector3f vVerts[2];
		Wml::ColorRGBA cColor;
	};
	std::vector<Line> m_vLines;

	struct LineLoop {
		GeomTag tag;
		int nStart;
		int nCount;
		Wml::ColorRGBA cColor;
	};
	std::vector<LineLoop> m_vLoops;
	std::vector<Wml::Vector3f> m_vLoopVerts;

};


}   // end namespace rmsui

#endif // __RMRSUI_GEOMETRY_SET_H__