// Ryan Schmidt - rms@unknownroad.com
// Copyright (c) 2005-2010
// Distributed under the Boost Software License, Version 1.0.
// http://www.boost.org/LICENSE_1_0.txt


#include "StdAfx.h"
#include "HUDButton.h"

using namespace rmsui;

#include <GL/gl.h>
#include <GL/glu.h>
#include <WmlExtVectorUtil.h>
#include <WmlIntrLin2Box2.h>

HUDButton::HUDButton( const std::string & name, HUDIcon * pIcon, bool bSendCaptureEvents, bool bSendCrossEvents )
	: HUDItem(name), m_bCapturingEnabled(bSendCaptureEvents), m_bCrossingEnabled(bSendCrossEvents)
{
	ASSERT(pIcon);
	m_pIcon = pIcon;

	m_bConsuming = false;
	m_vLastPoint = Wml::Vector2f::ZERO;
}

HUDButton::~HUDButton()
{
	if ( m_pIcon )
		delete m_pIcon;
}



HUDItem * HUDButton::BeginConsume( const Wml::Vector2f & vPoint )
{
	if ( m_bCapturingEnabled && Wml::Contained( Bounds(), vPoint.X(), vPoint.Y() ) ) {
		m_bConsuming = true;
		m_vLastPoint = vPoint;
		NotifyDragStart(vPoint);
		return this;
	} else 
		return NULL;
}

HUDItem * HUDButton::Consume( const Wml::Vector2f & vNextPoint )
{
	// compute delta here and send it to listeners...
	if ( m_bCapturingEnabled && m_bConsuming ) {
		Wml::Vector2f vDelta( vNextPoint - m_vLastPoint );
		NotifyDragDelta(vDelta);

		m_vLastPoint = vNextPoint;

		return this;
	} else 
		return NULL;
}

void HUDButton::EndConsume()
{
	NotifyDragEnd( m_vLastPoint );
	m_bConsuming = false;
}


HUDItem * HUDButton::TestLineStroke( const GestureLineStroke & stroke )
{
	if ( ! m_bCrossingEnabled)
		return NULL;

	Wml::Box2f box;
	box.Center().X() = 0.5f * (Bounds().GetXMax() + Bounds().GetXMin());
	box.Center().Y() = 0.5f * (Bounds().GetYMax() + Bounds().GetYMin());
	box.Axis(0) = Wml::Vector2f::UNIT_X;
	box.Axis(1) = Wml::Vector2f::UNIT_Y;
	box.Extent(0) = box.Center().X() - Bounds().GetXMin();
	box.Extent(1) = box.Center().Y() - Bounds().GetYMin();

	Wml::Segment2f seg;
	seg.Origin() = stroke.EndPoint1();
	seg.Direction() = stroke.EndPoint2() - stroke.EndPoint1();

	if ( Wml::TestIntersection(seg, box) == true ) {
		NotifyCross();
		return this;
	} else
		return NULL;
}


void HUDButton::Render()
{
	m_pIcon->Draw( Bounds() );
}


void HUDButton::NotifyCross()
{
	std::set<HUDEventHandler *>::iterator cur(m_vHandlers.begin()), end(m_vHandlers.end());
	while ( cur != end ) {
		HUDButtonEventHandler * pHandler = dynamic_cast<HUDButtonEventHandler *>(*cur++);
		if ( pHandler )
			pHandler->OnCross();
	}
}

void HUDButton::NotifyDragDelta( const Wml::Vector2f & vDelta )
{
	std::set<HUDEventHandler *>::iterator cur(m_vHandlers.begin()), end(m_vHandlers.end());
	while ( cur != end ) {
		HUDButtonEventHandler * pHandler = dynamic_cast<HUDButtonEventHandler *>(*cur++);
		if ( pHandler )
			pHandler->OnDragDelta(vDelta);
	}
}

void HUDButton::NotifyDragStart( const Wml::Vector2f & vStart )
{
	std::set<HUDEventHandler *>::iterator cur(m_vHandlers.begin()), end(m_vHandlers.end());
	while ( cur != end ) {
		HUDButtonEventHandler * pHandler = dynamic_cast<HUDButtonEventHandler *>(*cur++);
		if ( pHandler )
			pHandler->OnDragStart(vStart);
	}
}

void HUDButton::NotifyDragEnd( const Wml::Vector2f & vEnd )
{
	std::set<HUDEventHandler *>::iterator cur(m_vHandlers.begin()), end(m_vHandlers.end());
	while ( cur != end ) {
		HUDButtonEventHandler * pHandler = dynamic_cast<HUDButtonEventHandler *>(*cur++);
		if ( pHandler )
			pHandler->OnDragEnd(vEnd);
	}
}