// Ryan Schmidt - rms@unknownroad.com
// Copyright (c) 2005-2010
// Distributed under the Boost Software License, Version 1.0.
// http://www.boost.org/LICENSE_1_0.txt


#include "StdAfx.h"
#include "HUDPopUpToolbar.h"
#include "..\SystemConstants.h"

using namespace rmsui;

#include <GL/gl.h>
#include <GL/glu.h>
#include <WmlExtVectorUtil.h>
#include <WmlIntrLin2Box2.h>
#include <limits>

#include "rmsdebug.h"


class HUDPopUpOpenCloseButtonHandler : public HUDButtonEventHandler
{
public:
	HUDPopUpOpenCloseButtonHandler( HUDPopUpToolbar * pToolbar ) 
		{ m_pToolbar = pToolbar; }
	virtual ~HUDPopUpOpenCloseButtonHandler() {}

	virtual void OnDragStart( const Wml::Vector2f & vStartPos )	{ m_vStart = vStartPos; }
	virtual void OnDragDelta( const Wml::Vector2f & vDelta ) {}

	virtual void OnDragEnd( const Wml::Vector2f & vEndPos ) {
		Wml::Vector2f vDir( vEndPos - m_vStart );
		vDir.Normalize();
		float fVDot = vDir.Dot( Wml::Vector2f(0,1) );
		if ( m_pToolbar->GetOrientation() == HUDPopUpToolbar::Vertical ) {
			if ( m_pToolbar->IsExpanded() && fVDot > 0.5f )
				m_pToolbar->ForceContract();
			else if ( ! m_pToolbar->IsExpanded() && fVDot < -0.5f )
				m_pToolbar->ForceExpand();
		}
	}

	virtual void OnCross() {
		if ( m_pToolbar->IsExpanded() )
			m_pToolbar->ForceContract();
		else if ( ! m_pToolbar->IsExpanded() )
			m_pToolbar->ForceExpand();
	}

protected:
	Wml::Vector2f m_vStart;
	HUDPopUpToolbar * m_pToolbar;
};



HUDPopUpToolbar::HUDPopUpToolbar( const std::string & name, Orientation eOrientation )
	: HUDItem(name)
{
	m_eOrientation = eOrientation;
	m_fButtonSpacing = 0.025f;

	m_fMinButtonSize = 0.0f;

	m_bExpanded = false;
	m_nCurrentItem = std::numeric_limits<unsigned int>::max();

	m_bConsumingPossibleExpand = false;

	for ( int i = 0; i < 4; ++i )
		m_pArrowButtons[i] = NULL;

	InitializeToolbar();

	// create buttons
	static const float fButtonSize = 0.125f;
	rmsui::HUDFileIcon * pIcon = new rmsui::HUDFileIcon(  CSystemConstants::GetIconPath() + CString("PopUpDownArrow.png") );
	m_pArrowButtons[0] = new rmsui::HUDButton("PopUpArrowDownButton", pIcon, true, true);
	m_pArrowButtons[0]->Bounds() = Wml::AxisAlignedBox2f(0.0f, fButtonSize, 0.0f, 0.5f*fButtonSize);

	pIcon = new rmsui::HUDFileIcon(  CSystemConstants::GetIconPath() + CString("PopUpUpArrow.png") );
	m_pArrowButtons[1] = new rmsui::HUDButton("PopUpArrowUpButton", pIcon, true, true);
	m_pArrowButtons[1]->Bounds() = Wml::AxisAlignedBox2f(0.0f, fButtonSize, 0.0f, 0.5f*fButtonSize);

	// NOTE! ONLY ITERATING TO 2!!
	for ( int i = 0; i < 2; ++i ) {
		m_pArrowButtonHandlers[i] = new HUDPopUpOpenCloseButtonHandler( this );
		m_pArrowButtons[i]->AddListener( m_pArrowButtonHandlers[i] );
	}

}

HUDPopUpToolbar::~HUDPopUpToolbar()
{
}


void HUDPopUpToolbar::InitializeToolbar()
{
	// update bounds
	ComputeBounds();
}

void HUDPopUpToolbar::SetOrientation( HUDPopUpToolbar::Orientation eOrientation )
{
	m_eOrientation = eOrientation;
	RegenerateLayout();
}


void HUDPopUpToolbar::AppendItem( HUDItem * pItem )
{
	m_vItems.push_back(pItem);
	if ( m_nCurrentItem == std::numeric_limits<unsigned int>::max() )
		SetCurrentItem( 0 );

	InsertItem( pItem );

	ComputeBounds();
}


void HUDPopUpToolbar::InsertItem( HUDItem * pItem )
{
	switch ( m_eOrientation ) {
		case Horizontal:
			InsertItem_Horizontal(pItem);
			break;

		case Vertical:
			InsertItem_Vertical(pItem);
			break;
	}
}

void HUDPopUpToolbar::InsertItem_Horizontal( HUDItem * pItem )
{
	// get Item size
	Wml::AxisAlignedBox2f ItemBox = pItem->Bounds();
	float fWidth = Wml::GetDimension( ItemBox, 0 );
	float fHeight = Wml::GetDimension( ItemBox, 1 );

	// append to right of toolbar, and such that top of Item is in line with top of close Item
	float fX = Bounds().GetXMax() + m_fButtonSpacing;
	float fY = Bounds().GetYMax() - fHeight;

	pItem->Translate( fX, fY, false );
}

void HUDPopUpToolbar::InsertItem_Vertical( HUDItem * pItem )
{
	// get Item size
	Wml::AxisAlignedBox2f ItemBox = pItem->Bounds();
	float fWidth = Wml::GetDimension( ItemBox, 0 );
	float fHeight = Wml::GetDimension( ItemBox, 1 );

	// append to bottom of toolbar
	
	// resize X if necessary
	Wml::AxisAlignedBox2f toolbarBox = Bounds();
	if ( fWidth > Wml::GetDimension( toolbarBox, 0 ) ) {

		float fDeltaX = toolbarBox.XMin() + fWidth - toolbarBox.XMax();
		toolbarBox.XMax() = toolbarBox.XMin() + fWidth;
	}

	// resize Y to fit
	float fOldBottom = toolbarBox.YMin();
	toolbarBox.YMin() -= m_fButtonSpacing + fHeight;

	// ok set new bounds for toolbar
	Bounds() = toolbarBox;

	// now figure out button bounds
	pItem->Translate( toolbarBox.XMin(), toolbarBox.YMin(), false );
}


void HUDPopUpToolbar::UpdateArrowButtons_Vertical()
{
	// compute new vertical position
	Wml::AxisAlignedBox2f ItemBox = m_pArrowButtons[0]->Bounds();
	float fWidth = Wml::GetDimension( ItemBox, 0 );
	float fHeight = Wml::GetDimension( ItemBox, 1 );

	// find new Y bottom
	Wml::AxisAlignedBox2f toolbarBox = Bounds();
	float fNewYMin = toolbarBox.YMin() - (0.5f*m_fButtonSpacing + fHeight);

	// re-position toolbar buttons
	// TODO: should iterate to 4!!
	for ( int i = 0; i < 2; ++i )
		m_pArrowButtons[i]->Translate( toolbarBox.XMin(), fNewYMin, false );

}

void HUDPopUpToolbar::ClearItems()
{
	m_vItems.clear();
	m_nCurrentItem = std::numeric_limits<unsigned int>::max();

	ComputeBounds();

	//switch ( m_eOrientation ) {
	//	case Horizontal:
	//		Bounds().XMax() = Bounds().XMin();
	//		break;

	//	case Vertical:
	//		Bounds().YMin() = Bounds().YMax();
	//		break;
	//}
	
}


void HUDPopUpToolbar::Translate( float fX, float fY, bool bRelative )
{
	size_t nCount = m_vItems.size();

	// if not relative, figure out relative translation for entire box
	if ( ! bRelative ) {
		fX = fX - Bounds().GetXMin();
		fY = fY - Bounds().GetYMin(); 
	}

	for ( unsigned int i = 0; i < nCount; ++i ) {
		m_vItems[i]->Translate( fX, fY, true );
	}

	// compute new bounds
	HUDItem::Translate( fX, fY, true );
	ComputeBounds();
}




HUDItem * HUDPopUpToolbar::BeginConsume( const Wml::Vector2f & vPoint )
{
	HUDItem * pArrowItem = NULL;
	switch ( m_eOrientation ) {
		case Vertical:
			if ( m_bExpanded )
				pArrowItem = m_pArrowButtons[1]->BeginConsume(vPoint);
			else
				pArrowItem = m_pArrowButtons[0]->BeginConsume(vPoint);
	}
	if ( pArrowItem )
		return pArrowItem;

	if ( ! m_bExpanded ) {
			HUDItem * pItem = m_vItems[m_nCurrentItem]->BeginConsume(vPoint);
			if ( pItem != NULL )
				return pItem;
	} else {
		size_t nCount = m_vItems.size();
		for ( unsigned int i = 0; i < nCount; ++i ) {
			HUDItem * pItem = m_vItems[i]->BeginConsume(vPoint);
			if ( pItem != NULL )
				return pItem;
		}
	}
	return NULL;
}

HUDItem * HUDPopUpToolbar::Consume( const Wml::Vector2f & vNextPoint )
{
	if ( m_bConsumingPossibleExpand ) {

		if ( ! Wml::Contained( m_vItems[m_nCurrentItem]->Bounds(), vNextPoint.X(), vNextPoint.Y() ) ) 
			ExpandToolbar();

		return this;
	}
	return NULL;
}

void HUDPopUpToolbar::EndConsume()
{
	m_bConsumingPossibleExpand = false;
}


HUDItem * HUDPopUpToolbar::TestLineStroke( const GestureLineStroke & stroke )
{
	if ( ! m_bExpanded ) {
		HUDItem * pItem = m_vItems[m_nCurrentItem]->TestLineStroke(stroke);
		if ( pItem != NULL ) {
			return pItem;
		}
	} else {
		size_t nCount = m_vItems.size();
		for ( unsigned int i = 0; i < nCount; ++i ) {
			HUDItem * pItem = m_vItems[i]->TestLineStroke(stroke);
			if ( pItem != NULL ) {
				SetCurrentItem( i );
				ContractToolbar();
				return pItem;
			}
		}
	}

	// try open-close button last
	HUDItem * pArrowItem = NULL;
	switch ( m_eOrientation ) {
		case Vertical:
			if ( m_bExpanded )
				pArrowItem = m_pArrowButtons[1]->TestLineStroke(stroke);
			else
				pArrowItem = m_pArrowButtons[0]->TestLineStroke(stroke);
	}
	if ( pArrowItem )
		return pArrowItem;

	return NULL;
}


void HUDPopUpToolbar::Render()
{
	//const Wml::AxisAlignedBox2f & bounds = Bounds();
	//glColor3f(0,0,0);
	//glBegin(GL_LINE_LOOP);
	//glLineWidth(2.0f);
	//glVertex2f( bounds.GetXMin(), bounds.GetYMin() );
	//glVertex2f( bounds.GetXMax(), bounds.GetYMin() );
	//glVertex2f( bounds.GetXMax(), bounds.GetYMax() );
	//glVertex2f( bounds.GetXMin(), bounds.GetYMax() );
	//glEnd();

	if  ( ! m_bExpanded ) {
		m_vItems[m_nCurrentItem]->Render();
	} else {
		size_t nCount = m_vItems.size();
		for ( unsigned int i = 0; i < nCount; ++i )
			m_vItems[i]->Render();
	}

	// render arrow button
	switch ( m_eOrientation ) {
		case Vertical:
			if ( m_bExpanded )
				m_pArrowButtons[1]->Render();
			else
				m_pArrowButtons[0]->Render();
	}
}

void HUDPopUpToolbar::SetCurrentItem( unsigned int nItem )
{
	if ( m_nCurrentItem != std::numeric_limits<unsigned int>::max() ) {
		// need to copy bounds of old current item (!)
		Wml::Vector2f vCorner( 
			m_vItems[ m_nCurrentItem ]->Bounds().GetXMin(),
			m_vItems[ m_nCurrentItem ]->Bounds().GetYMin() );

		m_vItems[ nItem ]->Translate( vCorner.X(), vCorner.Y(), false );
	}

	m_nCurrentItem = nItem;
}

void HUDPopUpToolbar::ExpandToolbar()
{
	m_bExpanded = true;
	RegenerateLayout();
	NotifyToolbarChanged(true);
}

void HUDPopUpToolbar::ContractToolbar()
{
	m_bExpanded = false;
	RegenerateLayout();
	NotifyToolbarChanged(false);
}


void HUDPopUpToolbar::NotifyToolbarChanged(bool bExpanded)
{
	std::set<HUDEventHandler *>::iterator cur(m_vHandlers.begin()), end(m_vHandlers.end());
	while ( cur != end ) {
		IHUDPopUpToolbarEventHandler * pHandler = dynamic_cast<IHUDPopUpToolbarEventHandler *>(*cur++);
		if ( pHandler ) {
			if ( bExpanded )
				pHandler->OnHUDPopUpToolbarExpand();
			else
				pHandler->OnHUDPopUpToolbarContract();
		}
	}
}

void HUDPopUpToolbar::ComputeBounds()
{
	if  ( m_nCurrentItem == std::numeric_limits<unsigned int>::max() ) {
		m_bounds = Wml::AxisAlignedBox2f(0,0,0,0);
		return;
	}

	// compute bounds of current button
	m_bounds = m_vItems[m_nCurrentItem]->Bounds();

	if ( m_bExpanded ) {
		for ( unsigned int i = 0; i < m_vItems.size(); ++i )
			Wml::Union( m_bounds, m_vItems[i]->Bounds() );
	}

	// update arrow buttons
	if ( m_eOrientation == Vertical )
		UpdateArrowButtons_Vertical();
}


void HUDPopUpToolbar::RegenerateLayout( )
{
	// copy list of Items
	HUDItem * pCurrent = m_vItems[ m_nCurrentItem ];

	// save position of selected (main) button
	Wml::Vector2f vSaved( pCurrent->Bounds().GetXMin(),  pCurrent->Bounds().GetYMin() );

	// re-compute bounds
	InitializeToolbar();

	// add selected button
//	InsertItem( m_vItems[m_nCurrentItem] );
	m_bounds = pCurrent->Bounds();

	// now add all buttons, but skip current item
	for ( unsigned int i = 0; i < m_vItems.size(); ++i ) {
		if ( i != m_nCurrentItem )
			InsertItem( m_vItems[i] );
	}

	ComputeBounds();
//	Translate( vSaved.X(), vSaved.Y(), false );
}