// Ryan Schmidt - rms@unknownroad.com
// Copyright (c) 2005-2010
// Distributed under the Boost Software License, Version 1.0.
// http://www.boost.org/LICENSE_1_0.txt


#include "StdAfx.h"
#include "HUDToolbar.h"
#include "..\SystemConstants.h"

using namespace rmsui;

#include <GL/gl.h>
#include <GL/glu.h>
#include <WmlExtVectorUtil.h>
#include <WmlIntrLin2Box2.h>

#include "rmsdebug.h"


static const float s_fToolbarCloseButtonSize = 0.05f;


HUDToolbar::HUDToolbar( const std::string & name, bool bAllowClose, Orientation eOrientation )
	: HUDItem(name), 
	  m_closeButtonHandler(this),
	  m_pCloseButton(NULL)
{
	m_eOrientation = eOrientation;
	m_fButtonSpacing = 0.025f;

	InitializeToolbar();

	m_pCloseButton->AddListener( &m_closeButtonHandler );

	m_bAllowClose = bAllowClose;

	m_fMinButtonSize = 0.0f;
}

HUDToolbar::~HUDToolbar()
{
}


void HUDToolbar::InitializeToolbar()
{
	if ( m_pCloseButton == NULL ) {
		HUDIcon * m_pCloseIcon = new rmsui::HUDFileIcon(CSystemConstants::GetIconPath() + "closeX.png");
		m_pCloseButton = new rmsui::HUDButton( Name() + "CloseButton", m_pCloseIcon, true, true );
	}

	// create close button

	switch ( m_eOrientation ) {
		case Horizontal:
		{
			m_pCloseButton->Bounds() = Wml::AxisAlignedBox2f( -0.5f * s_fToolbarCloseButtonSize, 
															   0.5f * s_fToolbarCloseButtonSize, 
															   -s_fToolbarCloseButtonSize, 0.0f );
		}
		break;

		case Vertical:
		{
			if ( GetDockMode() == HUDItem::DockRight ) {
				m_pCloseButton->Bounds() = Wml::AxisAlignedBox2f( 0.0f, s_fToolbarCloseButtonSize,
					-0.5f*s_fToolbarCloseButtonSize, 0.5f*s_fToolbarCloseButtonSize );
			} else {
				m_pCloseButton->Bounds() = Wml::AxisAlignedBox2f( 0.0f, s_fToolbarCloseButtonSize,
					-0.5f*s_fToolbarCloseButtonSize, 0.5f*s_fToolbarCloseButtonSize );
			}
		}
		break;
	}

	// update bounds
	ComputeBounds();
}

void HUDToolbar::SetOrientation( HUDToolbar::Orientation eOrientation )
{
	m_eOrientation = eOrientation;
	RegenerateLayout();
}


void HUDToolbar::AppendItem( HUDItem * pItem )
{
	m_vItems.push_back(pItem);

	switch ( m_eOrientation ) {
		case Horizontal:
			InsertItem_Horizontal(pItem);
			break;

		case Vertical:
			InsertItem_Vertical(pItem);
			break;
	}

	ComputeBounds();
}


void HUDToolbar::InsertItem_Horizontal( HUDItem * pItem )
{
	// get Item size
	Wml::AxisAlignedBox2f ItemBox = pItem->Bounds();
	float fWidth = Wml::GetDimension( ItemBox, 0 );
	float fHeight = Wml::GetDimension( ItemBox, 1 );

	// append to right of toolbar, and such that top of Item is in line with top of close button
	float fX = Bounds().GetXMax() + m_fButtonSpacing;
	float fY = Bounds().GetYMax() - fHeight;

	pItem->Translate( fX, fY, false );
	
	float fRight = Bounds().GetXMax();
	float fBarRight = m_pCloseButton->Bounds().GetXMax();

	//// If this is first button, close button should go at it's top-left.
	//// it should already be at left, so just have to adjust to be at top...
	//if ( m_vItems.size() == 1 ) {
	//	float fCloseHeight = Wml::GetDimension( m_pCloseButton->Bounds(), 1 );
	//	m_pCloseButton->Translate( 
	//		m_pCloseButton->Bounds().GetXMin(), fY + fHeight - fCloseHeight, false );
	//}
}

void HUDToolbar::InsertItem_Vertical( HUDItem * pItem )
{
	// get Item size
	Wml::AxisAlignedBox2f ItemBox = pItem->Bounds();
	float fWidth = Wml::GetDimension( ItemBox, 0 );
	float fHeight = Wml::GetDimension( ItemBox, 1 );

	// append to bottom of toolbar
	
	// resize X if necessary
	Wml::AxisAlignedBox2f toolbarBox = Bounds();
	if ( fWidth > Wml::GetDimension( toolbarBox, 0 ) ) {

		float fDeltaX = toolbarBox.XMin() + fWidth - toolbarBox.XMax();
		toolbarBox.XMax() = toolbarBox.XMin() + fWidth;

		// need to bump up close button
		//Wml::Translate( m_pCloseButton->Bounds(), 0.0f, fDeltaX, true);
	}

	// resize Y to fit
	float fOldBottom = toolbarBox.YMin();
	toolbarBox.YMin() -= m_fButtonSpacing + fHeight;

	// ok set new bounds for toolbar
	Bounds() = toolbarBox;

	// now figure out button bounds
	pItem->Translate( toolbarBox.XMin(), toolbarBox.YMin(), false );
}


void HUDToolbar::ClearItems()
{
	m_vItems.clear();

	ComputeBounds();

	//switch ( m_eOrientation ) {
	//	case Horizontal:
	//		Bounds().XMax() = Bounds().XMin();
	//		break;

	//	case Vertical:
	//		Bounds().YMin() = Bounds().YMax();
	//		break;
	//}
	
}


void HUDToolbar::Translate( float fX, float fY, bool bRelative )
{
	size_t nCount = m_vItems.size();

	// if not relative, figure out relative translation for entire box
	if ( ! bRelative ) {
		fX = fX - Bounds().GetXMin();
		fY = fY - Bounds().GetYMin(); 
	}

	for ( unsigned int i = 0; i < nCount; ++i ) {
		m_vItems[i]->Translate( fX, fY, true );
	}

	// translate close button
	m_pCloseButton->Translate( fX, fY, true );

	// compute new bounds
	HUDItem::Translate( fX, fY, true );
}




HUDItem * HUDToolbar::BeginConsume( const Wml::Vector2f & vPoint )
{
	// try close button first
	HUDItem * pItem = m_pCloseButton->BeginConsume(vPoint);
	if ( pItem != NULL )
		return pItem;

	size_t nCount = m_vItems.size();
	for ( unsigned int i = 0; i < nCount; ++i ) {
		pItem = m_vItems[i]->BeginConsume(vPoint);
		if ( pItem != NULL )
			return pItem;
	}
	return NULL;
}

HUDItem * HUDToolbar::Consume( const Wml::Vector2f & vNextPoint )
{
	ASSERT(0);		// toolbar should not consume anything...
	return NULL;
}

void HUDToolbar::EndConsume()
{
	ASSERT(0);		// why would this be called??
}


HUDItem * HUDToolbar::TestLineStroke( const GestureLineStroke & stroke )
{
	// try close button first
	HUDItem * pItem = m_pCloseButton->TestLineStroke(stroke);
	if ( pItem != NULL ) {
		return pItem;
	}

	size_t nCount = m_vItems.size();
	for ( unsigned int i = 0; i < nCount; ++i ) {
		pItem = m_vItems[i]->TestLineStroke(stroke);
		if ( pItem != NULL ) {
			return pItem;
		}
	}
	return NULL;
}


void HUDToolbar::Render()
{
	m_pCloseButton->Render();

	size_t nCount = m_vItems.size();
	for ( unsigned int i = 0; i < nCount; ++i )
		m_vItems[i]->Render();
}



void HUDToolbar::CloseButtonEventHandler::OnDragDelta( const Wml::Vector2f & vDelta )
{
	m_pToolbar->Translate( vDelta.X(), vDelta.Y(), true );

	// notify listeners
	m_pToolbar->NotifyToolbarMove();
}

void HUDToolbar::CloseButtonEventHandler::OnCross()
{
	if ( m_pToolbar->m_bAllowClose ) {
		m_pToolbar->Visible() = false;

		// notify listeners
		m_pToolbar->NotifyToolbarHide();
	}
}




void HUDToolbar::NotifyToolbarMove()
{
	std::set<HUDEventHandler *>::iterator cur(m_vHandlers.begin()), end(m_vHandlers.end());
	while ( cur != end ) {
		IHUDToolbarEventHandler * pHandler = dynamic_cast<IHUDToolbarEventHandler *>(*cur++);
		if ( pHandler )
			pHandler->OnHUDToolbarMove();
	}
}


void HUDToolbar::NotifyToolbarHide()
{
	std::set<HUDEventHandler *>::iterator cur(m_vHandlers.begin()), end(m_vHandlers.end());
	while ( cur != end ) {
		IHUDToolbarEventHandler * pHandler = dynamic_cast<IHUDToolbarEventHandler *>(*cur++);
		if ( pHandler )
			pHandler->OnHUDToolbarHide();
	}
}


void HUDToolbar::ComputeBounds()
{
	m_bounds = m_pCloseButton->Bounds();

	if( m_eOrientation == Horizontal ) {
		float fHeight = Wml::GetDimension( Bounds(), 0 );
		if ( fHeight < m_fMinButtonSize )
			m_bounds.YMin() -= (m_fMinButtonSize - fHeight);
	}

	for ( unsigned int i = 0; i < m_vItems.size(); ++i )
		Wml::Union( m_bounds, m_vItems[i]->Bounds() );
}


void HUDToolbar::RegenerateLayout()
{
	// copy list of Items
	std::vector< HUDItem * > vSavedItems = m_vItems;

	// save position
	Wml::Vector2f vPosition(
		Bounds().GetXMin(), Bounds().GetYMin() );

	// clear Items
	ClearItems();

	// re-set close button
	InitializeToolbar();

	// re-add Items
	for ( unsigned int i = 0; i < vSavedItems.size(); ++i )
		AppendItem( vSavedItems[i] );

	Translate( vPosition.X(), vPosition.Y(), false );
}