// Ryan Schmidt - rms@unknownroad.com
// Copyright (c) 2005-2010
// Distributed under the Boost Software License, Version 1.0.
// http://www.boost.org/LICENSE_1_0.txt


#include "StdAfx.h"
#include "Image.h"

#include <png.h>
#include <WmlSegment2.h>
#include <WmlLine2.h>

#include "rmsdebug.h"

using namespace rmsui;


const unsigned int Image::RGB_CHANNEL_MAP[3] = {0,1,2};
const unsigned int Image::BGR_CHANNEL_MAP[3] = {2,1,0};
const unsigned int Image::GRAY_CHANNEL_MAP[3] = {0,0,0};

Image::Image(unsigned int nWidth, unsigned int nHeight, unsigned int nChannels)
{
	m_nWidth = nWidth;
	m_nHeight = nHeight;
	m_nChannels = nChannels;

	m_vBuffer.resize( nWidth * nHeight * nChannels );

	m_eMode = Replace;
}

Image::Image( const Image & copy )
{
	m_nWidth = copy.m_nWidth;
	m_nHeight = copy.m_nHeight;
	m_nChannels = copy.m_nChannels;
	m_vBuffer = copy.m_vBuffer;
	m_eMode = copy.m_eMode;
}



const unsigned char * Image::GetPixel( unsigned int x, unsigned int y ) const
{
	return &m_vBuffer[ y * m_nWidth * m_nChannels + x * m_nChannels ];
}

unsigned char * Image::GetPixel( unsigned int x, unsigned int y )
{
	return &m_vBuffer[ y * m_nWidth * m_nChannels + x * m_nChannels ];
}


bool Image::Load( const char * pFilename )
{
	if ( (strstr(pFilename,".png") != NULL )  )
		return LoadLibPNG( pFilename );
	else
		return LoadOpenCV( pFilename );
}

bool Image::Save( const char * pFilename ) const
{
	if ( m_nChannels < 3 ) {
		Image tmp( *this );
		tmp.ExpandChannels( 3 );
		return tmp.Save( pFilename );
	} else {
		return SaveOpenCV( pFilename );
	}
}



BOOL PngLoadImage (const char * pstrFileName, png_byte **ppbImageData,
                   unsigned int *piWidth, unsigned int *piHeight, int *piChannels, png_color *pBkgColor);

bool Image::LoadLibPNG(const char * pFilename)
{
	png_byte * pData = NULL;
	unsigned int nWidth;
	unsigned int nHeight;
	int nChannels;
	png_color backColor;

	BOOL bResult = PngLoadImage( pFilename, &pData, &nWidth, &nHeight, &nChannels, &backColor );
	ASSERT(bResult);

	InitFromBuffer( nWidth, nHeight, nChannels, pData, RGB_CHANNEL_MAP, false );

	free( pData );

	return true;
}

bool Image::LoadOpenCV(const char * pFilename)
{
	return false;
}


bool Image::SaveOpenCV(const char * pFilename) const
{
	return false;
}




// rgb is channel map [0,1,2], bgr is channel map [2,1,0]
void Image::InitFromBuffer( unsigned int nBufferWidth, unsigned int nBufferHeight, unsigned int nBufferChannels,
						    const unsigned char * pBuffer,
							const unsigned int nChannelMap[3], bool bAddAlpha )
{
	unsigned int nBufferSize = nBufferWidth * nBufferHeight * nBufferChannels;
	unsigned int nImgChannels = nBufferChannels;
	if ( nImgChannels < 4 && bAddAlpha ) 
		nImgChannels++;
	unsigned int nAlphaChannel = (nImgChannels-1);
	unsigned int nImgSize = nBufferWidth * nBufferHeight * nImgChannels;

	m_nWidth = nBufferWidth;
	m_nHeight = nBufferHeight;
	m_nChannels = nImgChannels;
	m_vBuffer.resize(nImgSize);

	for ( unsigned int i = 0; i < nBufferSize/nBufferChannels; ++i ) {
		unsigned int nImgIndex = i * nImgChannels;
		unsigned int nBufIndex = i * nBufferChannels;

		m_vBuffer[nImgIndex + nChannelMap[0]] = (unsigned char)pBuffer[nBufIndex ];
		if ( nImgChannels > 1 )
			m_vBuffer[nImgIndex + nChannelMap[1]] = (unsigned char)pBuffer[nBufIndex +1];
		if ( nImgChannels > 2 )
			m_vBuffer[nImgIndex + nChannelMap[2]] = (unsigned char)pBuffer[nBufIndex +2];
		
		if  ( nImgChannels == 4 || bAddAlpha ) {
			if ( nBufferChannels == 4 )
				m_vBuffer[nImgIndex + nAlphaChannel] = (unsigned char)pBuffer[nBufIndex +3];
			else
				m_vBuffer[nImgIndex + nAlphaChannel ] = 255;
		}
	}

}


void Image::ExpandChannels( unsigned int nNewNumChannels )
{
	if ( m_nChannels >= nNewNumChannels )
		return;

	unsigned int nPixels = m_nWidth * m_nHeight;
	unsigned int nImgSize = nPixels * nNewNumChannels;
	m_vBuffer.resize( nImgSize );

	// copy backwards...
	for ( int i = nPixels-1; i >= 0; --i ) {
		unsigned int nOldIndex = i*m_nChannels;
		unsigned int nNewIndex = i*nNewNumChannels;
		for ( unsigned int j = 0; j < nNewNumChannels; ++j ) {
			if ( j < m_nChannels  )
				m_vBuffer[ nNewIndex + j ] = m_vBuffer[ nOldIndex + j ];
			else
				m_vBuffer[ nNewIndex + j ] = m_vBuffer[ nOldIndex ];
		}
	}

	m_nChannels = nNewNumChannels;
}


void Image::Crop( const Wml::Polygon2f & vPolygon, const Wml::ColorRGBA & vCropColor )
{
	unsigned char nCropR = std::min( (unsigned char)255, std::max((unsigned char)0, (unsigned char)(vCropColor.R() * 255.0f) ) );
	unsigned char nCropG = std::min( (unsigned char)255, std::max((unsigned char)0, (unsigned char)(vCropColor.G() * 255.0f) ) );
	unsigned char nCropB = std::min( (unsigned char)255, std::max((unsigned char)0, (unsigned char)(vCropColor.B() * 255.0f) ) );
	unsigned char nCropA = std::min( (unsigned char)255, std::max((unsigned char)0, (unsigned char)(vCropColor.A() * 255.0f) ) );

	Wml::AxisAlignedBox2f vBounds;
	vPolygon.BoundingBox(vBounds);

	for ( unsigned int yi = 0 ; yi < m_nHeight; ++yi ) {
		float fY = (float)yi / (float)(m_nHeight-1);
		for ( unsigned int xi = 0; xi < m_nWidth; ++xi ) {
			float fX = (float)xi / (float)(m_nWidth-1);

			bool bCrop = false;
			if ( ! Wml::Contained( vBounds, fX, fY ) )
				bCrop = true;
			else if ( ! vPolygon.IsInside( Wml::Vector2f(fX,fY) ) )
				bCrop = true;

			if ( bCrop ) {
				unsigned int nIndex = (yi * m_nWidth * m_nChannels) + (xi * m_nChannels);
				m_vBuffer[ nIndex ] = nCropR;
				m_vBuffer[ nIndex + 1] = nCropG;
				m_vBuffer[ nIndex + 2] = nCropB;
				m_vBuffer[ nIndex + 3] = nCropA;
				
			}
		}
	}
		
}



void Image::AverageBlur( unsigned int nRadius )
{
	if ( m_nChannels < 3 )
		return;

	for ( unsigned int yi = 0 ; yi < m_nHeight; ++yi ) {
		for ( unsigned int xi = 0; xi < m_nWidth; ++xi ) {
			float fRedSum = 0, fBlueSum = 0, fGreenSum = 0, fAlphaSum = 0;
			unsigned int nPixels = 0;
			unsigned int nLeft = (unsigned int)std::max( (int)xi - (int)nRadius, 0 );
			unsigned int nRight = std::min( xi + nRadius, m_nWidth-1 );
			unsigned int nBottom = (unsigned int)std::max( (int)yi - (int)nRadius, 0 );
			unsigned int nTop = std::min( yi + nRadius, m_nHeight-1 );
			for ( unsigned int j = nBottom; j <= nTop; ++j ) {
				for ( unsigned int i = nLeft; i  <= nRight; ++i ) {
					const unsigned char * pPixel = GetPixel(i,j);
					fRedSum += pPixel[0];
					fGreenSum += pPixel[1];
					fBlueSum += pPixel[2];
					if ( m_nChannels == 4) 
						fAlphaSum += pPixel[3];
					++nPixels;
				}
			}
			unsigned char * pPixel = GetPixel(xi,yi);
			pPixel[0] = (unsigned char)( fRedSum / (float)nPixels );
			pPixel[1] = (unsigned char)( fGreenSum / (float)nPixels );
			pPixel[2] = (unsigned char)( fBlueSum / (float)nPixels );
			if ( m_nChannels == 4 )
				pPixel[3] = (unsigned char)( fAlphaSum / (float)nPixels );
		}
	}

}

void Image::RegionAverageBlur( const Wml::Polygon2f & vPolygon, unsigned int nRadius )
{
	if ( m_nChannels < 3 )
		return;

	Wml::AxisAlignedBox2f vBounds;
	vPolygon.BoundingBox(vBounds);

	for ( unsigned int yi = 0 ; yi < m_nHeight; ++yi ) {
		float fY = (float)yi / (float)(m_nHeight-1);
		for ( unsigned int xi = 0; xi < m_nWidth; ++xi ) {
			float fX = (float)xi / (float)(m_nWidth-1);

			bool bCrop = false;
			if ( ! Wml::Contained( vBounds, fX, fY ) )
				bCrop = true;
			else if ( ! vPolygon.IsInside( Wml::Vector2f(fX,fY) ) )
				bCrop = true;

			if ( !bCrop ) {

				float fRedSum = 0, fBlueSum = 0, fGreenSum = 0, fAlphaSum = 0;
				unsigned int nPixels = 0;
				unsigned int nLeft = (unsigned int)std::max( (int)xi - (int)nRadius, 0 );
				unsigned int nRight = std::min( xi + nRadius, m_nWidth-1 );
				unsigned int nBottom = (unsigned int)std::max( (int)yi - (int)nRadius, 0 );
				unsigned int nTop = std::min( yi + nRadius, m_nHeight-1 );
				for ( unsigned int j = nBottom; j <= nTop; ++j ) {
					for ( unsigned int i = nLeft; i  <= nRight; ++i ) {
						const unsigned char * pPixel = GetPixel(i,j);
						fRedSum += pPixel[0];
						fGreenSum += pPixel[1];
						fBlueSum += pPixel[2];
						if ( m_nChannels == 4) 
							fAlphaSum += pPixel[3];
						++nPixels;
					}
				}
				unsigned char * pPixel = GetPixel(xi,yi);
				pPixel[0] = (unsigned char)( fRedSum / (float)nPixels );
				pPixel[1] = (unsigned char)( fGreenSum / (float)nPixels );
				pPixel[2] = (unsigned char)( fBlueSum / (float)nPixels );
				if ( m_nChannels == 4 )
					pPixel[3] = (unsigned char)( fAlphaSum / (float)nPixels );
			}
		}
	}
}


void Image::Clear( const Wml::ColorRGBA & cClearColor, bool bAlphaOnly )
{
	for ( unsigned int yi = 0; yi < m_nHeight; ++yi ) {
		for ( unsigned int xi = 0; xi < m_nWidth; ++xi ) {
			unsigned char * pPixel = GetPixel(xi,yi);
			if ( !bAlphaOnly ) {
				for ( unsigned int ci = 0; ci < m_nChannels; ++ci )
					pPixel[ci] = Wml::ColorRGBA::ToByte(cClearColor[(int)ci]);
			} else {
				pPixel[ m_nChannels-1] = Wml::ColorRGBA::ToByte(cClearColor[3]);
			}
		}
	}
}

void Image::Invert( )
{
	unsigned int nChannels = std::min( m_nChannels, (unsigned int)3 );		// do not write to alpha channel
	for ( unsigned int yi = 0; yi < m_nHeight; ++yi ) {
		for ( unsigned int xi = 0; xi < m_nWidth; ++xi ) {
			unsigned char * pPixel = GetPixel(xi,yi);
			for ( unsigned int ci = 0; ci < nChannels; ++ci )
				pPixel[ci] = 255 - pPixel[ci];
		}
	}
}

void Image::Flip( bool bFlipX, bool bFlipY  )
{
	unsigned int nChannels = std::min( m_nChannels, (unsigned int)3 );		// do not write to alpha channel
	if ( bFlipY  &&  !bFlipX ) {
		for ( unsigned int yi = 0; yi < m_nHeight/2; ++yi ) {
			for ( unsigned int xi = 0; xi < m_nWidth; ++xi ) {
				unsigned char * pPixel1 = GetPixel(xi,yi);
				unsigned char * pPixel2 = GetPixel(xi, m_nHeight-yi-1);
				for ( unsigned int ci = 0; ci < nChannels; ++ci ) {
					unsigned char cTmp = pPixel1[ci];
					pPixel1[ci] = pPixel2[ci];
					pPixel2[ci] = cTmp;
				}
			}
		}
	} else if ( bFlipX  && !bFlipY ) {
		for ( unsigned int yi = 0; yi < m_nHeight; ++yi ) {
			for ( unsigned int xi = 0; xi < m_nWidth/2; ++xi ) {
				unsigned char * pPixel1 = GetPixel(xi,yi);
				unsigned char * pPixel2 = GetPixel(m_nWidth-xi-1, yi);
				for ( unsigned int ci = 0; ci < nChannels; ++ci ) {
					unsigned char cTmp = pPixel1[ci];
					pPixel1[ci] = pPixel2[ci];
					pPixel2[ci] = cTmp;
				}
			}
		}
	} else if ( bFlipX  && bFlipY && m_nHeight == m_nWidth ) {
		for ( unsigned int yi = 0; yi < m_nHeight; ++yi ) {
			for ( unsigned int xi = 0; xi < yi; ++xi ) {
				unsigned char * pPixel1 = GetPixel(xi,yi);
				unsigned char * pPixel2 = GetPixel(m_nHeight-yi-1, m_nWidth-xi-1);
				for ( unsigned int ci = 0; ci < nChannels; ++ci ) {
					unsigned char cTmp = pPixel1[ci];
					pPixel1[ci] = pPixel2[ci];
					pPixel2[ci] = cTmp;
				}
			}
		}
	}
		
}



void Image::CompositePixel( unsigned int nX, unsigned int nY, const Wml::ColorRGBA & cColor, float fAlpha )
{
	if ( nX >= m_nWidth || nY >= m_nHeight ) 
		return;

	unsigned char * pPixel = GetPixel( nX, nY );

	switch ( m_eMode ) {

		case Replace: 
		{
			for ( unsigned int j = 0; j < m_nChannels; ++j )
				pPixel[j] = Wml::ColorRGBA::ToByte( cColor[(int)j] );
		}
		break;

		case Max:
		{
			float fAlphaIn = cColor.A() * fAlpha;
			if (  Wml::ColorRGBA::ToByte(fAlphaIn) > pPixel[3] ) {
				for ( unsigned int j = 0; j < m_nChannels; ++j )
					pPixel[j] = Wml::ColorRGBA::ToByte( cColor[(int)j] * fAlpha );
			}
		}
		break;

		case Accumulate:
		{
			for ( unsigned int j = 0; j < 4; ++j ) {
				float fOldColor = pPixel[j] / 255.0f;
				float fSumColor = fOldColor + cColor[(int)j] * fAlpha;
				pPixel[j] = Wml::ColorRGBA::ToByte( fSumColor );
			}
		}
		break;

	}
}


void Image::DrawLine( float fFromX, float fFromY, float fToX, float fToY, const Wml::ColorRGBA & cColor, float fWidth )
{
	// figure out coords for a quad
	Wml::Vector2f vCorners[4];

	Wml::Vector2f vLine( fToX - fFromX, fToY - fFromY );
	Wml::Vector2f vLinePerp( -vLine.Y(), vLine.X() );
	vLinePerp.Normalize();
	fWidth *= 0.5f;

	vCorners[0] = Wml::Vector2f( fFromX + vLinePerp.X() * fWidth, fFromY + vLinePerp.Y() * fWidth );
	vCorners[1] = Wml::Vector2f( fToX + vLinePerp.X() * fWidth, fToY + vLinePerp.Y() * fWidth );
	vCorners[2] = Wml::Vector2f( fToX - vLinePerp.X() * fWidth, fToY - vLinePerp.Y() * fWidth );
	vCorners[3] = Wml::Vector2f( fFromX - vLinePerp.X() * fWidth, fFromY - vLinePerp.Y() * fWidth );

	FillQuadAA(vCorners, cColor);
}




struct LineInfo {
	Wml::Line2f vLine;
	float fYIntervalLow;
	float fYIntervalHigh;
};

static int qsort_ysortLine( const void * v1, const void * v2 )
{
	LineInfo * pSeg1 = (LineInfo *)v1;
	LineInfo * pSeg2 = (LineInfo *)v2;

	// sort by segment interval
	if ( pSeg1->fYIntervalLow < pSeg2->fYIntervalLow )
		return -1;
	else if ( pSeg1->fYIntervalHigh < pSeg2->fYIntervalHigh )
		return -1;
	else if ( pSeg1->fYIntervalLow == pSeg2->fYIntervalLow && pSeg1->fYIntervalHigh == pSeg2->fYIntervalHigh )
		return 0;
	else
		return 1;
}

// scan-convert a quad
void Image::FillQuad( const Wml::Vector2f vCorners[4], const Wml::ColorRGBA & cColor )
{
	LineInfo vLines[4];
	for ( int i = 0; i < 4; ++i ) {
		int iLow = i;
		int iHigh = (i+1)%4;
		bool bFlipped = false;
		if ( vCorners[iHigh].Y() < vCorners[iLow].Y() ) {
			iLow = (i+1)%4;
			iHigh = i;
			bFlipped = true;
		}

		vLines[i].vLine = Wml::Line2f( vCorners[iLow], vCorners[iHigh], true );
		vLines[i].fYIntervalLow = vCorners[iLow].Y();
		vLines[i].fYIntervalHigh = vCorners[iHigh].Y();
	}


	// sort vLine by Y intervals (first by lower miny, then lower maxy)
	qsort( vLines, 4, sizeof(LineInfo), qsort_ysortLine );

	// draw from min to max y
	float fMinY = vLines[0].fYIntervalLow;
	float fMaxY = vLines[3].fYIntervalHigh;

	int nMinY = (int)ceilf(fMinY);
	int nMaxY = (int)floorf(fMaxY);
	int nFirstI = 0;
	int nSecondI = 1;
	for ( int yi = nMinY; yi <= nMaxY; ++yi ) {

		// figure out sides of run for current row
		float fMinX = -1.0f, fMaxX = 0.0f;
		if ( yi == nMinY ) {			// bottom row is possibly horz - special case
			if ( vLines[0].vLine.Direction().Y() == 0 ) {
				fMinX = vLines[0].vLine.Origin().X();
				fMaxX = vLines[0].vLine.Origin().X() + vLines[0].vLine.Direction().X();

				nSecondI++;
				nFirstI = nSecondI-1;		// increment interval
			}
		} else if ( yi == nMaxY ) {		// horz case for top row
			if ( vLines[3].vLine.Direction().Y() == 0 ) {
				fMinX = vLines[3].vLine.Origin().X();
				fMaxX = vLines[3].vLine.Origin().X() + vLines[3].vLine.Direction().X();
			}
		}
	
		// no special case ocurred, find line isects
		if ( fMinX == -1.0f ) {
			float fYHeight = (float)yi;		// should we add 0.5f here ??

			// check if we need to increment interval
			if ( fYHeight > vLines[nFirstI].fYIntervalHigh ) {
				nSecondI++;
				nFirstI = nSecondI-1;
			}

			// ok, find 't' value for each line
			float fTFirst = (fYHeight - vLines[nFirstI].vLine.Origin().Y()) / vLines[nFirstI].vLine.Direction().Y();
			fMinX = vLines[nFirstI].vLine.Origin().X() + fTFirst * vLines[nFirstI].vLine.Direction().X();

			float fTSecond = (fYHeight - vLines[nSecondI].vLine.Origin().Y()) / vLines[nSecondI].vLine.Direction().Y();
			fMaxX = vLines[nSecondI].vLine.Origin().X() + fTSecond * vLines[nSecondI].vLine.Direction().X();
		}

		if ( fMaxX < fMinX ) {		// swap if order is wrong
			float fTmp = fMinX;
			fMinX = fMaxX;
			fMaxX = fTmp;
		}

		int nMinX = (int)ceilf( fMinX );
		int nMaxX = (int)floorf( fMaxX );
		for ( int xi = nMinX; xi <= nMaxX; ++xi ) {
			unsigned char * pPixel = GetPixel( xi, yi );
			for ( unsigned int j = 0; j < m_nChannels; ++j )
				pPixel[j] = Wml::ColorRGBA::ToByte(cColor[(int)j]);
		}
	}

}



void Image::FillQuadAA( const Wml::Vector2f vCorners[4], const Wml::ColorRGBA & cColor )
{
	if ( m_nChannels != 4 )		// require alpha channel for AA rendering
		return;

	// 4-bit line anti-aliasing table from Gupta Siggraph81 paper
	// first element is distance = 0/16, second is 1/16, etc. (24/16 == 0)
	static const float vGuptaTable[] = 
	{   0.780f, 0.775f, 0.760f, 0.736f, 0.703f, 
	    0.662f, 0.613f, 0.558f, 0.500f, 0.441f, 
		0.383f, 0.328f, 0.276f, 0.228f, 0.184f, 
		0.145f, 0.110f, 0.080f, 0.056f, 0.036f, 
		0.021f, 0.010f, 0.004f, 0.001f, 0.0f 
	};

	// 4-bit polygon anti-aliasing table from Gupta Siggraph81 paper
	// first element is distance = -16/16, last is distance 16/16
	static const float vGuptaTablePoly[] = 
	{   
		0.000f, 0.001f, 0.004f, 0.010f, 0.021f,
		0.036f, 0.056f, 0.080f, 0.110f, 0.145f,
		0.184f, 0.228f, 0.276f, 0.328f, 0.383f,
		0.441f, 0.500f, 0.559f, 0.617f, 0.672f,
		0.724f, 0.772f, 0.816f, 0.855f, 0.890f,
		0.920f, 0.944f, 0.964f, 0.979f, 0.990f,
		0.996f, 0.999f, 1.0f
	};

	// create list of lines (there are only 4...)
	LineInfo vLines[4];
	for ( int i = 0; i < 4; ++i ) {
		int iLow = i;
		int iHigh = (i+1)%4;
		bool bFlipped = false;
		if ( vCorners[iHigh].Y() < vCorners[iLow].Y() ) {
			iLow = (i+1)%4;
			iHigh = i;
			bFlipped = true;
		}

		vLines[i].vLine = Wml::Line2f( vCorners[iLow], vCorners[iHigh], true );

		// want to set normal as well because we're going to use the distance function later...
		vLines[i].vLine.Normal() = Wml::Vector2f(
			-vLines[i].vLine.Direction().Y(), vLines[i].vLine.Direction().X() );
		if ( bFlipped )
			vLines[i].vLine.Normal() *= -1.0f;
		vLines[i].vLine.Normal().Normalize();
		vLines[i].vLine.Constant() = vLines[i].vLine.Origin().Dot( vLines[i].vLine.Normal() );

		vLines[i].fYIntervalLow = vCorners[iLow].Y();
		vLines[i].fYIntervalHigh = vCorners[iHigh].Y();
	}


	// sort vLine by Y intervals (first by lower miny, then lower maxy)
	qsort( vLines, 4, sizeof(LineInfo), qsort_ysortLine );

	// draw from min to max y
	float fMinY = vLines[0].fYIntervalLow;
	float fMaxY = vLines[3].fYIntervalHigh;

	int nMinY = (int)ceilf(fMinY) - 1;	// these +1 bumps are for the antialiasing
	int nMaxY = (int)floorf(fMaxY) + 1;

	// handle straight lines at bottom and top
	bool bLowHorizontal = (vLines[0].vLine.Direction().Y() == 0);
	bool bHighHorizontal = ( vLines[3].vLine.Direction().Y() == 0 );

	int nFirstI = 0;
	int nSecondI = 1;
	for ( int yi = nMinY; yi <= nMaxY; ++yi ) {
		bool bIncremented = false;
		int nLastI;

		// figure out sides of run for current row
		float fMinX = -1.0f, fMaxX = 0.0f;

#if 1	// do horizontal-line special case tests
		if ( yi <= nMinY+1 && bLowHorizontal ) {
			fMinX = vLines[0].vLine.Origin().X();
			fMaxX = vLines[0].vLine.Origin().X() + vLines[0].vLine.Direction().X();
			if ( fMinX < fMaxX ) { fMinX -= 1; fMaxX += 1; } else { fMinX += 1; fMaxX -= 1; }
		} else if ( yi >= nMaxY-1 && bHighHorizontal ) {
			fMinX = vLines[3].vLine.Origin().X();
			fMaxX = vLines[3].vLine.Origin().X() + vLines[3].vLine.Direction().X();
			if ( fMinX < fMaxX ) { fMinX -= 1; fMaxX += 1; } else { fMinX += 1; fMaxX -= 1; }
		}
#endif

		// no special case ocurred, find line intersections
		if ( fMinX == -1.0f ) {
			float fYHeight = (float)yi;

			// check if we need to increment interval
			if ( (float)yi > vLines[nFirstI].fYIntervalHigh && nSecondI < 3) {
				nLastI = nFirstI;
				nSecondI++;
				nFirstI = nSecondI-1;
				bIncremented = true;
			}

			// ok, find 't' value for each line. we are offsetting by one pixel along the normal vector
			//  here so that the bounds come out right for anti-aliasing
			Wml::Line2f & vFirstLine = vLines[nFirstI].vLine;
			float fTFirst = (fYHeight - vFirstLine.Origin().Y() - vFirstLine.Normal().Y()) / vFirstLine.Direction().Y();
			fMinX = vFirstLine.Origin().X() + vFirstLine.Normal().X() + fTFirst * vFirstLine.Direction().X();

			Wml::Line2f & vSecondLine = vLines[nSecondI].vLine;
			float fTSecond = (fYHeight - vSecondLine.Origin().Y() - vSecondLine.Normal().Y()) / vSecondLine.Direction().Y();
			fMaxX = vSecondLine.Origin().X() + vSecondLine.Normal().X() + fTSecond * vSecondLine.Direction().X();
		}

		if ( fMaxX < fMinX ) {		// swap if order is wrong
			float fTmp = fMinX;
			fMinX = fMaxX;
			fMaxX = fTmp;
		}

		int nMinX = (int)ceilf( fMinX );
		int nMaxX = (int)floorf( fMaxX );
		for ( int xi = nMinX; xi <= nMaxX; ++xi ) {

			float fAlpha = 1.0f;

			// compute distance from lines that define the edges of the current run
			Wml::Vector2f vPoint((float)xi, (float)yi);
			float fDist1 = vLines[nFirstI].vLine.GetPseudodistance(vPoint);
			float fDist2 = vLines[nSecondI].vLine.GetPseudodistance(vPoint);
			float fMaxDist = std::max( fDist1, fDist2);

			// hacks for when we hit corners and actually need to test 3 edges...
			if ( (float)(yi+1) > vLines[nFirstI].fYIntervalHigh && nSecondI < 3) {
				float fDist3 = vLines[ nSecondI+1 ].vLine.GetPseudodistance(vPoint);
				fMaxDist = std::max( fMaxDist, fDist3 );
			}
			if ( bIncremented ) {
				float fDist3 = vLines[ nLastI ].vLine.GetPseudodistance(vPoint);
				fMaxDist = std::max( fMaxDist, fDist3 );
			}

			// use distance to compute anti-aliasing alpha value

#if 0		// line table
			if ( fMaxDist > 1.4)
				fAlpha = 0;
			else if ( fMaxDist > 0 )
				fAlpha = vGuptaTable[ (int)(fMaxDist * 16.0f ) ];

#else		// polygon table
			if ( fMaxDist > 1.0 )
				fAlpha =  0;
			else if ( fMaxDist > -1.0 ) {
				fMaxDist = -fMaxDist;
				fMaxDist += 1.0f;
				int nOffset = (int)(fMaxDist * 16.0f);
				if ( nOffset < 0 || nOffset > 32 )
					DebugBreak();
				fAlpha = vGuptaTablePoly[ nOffset ];
			}
#endif


			CompositePixel( xi, yi, cColor, fAlpha );
		}

	}

}


#define roundf(k) ( ((k)-floorf((k))) > 0.5f ? ceilf((k)) : floorf((k)) )

#define CircleComposite(cx,cy,nx,ny,c)   CompositePixel( (cx)+(nx), (cy)+(ny), c); \
										 CompositePixel( (cx)-(nx), (cy)+(ny), c); \
										 CompositePixel( (cx)-(nx), (cy)-(ny), c); \
										 CompositePixel( (cx)+(nx), (cy)-(ny), c); \
										 CompositePixel( (cx)+(ny), (cy)+(nx), c); \
										 CompositePixel( (cx)-(ny), (cy)+(nx), c); \
										 CompositePixel( (cx)-(ny), (cy)-(nx), c); \
										 CompositePixel( (cx)+(ny), (cy)-(nx), c);

#define CircleComposite2(cx,cy,nx,ny,c)  CompositePixel( (cx)+(nx), (cy)+(ny), c); \
										 CompositePixel( (cx)-(nx), (cy)-(ny), c); \
										 CompositePixel( (cx)+(ny), (cy)+(nx), c); \
										 CompositePixel( (cx)-(ny), (cy)-(nx), c); 

#define CircleComposite3(cx,cy,nx,ny,c)  CompositePixel( (cx)+(nx), (cy)+(ny), c); \
										 CompositePixel( (cx)-(nx), (cy)+(ny), c); \
										 CompositePixel( (cx)-(nx), (cy)-(ny), c); \
										 CompositePixel( (cx)+(nx), (cy)-(ny), c); 

void Image::DrawCircle( float fCenterX, float fCenterY, float fRadius, const Wml::ColorRGBA & cColor )
{
	int nCenterX = (int)roundf(fCenterX);
	int nCenterY = (int)roundf(fCenterY);
	int nRadius = (int)roundf(fRadius);

	int nX = 0;
	int nY = nRadius;
	int nD = 1 - nRadius;
	int nDeltaE = 3;
	int nDeltaSE = 5 - 2*nRadius;

	// plot first point
	CircleComposite2( nCenterX, nCenterY, nX, nY, cColor );

	while (nY > nX) {

		if ( nD < 0 ) {
			nD += nDeltaE;
			nDeltaE += 2;
			nDeltaSE += 2;
		} else {
			nD += nDeltaSE;
			nDeltaE += 2;
			nDeltaSE += 4;
			nY--;
		}
		nX++;

		if  ( nX == nY ) {
			CircleComposite3( nCenterX, nCenterY, nX, nY, cColor );
		} else if ( nX < nY ) {
			CircleComposite( nCenterX, nCenterY, nX, nY, cColor );
		}
	}
}



void Image::FillRow( int nXLow, int nXHigh, int nY, const Wml::ColorRGBA & cColor )
{
	for ( int i = nXLow; i <= nXHigh; ++i ) {
		CompositePixel( i, nY, cColor );
	}	
}

void Image::DrawDisc( float fCenterX, float fCenterY, float fRadius, const Wml::ColorRGBA & cColor )
{
	int nCenterX = (int)roundf(fCenterX);
	int nCenterY = (int)roundf(fCenterY);
	int nRadius = (int)roundf(fRadius);

	int nX = 0;
	int nY = nRadius;
	int nD = 1 - nRadius;
	int nDeltaE = 3;
	int nDeltaSE = 5 - 2*nRadius;

	while (nY > nX) {

		if ( nD < 0 ) {
			nD += nDeltaE;
			nDeltaE += 2;
			nDeltaSE += 2;

			// draw inner rows
			FillRow( nCenterX - nY, nCenterX + nY, nCenterY + nX, cColor );
			if  ( nX > 0 ) 
				FillRow( nCenterX - nY, nCenterX + nY, nCenterY - nX, cColor );

		} else {
			nD += nDeltaSE;
			nDeltaE += 2;
			nDeltaSE += 4;

			// draw outer rows
			FillRow( nCenterX - nX, nCenterX + nX, nCenterY + nY, cColor );
			FillRow( nCenterX - nX, nCenterX + nX, nCenterY - nY, cColor );

			// draw inner rows
			FillRow( nCenterX - nY, nCenterX + nY, nCenterY + nX, cColor );
			if  ( nX > 0 ) 
				FillRow( nCenterX - nY, nCenterX + nY, nCenterY - nX, cColor );

			nY--;
		}

		nX++;
	}

	if ( nY == nX ) {
		// draw outer rows
		FillRow( nCenterX - nX, nCenterX + nX, nCenterY + nY, cColor );
		FillRow( nCenterX - nX, nCenterX + nX, nCenterY - nY, cColor );
	}
}



void Image::FillAADiscRow( int nXLow, int nXHigh, int nPadX, int nY, const Wml::ColorRGBA & cColor,
						   float fX, float fY, float fRadius )
{
	// 4-bit polygon anti-aliasing table from Gupta Siggraph81 paper
	// first element is distance = -16/16, last is distance 16/16
	static const float vGuptaTablePoly[] = 
	{   
		0.000f, 0.001f, 0.004f, 0.010f, 0.021f,
		0.036f, 0.056f, 0.080f, 0.110f, 0.145f,
		0.184f, 0.228f, 0.276f, 0.328f, 0.383f,
		0.441f, 0.500f, 0.559f, 0.617f, 0.672f,
		0.724f, 0.772f, 0.816f, 0.855f, 0.890f,
		0.920f, 0.944f, 0.964f, 0.979f, 0.990f,
		0.996f, 0.999f, 1.0f
	};


	nXLow -= nPadX;
	nXHigh += nPadX;
	
	float fLineY = (float)nY;
	float fRadSqrLow = (fRadius - 1.0f)*(fRadius - 1.0f);
	float fRadSqrHigh = (fRadius + 1.0f)*(fRadius + 1.0f);
	
	for ( int i = nXLow; i <= nXHigh; ++i ) {

		Wml::Vector2f vPoint( (float)i, fLineY );
		float fDistSqr = (vPoint.X() - fX)*(vPoint.X() - fX) + (vPoint.Y() - fY)*(vPoint.Y() - fY);

		if ( fDistSqr < fRadSqrLow ) {
			CompositePixel( i, nY, cColor );
		} else if ( fDistSqr < fRadSqrHigh ) {

			// find real dist
			float fDist = sqrtf(fDistSqr) - fRadius;
			if ( fDist < 1.0  ) {
				fDist = -fDist;
				fDist += 1.0f;
				int nOffset = (int)(fDist * 16.0f);
				if ( nOffset < 0 || nOffset > 32 )
					DebugBreak();
				float fAlpha = vGuptaTablePoly[ nOffset ];
				CompositePixel( i, nY, cColor, fAlpha );
			}

		}
	}	
}


void Image::DrawDiscAA( float fCenterX, float fCenterY, float fRadius, const Wml::ColorRGBA & cColor )
{
	int nCenterX = (int)roundf(fCenterX);
	int nCenterY = (int)roundf(fCenterY);
	int nRadius = (int)roundf(fRadius);

	int nLastX = 0;
	int nX = 0;
	int nY = nRadius;
	int nD = 1 - nRadius;
	int nDeltaE = 3;
	int nDeltaSE = 5 - 2*nRadius;

	while (nY > nX) {

		if ( nD < 0 ) {
			nD += nDeltaE;
			nDeltaE += 2;
			nDeltaSE += 2;

			// draw inner rows
			FillAADiscRow( nCenterX - nY, nCenterX + nY, 1, nCenterY + nX, cColor, fCenterX, fCenterY, fRadius );
			if  ( nX > 0 ) 
				FillAADiscRow( nCenterX - nY, nCenterX + nY, 1, nCenterY - nX, cColor, fCenterX, fCenterY, fRadius );


		} else {
			nD += nDeltaSE;
			nDeltaE += 2;
			nDeltaSE += 4;

			// draw outer rows
			FillAADiscRow( nCenterX - nX, nCenterX + nX, 1, nCenterY + nY, cColor, fCenterX, fCenterY, fRadius );
			FillAADiscRow( nCenterX - nX, nCenterX + nX, 1, nCenterY - nY, cColor, fCenterX, fCenterY, fRadius );

			// draw inner rows
			FillAADiscRow( nCenterX - nY, nCenterX + nY, 1, nCenterY + nX, cColor, fCenterX, fCenterY, fRadius );
			if  ( nX > 0 ) 
				FillAADiscRow( nCenterX - nY, nCenterX + nY, 1, nCenterY - nX, cColor, fCenterX, fCenterY, fRadius );

			nY--;
			nLastX = nX;
		}

		nX++;
	}

	if ( nY == nX ) {
		// draw outer rows
		FillAADiscRow( nCenterX - nX, nCenterX + nX, 1, nCenterY + nY, cColor, fCenterX, fCenterY, fRadius );
		FillAADiscRow( nCenterX - nX, nCenterX + nX, 1, nCenterY - nY, cColor, fCenterX, fCenterY, fRadius );

	}
}



/***
The following code between the comments marked WILLEM_VAN_SCHAIK is distributed under this license:

// Copyright 2000, Willem van Schaik.  For conditions of distribution and
// use, see the copyright/license/disclaimer notice in png.h

****/

/* begin WILLEM_VAN_SCHAIK */

#define PNG_NO_STDIO

#include <cexcept.h>

define_exception_type(const char *);
extern struct exception_context the_exception_context[1];
struct exception_context the_exception_context[1];
png_const_charp msg;

static png_structp png_ptr = NULL;
static png_infop info_ptr = NULL;

// cexcept interface

static void
png_cexcept_error(png_structp png_ptr, png_const_charp msg)
{
   //if(png_ptr)
   //  ;
   {
      Throw msg;
   }
}

#ifdef PNG_NO_STDIO

static void
png_read_data(png_structp png_ptr, png_bytep data, png_size_t length)
{
   png_size_t check;

   /* fread() returns 0 on error, so it is OK to store this in a png_size_t
    * instead of an int, which is what fread() actually returns.
    */
   check = (png_size_t)fread(data, (png_size_t)1, length,
      (FILE *)png_ptr->io_ptr);

   if (check != length)
   {
      png_error(png_ptr, "Read Error");
   }
}

static void
png_write_data(png_structp png_ptr, png_bytep data, png_size_t length)
{
   png_uint_32 check;

   check = (png_uint_32)fwrite(data, 1, length, (FILE *)(png_ptr->io_ptr));
   if (check != length)
   {
      png_error(png_ptr, "Write Error");
   }
}

static void
png_flush(png_structp png_ptr)
{
   FILE *io_ptr;
   io_ptr = (FILE *)CVT_PTR((png_ptr->io_ptr));
   if (io_ptr != NULL)
      fflush(io_ptr);
}

#endif


// PNG image handler functions

BOOL PngLoadImage (const char * pstrFileName, png_byte **ppbImageData,
                   unsigned int *piWidth, unsigned int *piHeight, int *piChannels, png_color *pBkgColor)
{
    static FILE        *pfFile;
    png_byte            pbSig[8];
    int                 iBitDepth;
    int                 iColorType;
    double              dGamma;
//    png_color_16       *pBackground;
    png_uint_32         ulChannels;
    png_uint_32         ulRowBytes;
    png_byte           *pbImageData = *ppbImageData;
    static png_byte   **ppbRowPointers = NULL;
    int                 i;

    // open the PNG input file

    if (!pstrFileName)
    {
        *ppbImageData = pbImageData = NULL;
        return FALSE;
    }

    if (!(pfFile = fopen(pstrFileName, "rb")))
    {
        *ppbImageData = pbImageData = NULL;
        return FALSE;
    }

    // first check the eight byte PNG signature

    fread(pbSig, 1, 8, pfFile);
    if (!png_check_sig(pbSig, 8))
    {
        *ppbImageData = pbImageData = NULL;
        return FALSE;
    }

    // create the two png(-info) structures

    png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL,
      (png_error_ptr)png_cexcept_error, (png_error_ptr)NULL);
    if (!png_ptr)
    {
        *ppbImageData = pbImageData = NULL;
        return FALSE;
    }

    info_ptr = png_create_info_struct(png_ptr);
    if (!info_ptr)
    {
        png_destroy_read_struct(&png_ptr, NULL, NULL);
        *ppbImageData = pbImageData = NULL;
        return FALSE;
    }

    Try
    {
        // initialize the png structure
#if !defined(PNG_NO_STDIO)
        png_init_io(png_ptr, pfFile);
#else
        png_set_read_fn(png_ptr, (png_voidp)pfFile, png_read_data);
#endif
        
        png_set_sig_bytes(png_ptr, 8);
        
        // read all PNG info up to image data
        
        png_read_info(png_ptr, info_ptr);
        
        // get width, height, bit-depth and color-type
        
        png_get_IHDR(png_ptr, info_ptr, (png_uint_32 *)piWidth, (png_uint_32 *)piHeight, &iBitDepth,
            &iColorType, NULL, NULL, NULL);
        
        // expand images of all color-type and bit-depth to 3x8 bit RGB images
        // let the library process things like alpha, transparency, background
        
        if (iBitDepth == 16)
            png_set_strip_16(png_ptr);
        if (iColorType == PNG_COLOR_TYPE_PALETTE)
            png_set_expand(png_ptr);
        if (iBitDepth < 8)
            png_set_expand(png_ptr);
        if (png_get_valid(png_ptr, info_ptr, PNG_INFO_tRNS))
            png_set_expand(png_ptr);
        if (iColorType == PNG_COLOR_TYPE_GRAY ||
            iColorType == PNG_COLOR_TYPE_GRAY_ALPHA)
            png_set_gray_to_rgb(png_ptr);
        
        // set the background color to draw transparent and alpha images over.
        //if (png_get_bKGD(png_ptr, info_ptr, &pBackground))
        //{
        //    png_set_background(png_ptr, pBackground, PNG_BACKGROUND_GAMMA_FILE, 1, 1.0);
        //    pBkgColor->red   = (byte) pBackground->red;
        //    pBkgColor->green = (byte) pBackground->green;
        //    pBkgColor->blue  = (byte) pBackground->blue;
        //}
        //else
        //{
        //    pBkgColor = NULL;
        //}
		pBkgColor = NULL;

        // if required set gamma conversion
        if (png_get_gAMA(png_ptr, info_ptr, &dGamma))
            png_set_gamma(png_ptr, (double) 2.2, dGamma);
        
        // after the transformations have been registered update info_ptr data
        
        png_read_update_info(png_ptr, info_ptr);
        
        // get again width, height and the new bit-depth and color-type
        
        png_get_IHDR(png_ptr, info_ptr, (png_uint_32 *)piWidth, (png_uint_32 *)piHeight, &iBitDepth,
            &iColorType, NULL, NULL, NULL);
        
        
        // row_bytes is the width x number of channels
        
        ulRowBytes = png_get_rowbytes(png_ptr, info_ptr);
        ulChannels = png_get_channels(png_ptr, info_ptr);
        
        *piChannels = ulChannels;
        
        // now we can allocate memory to store the image
        
        if (pbImageData)
        {
            free (pbImageData);
            pbImageData = NULL;
        }
        if ((pbImageData = (png_byte *) malloc(ulRowBytes * (*piHeight)
                            * sizeof(png_byte))) == NULL)
        {
            png_error(png_ptr, "Visual PNG: out of memory");
        }
        *ppbImageData = pbImageData;
        
        // and allocate memory for an array of row-pointers
        
        if ((ppbRowPointers = (png_bytepp) malloc((*piHeight)
                            * sizeof(png_bytep))) == NULL)
        {
            png_error(png_ptr, "Visual PNG: out of memory");
        }
        
        // set the individual row-pointers to point at the correct offsets
        
        for (i = 0; i < (int)(*piHeight); i++)
            ppbRowPointers[i] = pbImageData + i * ulRowBytes;
        
        // now we can go ahead and just read the whole image
        
        png_read_image(png_ptr, ppbRowPointers);
        
        // read the additional chunks in the PNG file (not really needed)
        
        png_read_end(png_ptr, NULL);
        
        // and we're done
        
        free (ppbRowPointers);
        ppbRowPointers = NULL;
        
        // yepp, done
    }

    Catch (msg)
    {
        png_destroy_read_struct(&png_ptr, &info_ptr, NULL);

        *ppbImageData = pbImageData = NULL;
        
        if(ppbRowPointers)
            free (ppbRowPointers);

        fclose(pfFile);

        return FALSE;
    }

    fclose (pfFile);

    return TRUE;
}


BOOL PngSaveImage (PTSTR pstrFileName, png_byte *pDiData,
                   int iWidth, int iHeight, png_color bkgColor)
{
    const int           ciBitDepth = 8;
    const int           ciChannels = 3;

    static FILE        *pfFile;
    png_uint_32         ulRowBytes;
    static png_byte   **ppbRowPointers = NULL;
    int                 i;

    // open the PNG output file

    if (!pstrFileName)
        return FALSE;

    if (!(pfFile = fopen(pstrFileName, "wb")))
        return FALSE;

    // prepare the standard PNG structures

    png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL,
      (png_error_ptr)png_cexcept_error, (png_error_ptr)NULL);
    if (!png_ptr)
    {
        fclose(pfFile);
        return FALSE;
    }

    info_ptr = png_create_info_struct(png_ptr);
    if (!info_ptr) {
        fclose(pfFile);
        png_destroy_write_struct(&png_ptr, (png_infopp) NULL);
        return FALSE;
    }

    Try
    {
        // initialize the png structure
        
#if !defined(PNG_NO_STDIO)
        png_init_io(png_ptr, pfFile);
#else
        png_set_write_fn(png_ptr, (png_voidp)pfFile, png_write_data, png_flush);
#endif
        
        // we're going to write a very simple 3x8 bit RGB image
        
        png_set_IHDR(png_ptr, info_ptr, iWidth, iHeight, ciBitDepth,
            PNG_COLOR_TYPE_RGB, PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_BASE,
            PNG_FILTER_TYPE_BASE);
        
        // write the file header information
        
        png_write_info(png_ptr, info_ptr);
        
        // swap the BGR pixels in the DiData structure to RGB
        
        png_set_bgr(png_ptr);
        
        // row_bytes is the width x number of channels
        
        ulRowBytes = iWidth * ciChannels;
        
        // we can allocate memory for an array of row-pointers
        
        if ((ppbRowPointers = (png_bytepp) malloc(iHeight * sizeof(png_bytep))) == NULL)
            Throw "Visualpng: Out of memory";
        
        // set the individual row-pointers to point at the correct offsets
        
        for (i = 0; i < iHeight; i++)
            ppbRowPointers[i] = pDiData + i * (((ulRowBytes + 3) >> 2) << 2);
        
        // write out the entire image data in one call
        
        png_write_image (png_ptr, ppbRowPointers);
        
        // write the additional chunks to the PNG file (not really needed)
        
        png_write_end(png_ptr, info_ptr);
        
        // and we're done
        
        free (ppbRowPointers);
        ppbRowPointers = NULL;
        
        // clean up after the write, and free any memory allocated
        
        png_destroy_write_struct(&png_ptr, (png_infopp) NULL);
        
        // yepp, done
    }

    Catch (msg)
    {
        png_destroy_write_struct(&png_ptr, (png_infopp) NULL);

        if(ppbRowPointers)
            free (ppbRowPointers);

        fclose(pfFile);

        return FALSE;
    }
    
    fclose (pfFile);
    
    return TRUE;
}




/* end WILLEM_VAN_SCHAIK */

