// Ryan Schmidt - rms@unknownroad.com
// Copyright (c) 2005-2010
// Distributed under the Boost Software License, Version 1.0.
// http://www.boost.org/LICENSE_1_0.txt


#pragma once

#include <vector>
#include <WmlSerializer.h>

namespace rmsui {

class IParameter {
public:
	virtual ~IParameter() {};

	enum ParameterType {
		IntegerParameter,
		FloatParameter,
		StringParameter,
		FileParameter
	};

	virtual const CString & GetName() = 0;

	virtual ParameterType GetType() = 0;

	virtual int GetValueInteger() = 0;
	virtual float GetValueFloat() = 0;
	virtual void GetValueString( CString & string ) = 0;
	virtual void GetValueFile( CString & string) = 0;

	virtual void SetValueInteger( int nValue ) = 0;
	virtual void SetValueFloat( float fValue ) = 0;
	virtual void SetValueString( const CString & string ) = 0;
	virtual void SetValueFile( const CString & filename ) = 0;

	virtual void GetValue( CString & value ) = 0;
	virtual void SetValue( const CString & string ) = 0;

	virtual void SynchronizeParameter() = 0;

	/*
	 * serialization!
	 */
	virtual void Store( Wml::SerializerRecord * pRecord ) = 0;
	virtual void Restore( Wml::SerializerRecord * pRecord ) = 0;

	//! this hack prevents keyframing for this parameter. Necessary for "aggregate" parameters like the "position" of a composition...
	virtual bool GetNoKeyFrame() { return false; }
};



class ParameterList {
public:
	ParameterList() {}
	~ParameterList() {};

	void AddParameter( IParameter * pParam ) 
		{ m_vParams.push_back(pParam); }
	void RemoveAllParameters() 
		{ m_vParams.clear(); }

	const std::vector<IParameter *> & GetParameters() const
		{ return m_vParams; } 
	IParameter * GetParameter( unsigned int nParam ) const 
		{ return (nParam < m_vParams.size()) ? m_vParams[nParam] : NULL ; }

	IParameter * GetParameter( const char * pName ) const;

	/*
	 * serialization!
	 */
	virtual void Store( Wml::SerializerRecord * pRecord );
	virtual void Restore( Wml::SerializerRecord * pRecord );

protected:
	std::vector<IParameter *> m_vParams;
};


class IParameterTarget
{
public:
	virtual ~IParameterTarget() {};

	virtual const ParameterList & GetParameterSet() = 0;
};



class GenericParameter : public IParameter 
{
public:
	GenericParameter( const CString & paramName, ParameterType paramType );
	virtual ~GenericParameter() {};

	virtual const CString & GetName() { return m_paramName; }
	virtual ParameterType GetType() { return m_paramType; }

	virtual int GetValueInteger() { return 0; }
	virtual float GetValueFloat() { return 0.0f; }
	virtual void GetValueString( CString & string ) { string = CString(""); }
	virtual void GetValueFile( CString & string) { string = CString(""); }

	virtual void SetValueInteger( int nValue ) {}
	virtual void SetValueFloat( float fValue ) {}
	virtual void SetValueString( const CString & string ) {}
	virtual void SetValueFile( const CString & filename ) {}

	virtual void GetValue( CString & value );
	virtual void SetValue( const CString & string ){}

	//! this hack prevents keyframing for this parameter. Necessary for "aggregate" parameters like the "position" of a composition...
	virtual bool GetNoKeyFrame() { return m_bNoKeyFrame; }
	void SetNoKeyFrame( bool bNoKeyframe ) { m_bNoKeyFrame = bNoKeyframe; }

protected:
	CString m_paramName;
	ParameterType m_paramType;
	bool m_bNoKeyFrame;
};


class GenericFloatParameter : public GenericParameter 
{
public:
	GenericFloatParameter( const CString & paramName, float fInitialValue = 0.0 );
	virtual ~GenericFloatParameter() {};

	virtual float GetValueFloat() { return GetParameter(); }
	virtual void SetValueFloat( float fValue );

	virtual void GetValue( CString & value );
	virtual void SetValue( const CString & string );

	virtual void SynchronizeParameter();

	virtual float GetMinValue();

	/*
	 * serialization!
	 */
	virtual void Store( Wml::SerializerRecord * pRecord ) 
		{ pRecord->SetValue( (LPCSTR)GetName(), GetValueFloat() ); }
	virtual void Restore( Wml::SerializerRecord * pRecord )
		{ float fValue;  if ( pRecord->GetValue( (LPCSTR)GetName(), fValue, GetValueFloat() ) )  SetValueFloat(fValue); }

protected:
	float m_fParamValue;

	// override this if you need more complex validation than just min/max
	virtual bool Validate(float fValue);

	// these are the three functions that probably need to be implemented...
	virtual void GetMinMax( float & fMin, float & fMax );
	virtual void SendParameter( float fNewParameter ) = 0;
	virtual float GetParameter( ) = 0; 
};

class GenericIntegerParameter : public GenericParameter 
{
public:
	GenericIntegerParameter( const CString & paramName, int nInitialValue = 0 );
	virtual ~GenericIntegerParameter() {};

	virtual int GetValueInteger() { return GetParameter(); }
	virtual void SetValueInteger( int nValue );

	virtual void GetValue( CString & value );
	virtual void SetValue( const CString & string );

	virtual void SynchronizeParameter();

	/*
	 * serialization!
	 */
	virtual void Store( Wml::SerializerRecord * pRecord ) 
		{ pRecord->SetValue( (LPCSTR)GetName(), GetValueInteger() ); }
	virtual void Restore( Wml::SerializerRecord * pRecord )
		{ int nValue;  if ( pRecord->GetValue( (LPCSTR)GetName(), nValue, GetValueInteger()) )  SetValueInteger(nValue); }


protected:
	int m_nParamValue;

	// override this if you need more complex validation than just min/max
	virtual bool Validate(int nValue);

	// these are the three functions that probably need to be implemented...
	virtual void GetMinMax( int & nMin, int & nMax );
	virtual void SendParameter( int nNewParameter ) = 0;
	virtual int GetParameter( ) = 0; 
};


}  // end namespace rmsui