// Ryan Schmidt - rms@unknownroad.com
// Copyright (c) 2005-2010
// Distributed under the Boost Software License, Version 1.0.
// http://www.boost.org/LICENSE_1_0.txt


#ifndef _RMSUI_TOOL_H
#define _RMSUI_TOOL_H

#include <WmlRay3.h>
#include <Frame.h>
#include <string>
#include <set>
#include <vector>

#include  "ToolSet.h"


namespace rmsui
{


/**
 * interface that objects need to implement to be linked to a tool
 */
class ITool3DTarget
{
public:
	//! get the frame for this target
	virtual Wml::Frame GetFrame() = 0;

	//! translate the frame
	virtual void Translate( const Wml::Vector3f & vTranslate, bool bRelative = true ) = 0;

	//! rotate the frame
	virtual void Rotate( const Wml::Matrix3f & rotation ) = 0;

	//! test if target supports a tool
	virtual bool SupportsTool( const std::string & sToolName ) = 0;
};


/*
 * interface for objects that want to know when the tool did something
 */
class ITool3D;
class ITool3DListener
{
public:
	//! called when tool action occurs. Passes self to identify tool...
	virtual void OnToolActionApplied( ITool3D * pTool ) = 0;
};


// this is all double because gluProject / gluUnProject uses doubles. If I
//  write my own I can use floats...
class ViewState
{
public:
	unsigned int nViewport[4];

	double dEyeView_ProjectionMatrix[16];
	double dEyeView_ModelViewMatrix[16];

	double fFixedView_NearZ;
	double dFixedView_ProjectionMatrix[16];
	double dFixedView_ModelViewMatrix[16];

	float fCurrentTool_LookAt[4][4];
};


class SelectedTargetInfo
{
public:

	//! interaction point from previous frame
	Wml::Vector3f ptLastFrame;

	//! current eye view ray
	Wml::Ray3f eyeViewRay;

	//! current oriented fixed view ray
	Wml::Ray3f fixedViewOrientedRay;
};


/**
 * interface class for tool objects
 */
class ITool3D
{
public:
	ITool3D( const std::string & sToolName )
		: m_pTarget(NULL), m_sToolName(sToolName)
	{ }

	const std::string & GetToolName() const { return m_sToolName; }
	ToolSet::ToolType GetToolType() const { return m_nToolType; }

/*
 * virtual interface functions subclasses must implement
 */
	virtual void Render( const ViewState & viewState ) = 0;

	virtual bool Intersect( const Wml::Ray3f & ray, 
							Wml::Vector3f & ptPosition, float & fDepth ) = 0;

	virtual Wml::Vector3f Move( SelectedTargetInfo & targetInfo,
								const ViewState & viewState,
								bool bInitialize = false ) = 0;


	//! subclasses probably want to implement this function as well...
	virtual void TargetFrameChanged()
	{	
		m_frame = m_pTarget->GetFrame();
	}

/*
 * target functions
 */
	ITool3DTarget * GetTarget()
		{ return m_pTarget; }
		
	void SetTarget( ITool3DTarget * pTarget )
	{
		m_pTarget = pTarget;

		// notify tool that target frame changed
		TargetFrameChanged();
	}


/*
 * frame functions
 */
	const Wml::Vector3f & GetOrigin() const
		{ return m_frame.Origin(); }

	Wml::Vector3f GetAxis( Wml::Frame::FrameAxis eAxis ) const
		{ return m_frame.Axis( eAxis ); }

	const Wml::Frame & GetFrame() const
		{ return m_frame; }

	void TranslateFrame( const Wml::Vector3f & vTranslate, bool bRelative = true )
		{ m_pTarget->Translate(vTranslate, bRelative); NotifyToolActionOccured(); }

	void RotateFrame( const Wml::Matrix3f & rotation )
		{ m_pTarget->Rotate(rotation); NotifyToolActionOccured(); }


/*
 * event handling stuff
 */
	void RegisterListener( ITool3DListener * pListener );

protected:
	// tool ID
	ToolSet::ToolType m_nToolType;
	void SetToolType( ToolSet::ToolType eType ) { m_nToolType = eType; }
	friend class ToolSet;		// required to set tooltype

	std::string m_sToolName;

	//! target that tool acts on
	ITool3DTarget * m_pTarget;

	//! 3D tool frame-of-reference, initialized from target
	Wml::Frame m_frame;

	//! list of tool listeners
	std::set<ITool3DListener *> m_vListeners;

	void NotifyToolActionOccured();
};



class ToolTargetGroup : public ITool3DTarget
{
public:
	ToolTargetGroup() {}

	void AddTarget( ITool3DTarget * pTarget ) { m_vTargets.push_back(pTarget); }

	void SetToolTargets( const std::vector<ITool3DTarget *> & vToolTargets ) { m_vTargets = vToolTargets; }

	void Clear() { m_vTargets.clear(); }

	unsigned int Count() { return (unsigned int)m_vTargets.size(); }

	ITool3DTarget * GetTarget( unsigned int nIndex ) { return m_vTargets[nIndex]; }

	/*
	 * IToolTarget3D interface
	 */

	//! get the frame for this target
	virtual Wml::Frame GetFrame();

	//! translate the frame
	virtual void Translate( const Wml::Vector3f & vTranslate, bool bRelative = true );

	//! rotate the frame
	virtual void Rotate( const Wml::Matrix3f & rotation );

	//! test if target supports a tool
	virtual bool SupportsTool( const std::string & sToolName );

protected:
	std::vector<ITool3DTarget *> m_vTargets;
};



} // namespace rmsui

#endif // _RMSUI_TOOL_H