// Ryan Schmidt - rms@unknownroad.com
// Copyright (c) 2005-2010
// Distributed under the Boost Software License, Version 1.0.
// http://www.boost.org/LICENSE_1_0.txt


#include "StdAfx.h"
#include <gl/GL.h>
#include <gl/GLU.h>
#include <WmlDistLin3Lin3.h>

#include "Tool3DTranslation.h"
#include "rmsdebug.h"

using namespace rmsui;

LocalTranslationTool3D::LocalTranslationTool3D()
	: ITool3D("Translation3D")
{
	float fLength = 75.0;
	m_axisTargets[0] = Sphere( GetAxis(Wml::Frame::AxisX) * fLength, 15.0f );
	m_axisTargets[1] = Sphere( GetAxis(Wml::Frame::AxisY) * fLength, 15.0f );
	m_axisTargets[2] = Sphere( GetAxis(Wml::Frame::AxisZ) * fLength, 15.0f );
	m_planeTarget = Circle( Wml::Vector3f::ZERO, Wml::Vector3f::UNIT_Z, 15.0f );
	m_planeTargetVolume = Sphere( Wml::Vector3f::ZERO, 15.0f );
	m_nSelected = 0;
}

void LocalTranslationTool3D::TargetFrameChanged()
{	
	ITool3D::TargetFrameChanged();

	float fLength = 75.0;
	m_axisTargets[0].Geometry().Center() = GetAxis(Wml::Frame::AxisX) * fLength;
	m_axisTargets[1].Geometry().Center() = GetAxis(Wml::Frame::AxisY) * fLength;
	m_axisTargets[2].Geometry().Center() = GetAxis(Wml::Frame::AxisZ) * fLength;
}


void LocalTranslationTool3D::Render( const ViewState & viewState ) 
{
	// render axis lines
	float fLength = 75.0;
	glPushAttrib(GL_ENABLE_BIT);
	glDisable(GL_LIGHTING);
	glPointSize(5.0);
	glBegin(GL_POINTS);
	glColor3f(1.0, 1.0, 1.0);
	glVertex3f(0.0, 0.0, 0.0);
	glEnd();
	glBegin(GL_LINES);
	glColor3f(1.0, 0.0, 0.0);
	glVertex3f(0.0, 0.0, 0.0);
	Wml::Vector3f axis( GetAxis(Wml::Frame::AxisX) * fLength );
	glVertex3f(axis.X(), axis.Y(), axis.Z());
	glColor3f(0.0, 1.0, 0.0);
	glVertex3f(0.0, 0.0, 0.0);
	axis = GetAxis(Wml::Frame::AxisY) * fLength;
	glVertex3f(axis.X(), axis.Y(), axis.Z());
	glColor3f(0.0, 0.0, 1.0);
	glVertex3f(0.0, 0.0, 0.0);
	axis = GetAxis(Wml::Frame::AxisZ) * fLength;
	glVertex3f(axis.X(), axis.Y(), axis.Z());
	glEnd();

	// render view-facing circle that contains tumble target
	// need to remove lookat for this... (this is sort of hacky)
	// then orient in lookat direction (??)
	glPushAttrib(GL_ENABLE_BIT);
	glPopMatrix();
	glPushMatrix();
	m_planeTarget.Render( Wml::ColorRGB(0.2f,0.2f,0.2f) );
	glMultMatrixf( &viewState.fCurrentTool_LookAt[0][0] );
	glPopAttrib();

	glPopAttrib();

	for (int i = 0; i < 3; ++i) {
		if (i == m_nSelected )
			m_axisTargets[i].Render( Wml::ColorRGBA::RED );
		else
			m_axisTargets[i].Render( Wml::ColorRGBA::BLUE );
	}
}


bool LocalTranslationTool3D::Intersect( const Wml::Ray3f & ray, 
									    Wml::Vector3f & ptPosition, float & fDepth )
{
	float fNearestDepth = Wml::Mathf::MAX_REAL;
	Wml::Vector3f ptTemp;
	float fTempDepth;

	// try axis targets
	for (int i = 0; i < 3; ++i) {
		if (m_axisTargets[i].Intersect(ray, ptTemp, fTempDepth) && fTempDepth < fNearestDepth) {
			fNearestDepth = fTempDepth;
			ptPosition = ptTemp;
			m_nSelected = i;
		}
	}

	// try center target
	if ( m_planeTargetVolume.Intersect(ray, ptTemp, fTempDepth) && fTempDepth < fNearestDepth ) {
		fNearestDepth = fTempDepth;
		ptPosition = ptTemp;
		m_nSelected = 4;
	}

	if (fNearestDepth != Wml::Mathf::MAX_REAL) {
		fDepth = fNearestDepth;
		return true;
	} else {
		return false;
	}
}


Wml::Vector3f LocalTranslationTool3D::Move( SelectedTargetInfo & targetInfo,
										    const ViewState & viewState,
										    bool bInitialize )
{
	Wml::Vector3f ptClosest;
	if ( m_nSelected < 4 ) {
		// find closest point between ray and selected axis
		Wml::Line3f axisLine;
		axisLine.Origin() = Wml::Vector3f::ZERO;
		axisLine.Direction() = GetAxis( (Wml::Frame::FrameAxis)m_nSelected );
		axisLine.Direction().Normalize();

		// compute closest point on axis to ray
		float fLineCoeff, fRayCoeff;
		Wml::SqrDistance(axisLine, targetInfo.eyeViewRay, &fLineCoeff, &fRayCoeff);
		ptClosest = Wml::Vector3f( axisLine.Origin() + axisLine.Direction() * fLineCoeff );
	} else {

		// compute matrix that aligns Z axis with oriented ray direction
		Wml::Matrix3f matrix;
		Wml::Ray3f eyeRay( targetInfo.eyeViewRay );
		Wml::ComputeAlignZAxisMatrix( eyeRay.Direction(), matrix );

		Wml::Vector3f vAxis( matrix * Wml::Vector3f::UNIT_Z );
		vAxis.Normalize();

		// find intersection point between ray and selected plane
		const Wml::Vector3f & vPlaneNormal = vAxis;
		float fPlaneD = vPlaneNormal.Dot( GetOrigin() );

		float fDdotN = vPlaneNormal.Dot( eyeRay.Direction() );
		if (fabs(fDdotN) < Wml::Mathf::EPSILON)
			ptClosest = GetOrigin();
		else {
			// compute intersection point w/ plane
			float fT = (fPlaneD - vPlaneNormal.Dot( eyeRay.Origin()) ) / fDdotN;
			ptClosest = Wml::Vector3f( eyeRay.Origin() + eyeRay.Direction() * fT );
		}
	}

	Wml::Vector3f newPosition(GetOrigin());
	if (!bInitialize) {
		// compute distance between current projected position and last projected position
		Wml::Vector3f translate( ptClosest - targetInfo.ptLastFrame );

		// apply translation to object
		TranslateFrame(translate, true);
	}

	// update last projection (return value)
	targetInfo.ptLastFrame = ptClosest;
	
	return newPosition;
}
