// Ryan Schmidt - rms@unknownroad.com
// Copyright (c) 2005-2010
// Distributed under the Boost Software License, Version 1.0.
// http://www.boost.org/LICENSE_1_0.txt


#include "StdAfx.h"
#include <gl/GL.h>
#include <gl/GLU.h>

#include "ToolManager.h"
#include "Tool3DTranslation.h"
#include "Tool3DRotation.h"
#include "Tool3DSurfaceTranslation.h"
#include "GLUtil.h"

using namespace rmsui;


/*
 * IInputEventStream interface
 */
rmsui::IInputEventListener::EventState Tool3DManager::ProcessInputEvent( const rmsui::InputEventData & event )
{
	if ( event.LeftButton.bDown ) {

		if ( event.LeftButton.bWasDown == false ) {
			if ( PickTool( event.nMouseX, event.nMouseY ) ) {

				SetSelected(true);
				m_bInputCaptured = true;

				return IInputEventListener::StartCapture;
			} 

		} else if ( m_bInputCaptured ) {
			
			MoveTool( event.nMouseX, event.nMouseY );

			NotifyToolMoveEvent();

			return IInputEventListener::Consumed;
		}

	} else if ( event.LeftButton.bWasDown && m_bInputCaptured ) {
		SetSelected(false);
		m_bInputCaptured = false;
		return IInputEventListener::EndCapture;
	}

	return IInputEventListener::Ignored;
}


void Tool3DManager::OnToolActionApplied( ITool3D * pTool )
{
	UpdateTool();
}



/*
 * event support
 */
void Tool3DManager::NotifyToolMoveEvent()
{
	std::set<IToolActionEventListener *>::iterator cur(m_vListeners.begin()), end(m_vListeners.end());
	while ( cur != end ) {
		(*cur++)->OnToolMoveEvent();
	}
}


/*
 * Tool control
 */

void Tool3DManager::RegisterTool( ITool3D * pTool )
{
	m_toolSet.RegisterTool(pTool);

	// target of tools is always tool manager (which then re-directs to selected objects...)
	pTool->SetTarget( &m_vTargetGroup );

	pTool->RegisterListener( this );
}


void Tool3DManager::SelectTool( const std::string & sToolName )
{
	m_pTool = m_toolSet.FindTool( sToolName );
	if ( m_pTool == NULL )
		m_pTool = m_toolSet.GetDefaultTool();
	UpdateTool();
}

void Tool3DManager::SelectNextValidTool()
{
	if ( ! m_pTool ) {
		m_pTool = m_toolSet.GetDefaultTool();
		return;
	}

	ITool3D * pCurTool = m_pTool;
	m_pTool = m_toolSet.GetNextTool( pCurTool->GetToolType() );

	if ( m_vTargetGroup.Count() == 0 ) {		// assume any tool is OK
		UpdateTool();
		return;
	}

	while ( m_pTool != pCurTool ) {		// stop once we have cycled

		if ( m_vTargetGroup.SupportsTool( m_pTool->GetToolName() ) ) {
			UpdateTool();
			return;
		}

		m_pTool = m_toolSet.GetNextTool( m_pTool->GetToolType() );
	}

	// ok, if cur tool is valid, it is the only tool, so stick with it,
	//  otherwise drop to default tool
	if ( m_vTargetGroup.SupportsTool( pCurTool->GetToolName() ) )
		m_pTool = pCurTool;
	else
		m_pTool = m_toolSet.GetDefaultTool();

	UpdateTool();
}

void Tool3DManager::EnableCurrentTool( bool bEnable )
{
	m_bToolEnabled = bEnable; 

	// switch to next valid tool, if current one is not valid
	if ( m_pTool && ! m_vTargetGroup.SupportsTool( m_pTool->GetToolName() ) )
		SelectNextValidTool();
	else
		UpdateTool();
}

void Tool3DManager::UpdateTool()
{
	if ( m_vTargetGroup.Count() == 1 )
		m_pTool->SetTarget( m_vTargetGroup.GetTarget(0) );
	else
		m_pTool->SetTarget( &m_vTargetGroup );

	m_pTool->TargetFrameChanged();
}


/*
 * tool rendering support
 */

void Tool3DManager::Render( )
{
	if (m_pTool == NULL || !m_bToolEnabled)
		return;

	// save current opengl matrix stack
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();

	// set up fixed view
	SetupFixedSizeView( );

	glPushMatrix();		// save current modelview matrix

	Wml::Vector3f ptNearPlane;
	Wml::Vector3f rayDirection;
	ComputeFixedSizeTranslation(m_pTool, ptNearPlane, rayDirection);

	// now translate along the ray into the scene. The distance doesn't really matter,
	// as long as it is far enough to prevent anything from being clipped
	// [RMS: no idea why we need the 0.5 multiplier for ortho. Stuff gets clipped otherwise...?? ]
	if ( m_pCamera->GetUsePerspective() )
		ptNearPlane += rayDirection * (float)m_viewState.fFixedView_NearZ;
	else
		ptNearPlane += rayDirection * (float)m_viewState.fFixedView_NearZ * 0.5f;

	// now translate to the screen point
	glTranslated( ptNearPlane.X(), ptNearPlane.Y(), ptNearPlane.Z() );

	// now orient....do pushmatrix before in case tool needs to pop it
	glPushMatrix();
	GLUtil::MakeGLLookAtMatrix( m_pCamera->GetDirection(), m_pCamera->GetUp(),
								m_pCamera->GetLeft() * -1.0, m_viewState.fCurrentTool_LookAt );
	glMultMatrixf( &m_viewState.fCurrentTool_LookAt[0][0] );

	m_pTool->Render( m_viewState );

	glPopMatrix();		// pop the lookat transform

	glPopMatrix();		// restore previous modelview matrix

	// restore opengl matrix stack
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
}

void Tool3DManager::SetupFixedSizeView( )
{
	// read initial opengl viewport and matrices
	memcpy( m_viewState.nViewport, m_pCamera->GetViewport(), sizeof(unsigned int) * 4 );
	glGetDoublev( GL_MODELVIEW_MATRIX, m_viewState.dEyeView_ModelViewMatrix );
	glGetDoublev( GL_PROJECTION_MATRIX, m_viewState.dEyeView_ProjectionMatrix );

	// setup new projection
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	double fWidth = (double)m_pCamera->GetViewport()[2];
	double fHeight = (double)m_pCamera->GetViewport()[3];
	double fAspect = fWidth / fHeight;
	double fFov = (double)m_pCamera->GetHorzFovDegrees();
	double fNearZ = 
		( m_pCamera->GetUsePerspective() ) ? 
		(fHeight / 2.0) / tan( (fFov/2.0) * Wml::Mathd::DEG_TO_RAD ) :
		(fHeight / 2.0) / tan( (fFov/2.0) * Wml::Mathd::DEG_TO_RAD );
	
	// manually compute the frustum (need it for glortho)
	double fHalfAngleRadians = 0.5*fFov*Wml::Mathd::DEG_TO_RAD;
    double m_fFrustumT = fNearZ*tan(fHalfAngleRadians);
    double m_fFrustumB = -m_fFrustumT;
    double m_fFrustumR = fAspect*m_fFrustumT;
    double m_fFrustumL = -m_fFrustumR;

	if ( m_pCamera->GetUsePerspective() )
		glFrustum( m_fFrustumL, m_fFrustumR, m_fFrustumB, m_fFrustumT, fNearZ, 20.0 * fNearZ );
	else {
		// [RMS: this scale makes the ortho space about the same size as the perspective space... ]
		float fScale = 2.0f;
		glOrtho( m_fFrustumL * fScale, m_fFrustumR * fScale, m_fFrustumB * fScale, m_fFrustumT * fScale, fNearZ, 20.0 * fNearZ );
	}

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	// store fixed-view opengl matrices
	m_viewState.fFixedView_NearZ = fNearZ;
	glGetDoublev( GL_MODELVIEW_MATRIX, m_viewState.dFixedView_ModelViewMatrix );
	glGetDoublev( GL_PROJECTION_MATRIX, m_viewState.dFixedView_ProjectionMatrix );
}


void Tool3DManager::ComputeFixedSizeTranslation( const ITool3D * pTool,
												 Wml::Vector3f & nearPlanePoint, 
												 Wml::Vector3f & directionIntoScene )
{
	Wml::Vector3f ptOrigin( pTool->GetOrigin() );
	Wml::Vector3d ptScreen;

	// project origin point into the viewport
	gluProject( ptOrigin.X(), ptOrigin.Y(), ptOrigin.Z(), 
				m_viewState.dEyeView_ModelViewMatrix, 
				m_viewState.dEyeView_ProjectionMatrix,
				(const GLint *)m_viewState.nViewport, 
				& ptScreen.X(), & ptScreen.Y(), & ptScreen.Z() );

	// compute vector through screen-space reference point
	Wml::Vector3d rayOrigin, rayDirection;
	gluUnProject( ptScreen.X(), ptScreen.Y(), m_viewState.fFixedView_NearZ,
					m_viewState.dFixedView_ModelViewMatrix, 
					m_viewState.dFixedView_ProjectionMatrix,
					(const GLint *)m_viewState.nViewport, 
					& rayOrigin.X(), & rayOrigin.Y(), & rayOrigin.Z() );
	gluUnProject( ptScreen.X(), ptScreen.Y(), 2.0 * m_viewState.fFixedView_NearZ,
					m_viewState.dFixedView_ModelViewMatrix, 
					m_viewState.dFixedView_ProjectionMatrix,
					(const GLint *)m_viewState.nViewport, 
					& rayDirection.X(), & rayDirection.Y(), & rayDirection.Z() );
	rayDirection -= rayOrigin;
	rayDirection.Normalize();

	// screen space point has origin at bottom left, but perspective frustum has origin at
	// center of screen, so we have to shift half the vieport width/height. Set the
	// Z depth so it's sitting on the near Z clip plane
	ptScreen.X() -= (double)m_viewState.nViewport[2]/2.0;
	ptScreen.Y() -= (double)m_viewState.nViewport[3]/2.0;
	ptScreen.Z() =  - m_viewState.fFixedView_NearZ;

	nearPlanePoint.X() = (float)ptScreen.X();
	nearPlanePoint.Y() = (float)ptScreen.Y();
	nearPlanePoint.Z() = (float)ptScreen.Z();
	directionIntoScene.X() = (float)rayDirection.X();
	directionIntoScene.Y() = (float)rayDirection.Y();
	directionIntoScene.Z() = (float)rayDirection.Z();

	// [RMS: this seems to un-do the 2.0 scale of the frustum in SetupFixedSizeView() ]
	if (m_pCamera->GetUsePerspective() == false) {
		nearPlanePoint.X() *= 2.0f;
		nearPlanePoint.Y() *= 2.0f;
	}
}

void Tool3DManager::GenerateEyeRay( int nScreenX, int nScreenY, 
								    double dModelViewMatrix[16], double dProjectionMatrix[16], 
								    Wml::Ray3f & ray )
{
	Wml::Vector3d rayOrigin;
	Wml::Vector3d rayDirection;
	gluUnProject( (double)nScreenX, (double)nScreenY, 0.0,
					dModelViewMatrix, dProjectionMatrix,
					(const GLint *)m_viewState.nViewport, 
					& rayOrigin.X(), & rayOrigin.Y(), & rayOrigin.Z() );
	gluUnProject( (double)nScreenX, (double)nScreenY, 1.0,
					dModelViewMatrix, dProjectionMatrix,
					(const GLint *)m_viewState.nViewport, 
					& rayDirection.X(), & rayDirection.Y(), & rayDirection.Z() );
	rayDirection -= rayOrigin;
	rayDirection.Normalize();

	// create ray from eye through screen space point
	ray.Origin().X() = (float)rayOrigin.X();
	ray.Origin().Y() = (float)rayOrigin.Y();
	ray.Origin().Z() = (float)rayOrigin.Z();
	ray.Direction().X() = (float)rayDirection.X();
	ray.Direction().Y() = (float)rayDirection.Y();
	ray.Direction().Z() = (float)rayDirection.Z();
}

void Tool3DManager::TransformRayIntoToolSpace( const ITool3D * pTool, const Wml::Ray3f & initialRay,
											   const Wml::Matrix3f & matViewOrientation,
											   Wml::Ray3f & destRay )
{
	// compute translation applied to tool origin
	Wml::Vector3f ptNearPlane, vToolDirection;
	ComputeFixedSizeTranslation(pTool, ptNearPlane, vToolDirection);
	ptNearPlane += vToolDirection * (float)m_viewState.fFixedView_NearZ;

	// transform ray -> rotation * translation * ray
//	Wml::Ray3f transformedRay;	[RMS] todo remove
	destRay.Origin() = matViewOrientation * (initialRay.Origin() - ptNearPlane);
	destRay.Direction() = matViewOrientation * initialRay.Direction();
	destRay.Direction().Normalize();
}




/*
 * interaction support
 */

bool Tool3DManager::PickTool(int screenX, int screenY)
{
	if (m_pTool == NULL || !m_bToolEnabled)
		return false;
		
	// create eye ray in fixed view
	Wml::Ray3f ray;
	GenerateEyeRay(screenX, screenY, m_viewState.dFixedView_ModelViewMatrix, 
					m_viewState.dFixedView_ProjectionMatrix, ray);

	// compute view orientation matrix
	Wml::Matrix3f matOrientation;
	m_pCamera->GetFrameMatrix(matOrientation);
	matOrientation = matOrientation.Transpose();

	// transform fixed-view eye ray into local basis of tool
	Wml::Ray3f transformedRay;
	TransformRayIntoToolSpace(m_pTool, ray, matOrientation, transformedRay);

	// intersect ray, save hit if it's closer than closest
	float fDepth = Wml::Mathf::MAX_REAL;
	Wml::Vector3f ptHit;
	if ( m_pTool->Intersect( transformedRay, ptHit, fDepth) ) {

		// got a hit....

		// initialize movement
		GenerateEyeRay(screenX, screenY, m_viewState.dEyeView_ModelViewMatrix, 
						m_viewState.dEyeView_ProjectionMatrix, m_targetInfo.eyeViewRay);
		TransformRayIntoToolSpace( m_pTool, ray, matOrientation, m_targetInfo.fixedViewOrientedRay);
		m_pTool->Move(m_targetInfo, m_viewState, true);

		return true;
	}

	return false;
}

void Tool3DManager::MoveTool( int screenX, int screenY )
{
	int iWidth = m_pCamera->GetViewport()[2];
	int iHeight = m_pCamera->GetViewport()[3];

	// create initial pick rays
	GenerateEyeRay(screenX, screenY, m_viewState.dEyeView_ModelViewMatrix, 
					m_viewState.dEyeView_ProjectionMatrix, m_targetInfo.eyeViewRay);
	Wml::Ray3f fixedViewRay;
	GenerateEyeRay(screenX, screenY, m_viewState.dFixedView_ModelViewMatrix, 
					m_viewState.dFixedView_ProjectionMatrix, fixedViewRay);

	// compute view orientation matrix 
	Wml::Matrix3f matOrientation;
	m_pCamera->GetFrameMatrix(matOrientation);
	matOrientation = matOrientation.Transpose();

	// transform fixed-view eye ray into local basis of tool
	TransformRayIntoToolSpace(m_pTool, fixedViewRay, matOrientation, m_targetInfo.fixedViewOrientedRay);

	// compute new position from ray
	Wml::Vector3f & newPosition = 
		m_pTool->Move( m_targetInfo, m_viewState );
}


