// Ryan Schmidt - rms@unknownroad.com
// Copyright (c) 2004-2010
// Distributed under the Boost Software License, Version 1.0.
// http://www.boost.org/LICENSE_1_0.txt

#include "StdAfx.h"
#include "WmlExtVectorUtil.h"

using namespace Wml;

#include <algorithm>


template <class Real>
void Wml::ComputeAlignZAxisMatrix( const Vector3<Real> & vAlignWith,
								  Matrix3<Real> & matrix, bool bInvert )
{
	// compute cosine of angle between vectors
	Real axisDot = vAlignWith.Dot( Vector3<Real>::UNIT_Z );

	// compute rotation axis
	Vector3<Real> axisCross( Vector3<Real>::UNIT_Z.Cross( vAlignWith ) );

	Real fInverter = (bInvert) ? (Real)-1 : (Real)1;

	// apply rotation if necessary
	if (axisCross.SquaredLength() > Wml::Math<Real>::EPSILON) {

		// compute normalized axis and angle, then create rotation around axis
		axisCross.Normalize();
		Real fAngle = Math<Real>::ACos( axisDot / vAlignWith.Length() );
		matrix.FromAxisAngle( axisCross, fAngle * fInverter );

	} else if (axisDot < (Real)0) {
		matrix.FromAxisAngle( Vector3<Real>::UNIT_X, (Real)180 * Math<Real>::DEG_TO_RAD * fInverter );
	} else {
		matrix = Matrix3<Real>::IDENTITY;
	}
}


template <class Real>
void Wml::ComputeAlignAxisMatrix( const Vector3<Real> & vInitial,
								  const Vector3<Real> & vAlignWith, Matrix3<Real> & matrix )
{
	// compute cosine of angle between vectors
	Real axisDot = vAlignWith.Dot( vInitial );

	// compute rotation axis
	Vector3<Real> axisCross( vInitial.Cross( vAlignWith ) );

	// apply rotation if necessary
	if (axisCross.SquaredLength() > Wml::Math<Real>::EPSILON) {

		// compute normalized axis and angle, then create rotation around axis
		axisCross.Normalize();
		Real fAngle = Math<Real>::ACos( axisDot / vAlignWith.Length() );
		matrix.FromAxisAngle( axisCross, fAngle );

	} else if (axisDot < (Real)0) {

		// find some perpendicular vectors
		Wml::Vector3<Real> vPerp1, vPerp2;
		ComputePerpVectors( vInitial, vPerp1, vPerp2 );

		matrix.FromAxisAngle( vPerp1, (Real)180 * Math<Real>::DEG_TO_RAD );
	} else {
		matrix = Matrix3<Real>::IDENTITY;
	}
}


template <class Real>
void Wml::ComputePerpVectors( const Vector3<Real> & vIn,
							  Vector3<Real> & vOut1, Vector3<Real> & vOut2, bool bInIsNormalized )
{
	Wml::Vector3<Real> vPerp(vIn);
	if ( ! bInIsNormalized )
		vPerp.Normalize();

	if ( Wml::Math<Real>::FAbs(vPerp.X()) >= Wml::Math<Real>::FAbs(vPerp.Y())
		 &&   Wml::Math<Real>::FAbs(vPerp.X()) >= Wml::Math<Real>::FAbs(vPerp.Z()) )
    {
        vOut1.X() = -vPerp.Y();
        vOut1.Y() = vPerp.X();
        vOut1.Z() = (Real)0.0;
    }
    else
    {
        vOut1.X() = (Real)0.0;
        vOut1.Y() = vPerp.Z();
        vOut1.Z() = -vPerp.Y();
    }

    vOut1.Normalize();
    vOut2 = vPerp.Cross(vOut1);	
}


template<class Real>
Real Wml::Clamp( const Real & fValue, const Real & fMin, const Real & fMax )
{
	if ( fValue < fMin )
		return fMin;
	else if ( fValue > fMax )
		return fMax;
	else
		return fValue;
}


Wml::Vector3d Wml::VectorCastfd( const Wml::Vector3f & vec )
{
	return Wml::Vector3d( (double)vec.X(), (double)vec.Y(), (double)vec.Z() );
}

Wml::Vector3f Wml::VectorCastdf( const Wml::Vector3d & vec )
{
	return Wml::Vector3f( (float)vec.X(), (float)vec.Y(), (float)vec.Z() );
}



template <class Real>
void Wml::ToGLMatrix( const Matrix3<Real> & matrix, Real glMatrix[16] )
{
	for (int r = 0; r < 3; ++r)
		for (int c = 0; c < 4; ++c)
			glMatrix[c*4 + r] = (c < 3) ? matrix(r,c) : 0;
	glMatrix[3] = glMatrix[7] = glMatrix[11] = 0;
	glMatrix[15] = 1;
}




template <class Real>
Wml::Vector3<Real> Wml::Center( const Wml::AxisAlignedBox3<Real> & box )
{
	Wml::Vector3<Real> v;
	v.X() = (Real)0.5 * ( box.GetXMax() + box.GetXMin() );
	v.Y() = (Real)0.5 * ( box.GetYMax() + box.GetYMin() );
	v.Z() = (Real)0.5 * ( box.GetZMax() + box.GetZMin() );
	return v;
}

template <class Real>
Wml::Vector2<Real> Wml::Center( const Wml::AxisAlignedBox2<Real> & box )
{
	Wml::Vector2<Real> v;
	v.X() = (Real)0.5 * ( box.GetXMax() + box.GetXMin() );
	v.Y() = (Real)0.5 * ( box.GetYMax() + box.GetYMin() );
	return v;
}


template <class Real>
Real Wml::GetDimension( const Wml::AxisAlignedBox3<Real> & box, int nDimension )
{
	switch(nDimension) {
		case 0:
			return box.GetXMax() - box.GetXMin();
		case 1:
			return box.GetYMax() - box.GetYMin();
		case 2:
			return box.GetZMax() - box.GetZMin();
		default:
			return (Real)0;
	};
}


template <class Real>
Real Wml::GetDimension( const Wml::AxisAlignedBox2<Real> & box, int nDimension )
{
	switch(nDimension) {
		case 0:
			return box.GetXMax() - box.GetXMin();
		case 1:
			return box.GetYMax() - box.GetYMin();
		default:
			return (Real)0;
	};
}


template <class Real>
Real Wml::MaxDimension( const Wml::AxisAlignedBox3<Real> & box )
{
	Real width = box.GetXMax() - box.GetXMin();
	Real height = box.GetYMax() - box.GetYMin();
	Real depth = box.GetZMax() - box.GetZMin();
	return std::max<Real>( width, std::max<Real>(height,depth) );
}


void Wml::Union( Wml::AxisAlignedBox3f & dest, const Wml::AxisAlignedBox3f & with )
{
	if ( with.GetXMin() < dest.XMin() )
		dest.XMin() = with.GetXMin();
	if ( with.GetXMax() > dest.XMax() )
		dest.XMax() = with.GetXMax();
	if ( with.GetYMin() < dest.YMin() )
		dest.YMin() = with.GetYMin();
	if ( with.GetYMax() > dest.YMax() )
		dest.YMax() = with.GetYMax();
	if ( with.GetZMin() < dest.ZMin() )
		dest.ZMin() = with.GetZMin();
	if ( with.GetZMax() > dest.ZMax() )
		dest.ZMax() = with.GetZMax();
}

void WMLEXT_ITEM Wml::Union( Wml::AxisAlignedBox3f & dest, const Wml::Vector3f & point )
{
	if ( point.X() < dest.XMin() )
		dest.XMin() = point.X();
	if ( point.X() > dest.XMax() )
		dest.XMax() = point.X();
	if ( point.Y() < dest.YMin() )
		dest.YMin() = point.Y();
	if ( point.Y() > dest.YMax() )
		dest.YMax() = point.Y();
	if ( point.Z() < dest.ZMin() )
		dest.ZMin() = point.Z();
	if ( point.Z() > dest.ZMax() )
		dest.ZMax() = point.Z();
}

void Wml::Union( Wml::AxisAlignedBox2f & dest, const Wml::AxisAlignedBox2f & with )
{
	if ( with.GetXMin() < dest.XMin() )
		dest.XMin() = with.GetXMin();
	if ( with.GetXMax() > dest.XMax() )
		dest.XMax() = with.GetXMax();
	if ( with.GetYMin() < dest.YMin() )
		dest.YMin() = with.GetYMin();
	if ( with.GetYMax() > dest.YMax() )
		dest.YMax() = with.GetYMax();
}


float Wml::Volume( const Wml::AxisAlignedBox3f & box )
{
	return ( box.GetXMax() - box.GetXMin() ) * (box.GetYMax() - box.GetYMin()) * (box.GetZMax() - box.GetZMin());
}

template <class Real>
void Wml::FitAABox( const Wml::Vector3<Real> & vCenter, Real fRadius, Wml::AxisAlignedBox3<Real> & aaBox )
{
	aaBox.XMin() = vCenter.X() - fRadius;
	aaBox.XMax() = vCenter.X() + fRadius;
	aaBox.YMin() = vCenter.Y() - fRadius;
	aaBox.YMax() = vCenter.Y() + fRadius;
	aaBox.ZMin() = vCenter.Z() - fRadius;
	aaBox.ZMax() = vCenter.Z() + fRadius;
}



template <class Real>
void Wml::ConvertBox( const Wml::AxisAlignedBox3<Real> & fromAABox, Wml::Box3<Real> & toOBox )
{
	toOBox.Center() = Wml::Vector3<Real>(
		(Real)0.5 * (fromAABox.GetXMin() + fromAABox.GetXMax()),
		(Real)0.5 * (fromAABox.GetYMin() + fromAABox.GetYMax()),
		(Real)0.5 * (fromAABox.GetZMin() + fromAABox.GetZMax()) );

	toOBox.Axis(0) = Wml::Vector3<Real>::UNIT_X;
	toOBox.Axis(1) = Wml::Vector3<Real>::UNIT_Y;
	toOBox.Axis(2) = Wml::Vector3<Real>::UNIT_Z;

	toOBox.Extent(0) = fromAABox.GetXMax() - toOBox.Center().X();
	toOBox.Extent(1) = fromAABox.GetYMax() - toOBox.Center().Y();
	toOBox.Extent(2) = fromAABox.GetZMax() - toOBox.Center().Z();
}


template <class Real>
void Wml::ScaleBox( Wml::Box3<Real> & toOBox, Real fScaleX, Real fScaleY, Real fScaleZ )
{
	toOBox.Extent(0) *= fScaleX;
	toOBox.Extent(1) *= fScaleY;
	toOBox.Extent(2) *= fScaleZ;
}




bool Wml::Contained( const Wml::AxisAlignedBox3f & box, float fX, float fY, float fZ )
{
	return (fX >= box.GetXMin() && fX <= box.GetXMax() 
		&& fY >= box.GetYMin() && fY <= box.GetYMax()
		&& fZ >= box.GetZMin() && fZ <= box.GetZMax() );
}

bool Wml::Contained( const Wml::AxisAlignedBox2f & box, float fX, float fY )
{
	return (fX >= box.GetXMin() && fX <= box.GetXMax() 
		&& fY >= box.GetYMin() && fY <= box.GetYMax() );
}


template<class Real>
void Wml::Translate( Wml::AxisAlignedBox2<Real> & box, Real fX, Real fY, bool bRelative )
{
	if ( bRelative ) {
		box.XMin() += fX;
		box.XMax() += fX;
		box.YMin() += fY;
		box.YMax() += fY;
	} else {
		Real fWidth = Wml::GetDimension(box, 0);
		Real fHeight = Wml::GetDimension(box, 1);
		box.XMin() = fX;
		box.XMax() = fX + fWidth;
		box.YMin() = fY;
		box.YMax() = fY + fHeight;
	}
}


namespace Wml
{
template WMLEXT_ITEM void ToGLMatrix( const Matrix3<float> & matrix, float glMatrix[16] );
template WMLEXT_ITEM void ToGLMatrix( const Matrix3<double> & matrix, double glMatrix[16] );

template WMLEXT_ITEM void ComputeAlignZAxisMatrix( const Vector3<float> & vAlignWith, Matrix3<float> & matrix, bool bInvert );
template WMLEXT_ITEM void ComputeAlignZAxisMatrix( const Vector3<double> & vAlignWith, Matrix3<double> & matrix, bool bInvert );

template WMLEXT_ITEM void ComputeAlignAxisMatrix( const Vector3<float> & vInitial, const Vector3<float> & vAlignWith, Matrix3<float> & matrix );
template WMLEXT_ITEM void ComputeAlignAxisMatrix( const Vector3<double> & vInitial, const Vector3<double> & vAlignWith, Matrix3<double> & matrix );

template WMLEXT_ITEM void ComputePerpVectors( const Vector3<float> & vIn, Vector3<float> & vOut1, Vector3<float> & vOut2, bool bInIsNormalized );
template WMLEXT_ITEM void ComputePerpVectors( const Vector3<double> & vIn, Vector3<double> & vOut1, Vector3<double> & vOut2, bool bInIsNormalized );

template WMLEXT_ITEM Wml::Vector3<float> Wml::Center( const Wml::AxisAlignedBox3<float> & box );
template WMLEXT_ITEM Wml::Vector3<double> Wml::Center( const Wml::AxisAlignedBox3<double> & box );

template WMLEXT_ITEM Wml::Vector2<float> Wml::Center( const Wml::AxisAlignedBox2<float> & box );
template WMLEXT_ITEM Wml::Vector2<double> Wml::Center( const Wml::AxisAlignedBox2<double> & box );


template WMLEXT_ITEM void Wml::FitAABox( const Wml::Vector3<float> & vCenter, float fRadius, Wml::AxisAlignedBox3<float> & aaBox );
template WMLEXT_ITEM void Wml::FitAABox( const Wml::Vector3<double> & vCenter, double fRadius, Wml::AxisAlignedBox3<double> & aaBox );

template WMLEXT_ITEM void Wml::ConvertBox( const Wml::AxisAlignedBox3<float> & fromAABox, Wml::Box3<float> & toOBox );
template WMLEXT_ITEM void Wml::ConvertBox( const Wml::AxisAlignedBox3<double> & fromAABox, Wml::Box3<double> & toOBox );

template WMLEXT_ITEM void Wml::ScaleBox( Wml::Box3<float> & toOBox, float fScaleX, float fScaleY, float fScaleZ );
template WMLEXT_ITEM void Wml::ScaleBox( Wml::Box3<double> & toOBox, double fScaleX, double fScaleY, double fScaleZ );

template WMLEXT_ITEM float Wml::MaxDimension( const Wml::AxisAlignedBox3<float> & box );
template WMLEXT_ITEM double Wml::MaxDimension( const Wml::AxisAlignedBox3<double> & box );

template WMLEXT_ITEM float Wml::GetDimension( const Wml::AxisAlignedBox3<float> & box, int nDimension );
template WMLEXT_ITEM double Wml::GetDimension( const Wml::AxisAlignedBox3<double> & box, int nDimension );

template WMLEXT_ITEM float Wml::GetDimension( const Wml::AxisAlignedBox2<float> & box, int nDimension );
template WMLEXT_ITEM double Wml::GetDimension( const Wml::AxisAlignedBox2<double> & box, int nDimension );

template WMLEXT_ITEM void Wml::Translate( Wml::AxisAlignedBox2<float> & box, float fX, float fY, bool bRelative );
template WMLEXT_ITEM void Wml::Translate( Wml::AxisAlignedBox2<double> & box, double fX, double fY, bool bRelative );

template WMLEXT_ITEM float Wml::Clamp( const float & fValue, const float & fMin, const float & fMax );
template WMLEXT_ITEM double Wml::Clamp( const double & fValue, const double & fMin, const double & fMax );
}

