// Ryan Schmidt - rms@unknownroad.com
// Copyright (c) 2004-2010
// Distributed under the Boost Software License, Version 1.0.
// http://www.boost.org/LICENSE_1_0.txt


#include "StdAfx.h"
#include "WmlPolyLine3.h"
#include "WmlSegment3.h"

#include <WmlDistVec3Lin3.h>
#include <limits>
#include <fstream>

#include "rmsdebug.h"

using namespace Wml;


template <class Real>
PolyLine3<Real>::PolyLine3()
{
}

template <class Real>
PolyLine3<Real>::PolyLine3( const PolyLine3<Real> & copy )
{
	m_vVertices = copy.m_vVertices;
}

template <class Real>
PolyLine3<Real>::~PolyLine3()
{
}

template <class Real>
void PolyLine3<Real>::Reverse()
{
	size_t nCount = m_vVertices.size();
	for (unsigned int i = 0; i < nCount/2; ++i) {
		Wml::Vector3<Real> tmp = m_vVertices[i];
		m_vVertices[i] = m_vVertices[ (nCount-1)-i ];
		m_vVertices[ (nCount-1)-i ] = tmp;
	}	
}


// bounding things
template <class Real>
void PolyLine3<Real>::BoundingBox( Wml::AxisAlignedBox3<Real> & box ) const
{
	if( IsEmpty() ) {
		box = Wml::AxisAlignedBox3<Real>();
		return;
	}

	box = Wml::AxisAlignedBox3<Real>( m_vVertices[0].X(), m_vVertices[0].X(), 
		m_vVertices[0].Y(), m_vVertices[0].Y(), m_vVertices[0].Z(), m_vVertices[0].Z() );
	size_t nVertices = m_vVertices.size();
	for (unsigned int i = 1; i < nVertices; ++i) {
		const Wml::Vector3<Real> & vtx = m_vVertices[i];
		if ( vtx.X() < box.XMin() ) 
			box.XMin() = vtx.X();
		else if (vtx.X() > box.XMax() )
			box.XMax() = vtx.X();
		if ( vtx.Y() < box.YMin() ) 
			box.YMin() = vtx.Y();
		else if (vtx.Y() > box.YMax() )
			box.YMax() = vtx.Y();
		if ( vtx.Z() < box.ZMin() ) 
			box.ZMin() = vtx.Z();
		else if (vtx.Z() > box.ZMax() )
			box.ZMax() = vtx.Z();
	}
}

// translation / rotation functions
template <class Real>
void PolyLine3<Real>::Transform( const Wml::Matrix3<Real> & transform )
{
	size_t nCount = m_vVertices.size();
	for (unsigned int i = 0; i < nCount; ++i)
		m_vVertices[i] = transform * m_vVertices[i];
}

template <class Real>
void PolyLine3<Real>::Translate( Real fX, Real fY, Real fZ )
{
	size_t nCount = m_vVertices.size();
	for (unsigned int i = 0; i < nCount; ++i) {
		m_vVertices[i].X() += fX;
		m_vVertices[i].Y() += fY;
		m_vVertices[i].Z() += fZ;
	}
}




// Polygon simplification
// code adapted from: http://softsurfer.com/Archive/algorithm_0205/algorithm_0205.htm

// dist_Point_to_Segment(): get the distance of a point to a segment.
//    Input:  a Point P and a Segment S (in any dimension)
//    Return: the shortest distance from P to S
template<class Real>
static Real
dist_Point_to_Segment( Wml::Vector3<Real> & P, Wml::Segment3<Real> & S)
{
	Wml::Vector3<Real> & v = S.Direction();
	Wml::Vector3<Real> & w = P - S.Origin();

    Real c1 = w.Dot(v);
    if ( c1 <= 0 )
		return w.Length();

    Real c2 = v.Dot(v);
    if ( c2 <= c1 )
		return (P - (S.Origin() + S.Direction())).Length();

    Real b = c1 / c2;
	Wml::Vector3<Real> Pb = S.Origin() + v * b;
	return (P - Pb).Length();
}


// simplifyDP():
//  This is the Douglas-Peucker recursive simplification routine
//  It just marks vertices that are part of the simplified polyline
//  for approximating the polyline subchain v[j] to v[k].
//    Input:  tol = approximation tolerance
//            v[] = polyline array of vertex points
//            j,k = indices for the subchain v[j] to v[k]
//    Output: mk[] = array of markers matching vertex array v[]
template<class Real>
static void
simplifyDP( Real tol, std::vector<Wml::Vector3<Real> > & v, int j, int k, std::vector<bool> & mk )
{
    if (k <= j+1) // there is nothing to simplify
        return;

    // check for adequate approximation by segment S from v[j] to v[k]
    int     maxi = j;          // index of vertex farthest from S
    Real   maxd2 = 0;         // distance squared of farthest vertex
    Real   tol2 = tol * tol;  // tolerance squared
	Wml::Segment3<Real> S;    // segment from v[j] to v[k]
		S.Origin() = v[j];
		S.Direction() = v[k] - v[j];
	Wml::Vector3<Real> u( S.Direction() );  // segment direction vector
	Real cu = u.SquaredLength();            // segment length squared

    // test each vertex v[i] for max distance from S
    // compute using the Feb 2001 Algorithm's dist_Point_to_Segment()
    // Note: this works in any dimension (2D, 3D, ...)
	Wml::Vector3<Real> w;
	Wml::Vector3<Real> Pb;   // base of perpendicular from v[i] to S
	Real b, cw, dv2;         // dv2 = distance v[i] to S squared

    for (int i = j+1; i < k; i++)
    {
        // compute distance squared
		w = v[i] - S.Origin();
        cw = w.Dot(u);
        if ( cw <= 0 )
			dv2 = (v[i] - S.Origin()).SquaredLength();
        else if ( cu <= cw )
			dv2 = (v[i] - (S.Origin() + S.Direction())).SquaredLength();
        else {
            b = cw / cu;
            Pb = S.Origin() + u * b;
			dv2 = (v[i] - Pb).SquaredLength();
        }
        // test with current max distance squared
        if (dv2 <= maxd2)
            continue;
        // v[i] is a new max vertex
        maxi = i;
        maxd2 = dv2;
    }
    if (maxd2 > tol2)        // error is worse than the tolerance
    {
        // split the polyline at the farthest vertex from S
        mk[maxi] = true;      // mark v[maxi] for the simplified polyline
        // recursively simplify the two subpolylines at v[maxi]
        simplifyDP( tol, v, j, maxi, mk );  // polyline v[j] to v[maxi]
        simplifyDP( tol, v, maxi, k, mk );  // polyline v[maxi] to v[k]
    }
    // else the approximation is OK, so ignore intermediate vertices
    return;
}



template <class Real>
void PolyLine3<Real>::Simplify( float fToleranceCoeff, bool bSimplifyStraightLines )
{
	Wml::AxisAlignedBox3<Real> box;
	BoundingBox(box);
	Real maxDim = std::max<Real>( box.GetXMax() - box.GetXMin(), box.GetYMax() - box.GetYMin() );
	maxDim = std::max<Real>( box.GetZMax() - box.GetZMin(), maxDim );
	Real fTol = maxDim * fToleranceCoeff;

	int n = (int)m_vVertices.size();

    int    i, k, pv;            // misc counters
    Real  tol2 = fTol * fTol;       // tolerance squared
	std::vector<Wml::Vector3<Real> > vt(n);  // vertex buffer
	std::vector<bool> mk(n, ! bSimplifyStraightLines);			 // marker buffer

    // STAGE 1.  Vertex Reduction within tolerance of prior vertex cluster
    vt[0] = m_vVertices[0];              // start at the beginning
    for (i = k = 1, pv = 0; i < n; i++) {
		if ( (m_vVertices[i] - m_vVertices[pv]).SquaredLength() < tol2 )
            continue;
        vt[k++] = m_vVertices[i];
        pv = i;
    }
    if (pv < n-1)
        vt[k++] = m_vVertices[n-1];      // finish at the end

    // STAGE 2.  Douglas-Peucker polyline simplification
	if (bSimplifyStraightLines) {
		mk[0] = mk[k-1] = 1;       // mark the first and last vertices
		simplifyDP( fTol * (Real)0.25, vt, 0, k-1, mk );
	}

	// copy marked vertices back to this polygon
	m_vVertices.resize(0);
	for (i = 0; i < k; ++i) {
		if (mk[i])
			m_vVertices.push_back( vt[i] );
	}

	return;
}

template <class Real>
Real PolyLine3<Real>::SqrDistance( const Wml::Vector3<Real> & vPoint, Wml::Vector3<Real> * pNearest = NULL )
{
	Real fMinSqrDist = std::numeric_limits<Real>::max();

	Real fMinParam = 0.0f;
	unsigned int nMinIndex = std::numeric_limits<unsigned int>::max();

	size_t nCount = m_vVertices.size();
	for ( unsigned int i = 0; i < nCount; ++i ) {

		Wml::Segment3<Real> seg;
		seg.Origin() = m_vVertices[i];
		seg.Direction() = m_vVertices[ (i+1)%nCount ] - seg.Origin();

		Real fParam = 0;
		Real fSqrDist = Wml::SqrDistance( vPoint, seg, &fParam );

		if ( fSqrDist < fMinSqrDist ) {
			fMinSqrDist = fSqrDist;
			fMinParam = fParam;
			nMinIndex = i;
		}
	}

	if ( pNearest != NULL && nMinIndex != std::numeric_limits<unsigned int>::max() ) {
		Wml::Segment3<Real> seg;
		seg.Origin() = m_vVertices[nMinIndex];
		seg.Direction() = m_vVertices[ (nMinIndex+1)%nCount ] - seg.Origin();

		*pNearest = seg.Origin() + fMinParam * seg.Direction();
	}

	return fMinSqrDist;
}


template <class Real>
bool PolyLine3<Real>::Load( const char * pFilename )
{
	std::ifstream in(pFilename);
	if (!in)
		return false;

	unsigned int nCount;
	in >> nCount;

	m_vVertices.resize(nCount);

	for ( unsigned int i = 0; i < nCount; ++i ) {

		Real fX, fY, fZ;
		in >> fX >> fY >> fZ;
		if (!in) {
			in.close();
			return false;
		}

		m_vVertices[i] = Wml::Vector3<Real>(fX,fY,fZ);
	}

	in.close();
	return true;
}

template <class Real>
bool PolyLine3<Real>::Save( const char * pFilename )
{
	std::ofstream out(pFilename);
	if (!out)
		return false;

	size_t nCount = m_vVertices.size();
	out << (unsigned int)nCount << std::endl;

	for ( unsigned int i = 0; i < nCount; ++i ) {
		out << m_vVertices[i].X() << " " << m_vVertices[i].Y() << " " << m_vVertices[i].Z() << std::endl;
	}

	bool bReturn = (out) ? true : false;
	out.close();
	return bReturn;
}



//----------------------------------------------------------------------------
// explicit instantiation
//----------------------------------------------------------------------------
namespace Wml
{
template WML_ITEM class PolyLine3<float>;
template WML_ITEM class PolyLine3<double>;
}
//----------------------------------------------------------------------------
