// Ryan Schmidt - rms@unknownroad.com
// Copyright (c) 2004-2010
// Distributed under the Boost Software License, Version 1.0.
// http://www.boost.org/LICENSE_1_0.txt


#include "StdAfx.h"

#include "WmlSerializerRecord.h"

#include <limits>

using namespace Wml;


const Wml::SerializerRecord::RecordType Wml::SerializerRecord::UnknownType = std::numeric_limits<unsigned int>::max();

// [RMS: leave first 100 entries for "special" records...]
SerializerRecord::RecordID SerializerRecord::s_recordIDGen = 100;

SerializerRecord::RecordID SerializerRecord::GenUniqueRecordID()
{
	return s_recordIDGen++;
}


SerializerRecord::SerializerRecord( RecordType nType )
	: m_recordID( GenUniqueRecordID() ), m_recordType(nType)
{
}

SerializerRecord::SerializerRecord( RecordID nID, RecordType nType )
	: m_recordID(nID), m_recordType(nType)
{
}

SerializerRecord::~SerializerRecord()
{
}


template<typename T>
void AppendValue( std::vector<unsigned char> & vData, const T & value )
{
	size_t nCur = vData.size();
	int nBytes = sizeof(T);
	vData.resize( nCur + nBytes );
	
	memcpy( &vData[nCur], &value, nBytes );
}

template<typename T>
void AppendValue( std::vector<unsigned char> & vData, const T * pValue, unsigned int nCount )
{
	size_t nCur = vData.size();
	int nBytes = sizeof(T) * nCount;
	vData.resize( nCur + nBytes );
	
	memcpy( &vData[nCur], pValue, nBytes );
}


#define SIMPLE_SETVALUE( key, value ) \
	std::vector<unsigned char> vTemp; \
	AppendValue( vTemp, value ); \
	m_values[ key ] = vTemp;

#define POINTER_SETVALUE( key, pValue, count ) \
	std::vector<unsigned char> vTemp; \
	AppendValue( vTemp, pValue, count ); \
	m_values[ key ] = vTemp;




template<typename T>
unsigned int ReadValue( const std::vector<unsigned char> & vData, unsigned int nOffset, T & value )
{
	unsigned int nBytes = sizeof(T);
	
	memcpy( &value, &vData[nOffset], nBytes );
	return nOffset + nBytes;
}

template<typename T>
unsigned int ReadValue( const std::vector<unsigned char> & vData, unsigned int nOffset, T * pValue, unsigned int nCount )
{
	unsigned int nBytes = sizeof(T) * nCount;
	
	memcpy( pValue, &vData[nOffset], nBytes );
	return nOffset + nBytes;
}


#define SIMPLE_GETVALUE( key, value, defaultValue ) \
	if ( m_values.find( key ) != m_values.end() ) { \
		ReadValue( m_values[key], 0, value ); \
		return true; \
	} else { \
		value = defaultValue; \
		return false; \
	} 


void SerializerRecord::SetValue( const std::string & key, unsigned int nValue )
{
	SIMPLE_SETVALUE( key, nValue );
}


void SerializerRecord::SetValue( const std::string & key, int nValue )
{
	SIMPLE_SETVALUE( key, nValue );
}


void SerializerRecord::SetValue( const std::string & key, float fValue )
{
	SIMPLE_SETVALUE( key, fValue );
}


void SerializerRecord::SetValue( const std::string & key, double dValue )
{
	SIMPLE_SETVALUE( key, dValue );
}


void SerializerRecord::SetValue( const std::string & key, const Vector3f & vValue )
{
	POINTER_SETVALUE( key, (const float *)(vValue), 3 );
}

void SerializerRecord::SetValue( const std::string & key, const Vector2f & vValue )
{
	POINTER_SETVALUE( key, (const float *)(vValue), 2 );
}

void SerializerRecord::SetValue( const std::string & key, const Frame3f & vValue )
{
	std::vector<unsigned char> vTemp;
	AppendValue( vTemp, (const float *)vValue.FrameMatrix(), 9 );
	AppendValue( vTemp, (const float *)vValue.FrameInverse(), 9 );
	AppendValue( vTemp, (const float *)vValue.Origin(), 3 );
	m_values[ key ] = vTemp;	
}

void SerializerRecord::SetValue( const std::string & key, AxisAlignedBox2f & vBox )
{
	std::vector<unsigned char> vTemp;
	AppendValue( vTemp, vBox.XMin() );
	AppendValue( vTemp, vBox.XMax() );
	AppendValue( vTemp, vBox.YMin() );
	AppendValue( vTemp, vBox.YMax() );
	m_values[ key ] = vTemp;	
}




void SerializerRecord::SetValue( const std::string & key, const std::vector<float> & vValue )
{
	std::vector<unsigned char> vTemp;
	AppendValue( vTemp, (unsigned int)vValue.size() );
	AppendValue( vTemp, &vValue[0], (unsigned int)vValue.size() );
	m_values[ key ] = vTemp;
}

void SerializerRecord::SetValue( const std::string & key, const std::vector<SerializerRecord::RecordID> & vValue )
{
	std::vector<unsigned char> vTemp;
	AppendValue( vTemp, (unsigned int)vValue.size() );
	AppendValue( vTemp, &vValue[0], (unsigned int)vValue.size() );
	m_values[ key ] = vTemp;
}

void SerializerRecord::SetValue( const std::string & key, const std::vector<Wml::Vector3f> & vValue )
{
	std::vector<unsigned char> vTemp;
	unsigned int nCount = (unsigned int)vValue.size();
	AppendValue( vTemp, nCount );
	for ( unsigned int i = 0; i < nCount; ++i )
		AppendValue( vTemp, (const float *)vValue[i], 3 );
	m_values[ key ] = vTemp;
}

void SerializerRecord::SetValue( const std::string & key, const std::vector<Wml::Vector2f> & vValue )
{
	std::vector<unsigned char> vTemp;
	unsigned int nCount = (unsigned int)vValue.size();
	AppendValue( vTemp, nCount );
	for ( unsigned int i = 0; i < nCount; ++i )
		AppendValue( vTemp, (const float *)vValue[i], 2 );
	m_values[ key ] = vTemp;
}


void SerializerRecord::SetValue( const std::string & key, const std::string & value )
{
	std::vector<unsigned char> vTemp;
	unsigned int nCount = (unsigned int)value.length() + 1;
	AppendValue( vTemp, nCount );
	AppendValue( vTemp, value.c_str(), nCount );
	m_values[ key ] = vTemp;
}




bool SerializerRecord::GetValue( const std::string & key, unsigned int & nValue, unsigned int nDefault )
{
	SIMPLE_GETVALUE( key, nValue, nDefault );
}


bool SerializerRecord::GetValue( const std::string & key, int & nValue, int nDefault )
{
	SIMPLE_GETVALUE( key, nValue, nDefault );
}


bool SerializerRecord::GetValue( const std::string & key, float & fValue, float fDefault )
{
	SIMPLE_GETVALUE( key, fValue, fDefault );
}


bool SerializerRecord::GetValue( const std::string & key, double & dValue, double dDefault )
{
	SIMPLE_GETVALUE( key, dValue, dDefault );
}


bool SerializerRecord::GetValue( const std::string & key, Vector3f & vValue, const Vector3f & vDefault )
{
	if ( m_values.find( key ) != m_values.end() ) {
		ReadValue( m_values[key], 0, (float *)vValue, 3 );
		return true;
	} else {
		vValue = vDefault;
		return false;
	} 
}

bool SerializerRecord::GetValue( const std::string & key, Vector2f & vValue, const Vector2f & vDefault )
{
	if ( m_values.find( key ) != m_values.end() ) {
		ReadValue( m_values[key], 0, (float *)vValue, 2 );
		return true;
	} else {
		vValue = vDefault;
		return false;
	} 
}

bool SerializerRecord::GetValue( const std::string & key, Frame3f & vFrame, const Frame3f & vDefault )
{
	if ( m_values.find( key ) != m_values.end() ) {
		Wml::Matrix3f matFrame, matInverse;
		Wml::Vector3f ptOrigin;

		unsigned int nOffset = ReadValue( m_values[key], 0, (float *)matFrame, 9 );
		nOffset = ReadValue( m_values[key], nOffset, (float *)matInverse, 9 );
		nOffset = ReadValue( m_values[key], nOffset, (float *)ptOrigin, 3 );

		vFrame = Wml::Frame3f(matFrame, matInverse, ptOrigin);

		return true;
	} else {
		vFrame = vDefault;
		return false;
	} 
}

bool SerializerRecord::GetValue( const std::string & key, AxisAlignedBox2f & vBox, const AxisAlignedBox2f & vDefault )
{
	if ( m_values.find( key ) != m_values.end() ) {

		unsigned int nOffset = ReadValue( m_values[key], 0, vBox.XMin() );
		nOffset = ReadValue( m_values[key], nOffset, vBox.XMax() );
		nOffset = ReadValue( m_values[key], nOffset, vBox.YMin() );
		nOffset = ReadValue( m_values[key], nOffset, vBox.YMax() );

		return true;
	} else {
		vBox = vDefault;
		return false;
	} 
}


bool SerializerRecord::GetValue( const std::string & key, std::vector<float> & vValue )
{
	if ( m_values.find( key ) != m_values.end() ) {

		unsigned int nCount;
		unsigned int nOffset = ReadValue( m_values[key], 0, nCount );

		if ( nCount > vValue.capacity() )
			return false;

		vValue.resize(nCount);
		ReadValue( m_values[key], nOffset, &vValue[0], nCount );

		return true;
	} else {
		return false;
	} 
}


bool SerializerRecord::GetValue( const std::string & key, std::vector<SerializerRecord::RecordID> & vValue )
{
	if ( m_values.find( key ) != m_values.end() ) {

		unsigned int nCount;
		unsigned int nOffset = ReadValue( m_values[key], 0, nCount );

		if ( nCount > vValue.capacity() )
			return false;

		vValue.resize(nCount);
		ReadValue( m_values[key], nOffset, &vValue[0], nCount );

		return true;
	} else {
		return false;
	} 
}


bool SerializerRecord::GetValue( const std::string & key, std::vector<Wml::Vector3f> & vValue )
{
	if ( m_values.find( key ) != m_values.end() ) {

		unsigned int nCount;
		unsigned int nOffset = ReadValue( m_values[key], 0, nCount );

		if ( nCount > vValue.capacity() )
			return false;

		vValue.resize(nCount);

		float buf[3];
		for ( unsigned int i = 0; i < nCount; ++i ) {
			nOffset = ReadValue( m_values[key], nOffset, buf, 3 );
			vValue[i] = Wml::Vector3f( buf[0], buf[1], buf[2] );
		}

		return true;
	} else {
		return false;
	} 
}

bool SerializerRecord::GetValue( const std::string & key, std::vector<Wml::Vector2f> & vValue )
{
	if ( m_values.find( key ) != m_values.end() ) {

		unsigned int nCount;
		unsigned int nOffset = ReadValue( m_values[key], 0, nCount );

		if ( nCount > vValue.capacity() )
			return false;

		vValue.resize(nCount);

		float buf[3];
		for ( unsigned int i = 0; i < nCount; ++i ) {
			nOffset = ReadValue( m_values[key], nOffset, buf, 2 );
			vValue[i] = Wml::Vector2f( buf[0], buf[1] );
		}

		return true;
	} else {
		return false;
	} 
}

bool SerializerRecord::GetValue( const std::string & key, std::string & value )
{
	if ( m_values.find( key ) != m_values.end() ) {

		unsigned int nCount;
		unsigned int nOffset = ReadValue( m_values[key], 0, nCount );

		std::vector<char> vString(nCount);
		ReadValue( m_values[key], nOffset, &vString[0], nCount );

		value = std::string( &vString[0] );

		return true;
	} else {
		return false;
	} 
}



void SerializerRecord::GetEmptyRecordBytes( std::vector< unsigned char > & vBytes )
{
	AppendValue( vBytes, 0 );

	// invalid type (hopefully nobody uses this...)
	AppendValue( vBytes, std::numeric_limits<unsigned int>::max() );

	// zero records
	AppendValue( vBytes, 0 );
}


bool SerializerRecord::GetBytes( std::vector<unsigned char> & vBytes )
{
	AppendValue( vBytes, m_recordID );
	AppendValue( vBytes, m_recordType );

	AppendValue( vBytes, (unsigned int)m_values.size() );

	std::map< std::string, std::vector<unsigned char> >::iterator
		cur( m_values.begin() ), end( m_values.end() );
	while ( cur != end ) {

		const std::string & name = (*cur).first;
		const std::vector<unsigned char> & value = (*cur).second;
		
		unsigned int nLen = (unsigned int)strlen( name.c_str() );
		AppendValue( vBytes, nLen );
		AppendValue( vBytes, name.c_str(), nLen );

		nLen = (unsigned int)value.size();
		AppendValue( vBytes, nLen );
		AppendValue( vBytes, &value[0], nLen );

		++cur;
	}

	return true;
}

bool SerializerRecord::SetBytes( const std::vector<unsigned char> & vBytes )
{
	unsigned int nOffset = 0;
	nOffset = ReadValue( vBytes, nOffset, m_recordID );
	nOffset = ReadValue( vBytes, nOffset, m_recordType );

	unsigned int nMapEntries;
	nOffset = ReadValue( vBytes, nOffset, nMapEntries );

	std::vector<char> namebuf;
	std::vector<unsigned char> buf;
	for ( unsigned int i = 0; i < nMapEntries; ++i ) {

		unsigned int nStrLen;
		nOffset = ReadValue( vBytes, nOffset, nStrLen);

		namebuf.resize( nStrLen + 1 );
		nOffset = ReadValue( vBytes, nOffset, & namebuf[0], nStrLen );
		namebuf[ nStrLen ] = '\0';
		std::string name( & namebuf[0] );

		unsigned int nDataLen;
		nOffset = ReadValue( vBytes, nOffset, nDataLen );

		buf.resize( nDataLen + 1 );
		nOffset = ReadValue( vBytes, nOffset, & buf[0], nDataLen );

		m_values[name] = buf;
	}

	return true;
}
